-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--							    "BuiltIns.hs":								 --
-- 			  Implementation of various Aldor Built-In operations			 --
-- 						 L.A.Timochouk,  2000--2002							 --
-- ========================================================================= --
-- RATIONALE:
-- We treat many standard Aldor domains (such as Integer)  as elementary
--		built-in types, as their internal structure is not representable
--		within our selected framework   (e.g., it may  refer to machine-
--		specific underlying types, or require imperative evaluation). We
--		therefore need to maintain a uniform list of supported  built-in
--		operations; as this list will never be completely stable,  it is
--		encapsulated in a separate module.
-- THUS: built-in functions  belong to those (few) pre-defined  imported
--		domains for which we know how to do evaluation.   Other imported
--		domains are viewed as purely abstract types without type defini-
--		tions or data evaluation.

module BuiltIns
	   (
		  dontImport, isBuiltIn, biBinds, biInfo
	   )
where

import DataTypes
import Data.FiniteMap

modName :: String
modName = "BuiltIns"


-- ================= --
--	"dontImport":    --
-- ================= --
--  "Standard" domains and libraries which should NOT be imported explicitly:
--
dontImport :: [QName]
dontImport  = [["Integer"], ["Float"],     ["Boolean"], ["FormattedOutput"],
			   ["AxlLib"],  ["BLInteger"], ["BMFloat"], ["BLErrorPackage"],
			   ["BlLib"] ]

-- =============== --
--  "builtIns_L":  --
-- =============== --
--	The list of supported built-in functions in a "source" form (not optimised
--		for efficient look-up):
--
type BIL_Entry = (String, [AldorType], AldorType,  BI_Body)
--				  OpCode  ArgTypes_if_known  ResType_if_known  Operation_Lambda

builtIns_L:: [BIL_Entry]

builtIns_L = [
	-- ======================== --
	--	"BooleanT" Operations:  --
	-- ======================== --
	--	Equality:
	--
	("=",  [BooleanT,		 BooleanT], 	   BooleanT,
		 \ [BooleanVal a,    BooleanVal b] ->  BooleanVal (a == b)),

	("~=", [BooleanT,        BooleanT],        BooleanT,
	     \ [BooleanVal a,    BooleanVal b] ->  BooleanVal (a /= b)),

	--	Logical functions:
	--	"And" and "Or" are expressions as they require special evaluation
	--		rules; only "Not" is a built-in function:
	--
	("not",[BooleanT],	   	  BooleanT,
		 \ [BooleanVal x]  -> BooleanVal (not x)),


	-- ======================== --
	--	"IntegerT" Operations:  --
	-- ======================== --
	---------------------------------
	--	UNARY  Integer operations: --
   	---------------------------------
	("+",  [IntegerT],       IntegerT,
		 \ [IntegerVal  x]-> IntegerVal  x),

	("-",  [IntegerT],  	 IntegerT,
		 \ [IntegerVal  x]-> IntegerVal (-x)),

	("abs",[IntegerT],  	 IntegerT,
		 \ [IntegerVal x] -> IntegerVal (abs x)),

	---------------------------------
	--	BINARY Integer operations: --
	---------------------------------
   	("+",  [IntegerT,   	 IntegerT],   	   IntegerT,
		 \ [IntegerVal  x,   IntegerVal  y] -> IntegerVal (x+y) ),

	("-",  [IntegerT,   	 IntegerT],   	   IntegerT,
		 \ [IntegerVal  x,   IntegerVal  y] -> IntegerVal (x-y) ),

	("*",  [IntegerT,   	 IntegerT],   	   IntegerT,
		 \ [IntegerVal  x,   IntegerVal  y] -> IntegerVal (x*y) ),
	--
	--	"quo", rem": Integer Division with rounding towards 0:
	--
	("quo",[IntegerT,   	 IntegerT],   	   IntegerT,
		 \ [IntegerVal  x,   IntegerVal  y] ->
		   if y /= 0 then IntegerVal (x `quot` y)
					 else error2 "quo" "Division by 0"
	),

	("rem",[IntegerT,   	 IntegerT],        IntegerT,
		 \ [IntegerVal  x,   IntegerVal  y] ->
		   if y /= 0 then IntegerVal (x `rem`  y)
					 else error2 "rem" "Division by 0"
	),
	--
	--	"mod": Integer Division with rounding towards -infinity
	--		(but there is no "div" operation!):
	--
	("mod",[IntegerT,   	 IntegerT],   	   IntegerT,
		 \ [IntegerVal  x,   IntegerVal  y] ->
		   if y /= 0 then IntegerVal (x `mod`  y)
					 else error2 "mod" "Division by 0"
	),
	--
	--	Integer Exponentiation:
	--
	("^",  [IntegerT,   	 IntegerT],   	   IntegerT,
		 \ [IntegerVal  x,   IntegerVal  y] ->

		   case (  abs x, y) of
				(0, n)    -> if  n <= 0
							 then error2 "^\"--Integer" ("0^("++(show n)++")")
							 else IntegerVal (0)

				(1, n)    -> IntegerVal (x ^  n)  -- always exists

				(m, n)    -> if  n <  0
							 then error2 "^\"--Integer"
										 ((show m)++"^("++(show n)++")")
							 else IntegerVal (m ^ n)
	),
	--
	--	Integer Comparisons:
	--
	("=",  [IntegerT, 	   IntegerT],	 	 BooleanT,
		 \ [IntegerVal  x, IntegerVal  y] -> BooleanVal (x == y)),

	("~=", [IntegerT, 	   IntegerT], 		 BooleanT,
		 \ [IntegerVal  x, IntegerVal  y] -> BooleanVal (x /= y)),

	(">=", [IntegerT, 	   IntegerT],   	 BooleanT,
		 \ [IntegerVal  x, IntegerVal  y] -> BooleanVal (x >= y)),

	("<=", [IntegerT, 	   IntegerT],  		 BooleanT,
		 \ [IntegerVal  x, IntegerVal  y] -> BooleanVal (x <= y)),

	("<",  [IntegerT, 	   IntegerT],   	 BooleanT,
		 \ [IntegerVal  x, IntegerVal  y] -> BooleanVal (x <  y)),

	(">",  [IntegerT, 	   IntegerT],   	 BooleanT,
		 \ [IntegerVal  x, IntegerVal  y] -> BooleanVal (x >  y)),

	-- ====================== --
	--	"FloatT" Operations:  --
	-- ====================== --
	----------------------------------
	--	UNARY  Floating Operations: --
	----------------------------------
	("+",  [FloatT],  	   FloatT,
		 \ [FloatVal x]  -> FloatVal   x ),

	("-",  [FloatT],  	   FloatT,
		 \ [FloatVal x]  -> FloatVal (-x)),

	("abs",[FloatT],  	   FloatT,
		 \ [FloatVal x]  -> FloatVal (abs x)),

	----------------------------------
	--  BINARY Floating Operations: --
	----------------------------------
	("+",  [FloatT,   	   FloatT],     	 FloatT,
		 \ [FloatVal  x,   FloatVal  y]   -> FloatVal (x+y)),

	("-",  [FloatT,   	   FloatT],          FloatT,
		 \ [FloatVal  x,   FloatVal  y]   -> FloatVal (x-y)),

	("*",  [FloatT,   	   FloatT],     	 FloatT,
	     \ [FloatVal  x,   FloatVal  y]   -> FloatVal (x*y)),

	("/",  [FloatT,   	   FloatT],     	 FloatT,
	     \ [FloatVal  x,   FloatVal  y]   ->
		   if  y /= 0.0 then  FloatVal (x / y)
						else  error2   "\"/\"--Float" "Division by 0"
	),
	--
	--	Exponentiation with Integer Power:
	--
	("^",  [FloatT,   	   IntegerT],   	 FloatT,
	     \ [FloatVal  x,   IntegerVal  y] ->
		   if  x == 0.0 && y <= 0
				then error2  "\"^\"--Float" ((show x)++"^("++(show y)++")")
				else FloatVal (x ^ y)
	),
	--
	--	Exponentiation with Floating-Point Power:
	--
	("^",  [FloatT,   	   FloatT],     	 FloatT,
		 \ [FloatVal  x,   FloatVal  y]   ->
		   if  x <= 0.0
				then error2  "\"^\"--Float" ((show x)++"^("++(show y)++")")
				else FloatVal (x ** y)
	),
	--
	--  Floating-Point Comparisons:
	--
	("=",  [FloatT,   	   FloatT],     	 BooleanT,
	     \ [FloatVal  x,   FloatVal  y]   -> BooleanVal (x == y)),

	("~=", [FloatT,   	   FloatT],     	 BooleanT,
		 \ [FloatVal  x,   FloatVal  y]   -> BooleanVal (x /= y)),

	(">=", [FloatT,   	   FloatT],     	 BooleanT,
		 \ [FloatVal  x,   FloatVal  y]   -> BooleanVal (x >= y)),

	("<=", [FloatT,   	   FloatT],     	 BooleanT,
	     \ [FloatVal  x,   FloatVal  y]   -> BooleanVal (x <= y)),

	("<",  [FloatT,   	   FloatT],     	 BooleanT,
		 \ [FloatVal  x,   FloatVal  y]   -> BooleanVal (x <  y)),

	(">",  [FloatT,   	   FloatT],     	 BooleanT,
		 \ [FloatVal  x,   FloatVal  y]   -> BooleanVal (x >  y)),


	-- ======================= --
	--	"StringT" Operations:  --
	-- ======================= --
	--	Equality:
	--
	("=",  [StringT,	   StringT], 	   	 BooleanT,
		 \ [StringVal s,   StringVal t]   -> BooleanVal (s == t)),

	("~=", [StringT,	   StringT], 	   	 BooleanT,
		 \ [StringVal s,   StringVal t]   -> BooleanVal (s /= t))
	--
	--	XXX: More to come here: e.g. concatenation

	--
	--	That's all for the moment!..
	--
   ]

-- ============= --
-- "builtIns_O": --
-- "builtIns_N": --
-- ============= --
-- "builtIns_O":   the same as "builtIns_L", but as a map (dictionary)
--		OpCode => [(BINo,   Type,  Body)]     suitable for fast
--		names look-up;
-- "builtIns_N":   like "builtIns_O",   but as a map
--		BINo   =>  (OpCode, Type,  Body)
--		used for faster evaluation and type-checking.
--	NB: the maps will be contructed once, during the first evaluation
--		of their definitions, and re-used in the subsequent calls:
--
type	     BIDO_Entry = (BINo,  AldorType,  BI_Body)
type         BI_DictO   = FiniteMap   String  [BIDO_Entry]
builtIns_O:: BI_DictO

builtIns_O = fst (foldl attachBI (emptyFM, 0) builtIns_L)
	where
	attachBI:: (BI_DictO, BINo) -> BIL_Entry ->(BI_DictO, BINo)

	attachBI   (fm, bino) (op_code, arg_types, retype, op_body)
		--	If this opcode is already in the map, merge the contents:
		|  elemFM   op_code  fm =
		   (addToFM_C appendOverLoaded fm
					op_code    [(bino,  BuiltInT arg_types retype,  op_body)],
		    bino+1)

		--	Otherwise, create a new map entry:
		|  otherwise =
		   (addToFM fm op_code [(bino,  BuiltInT arg_types retype,  op_body)],
			bino+1)

		where
		--	"appendOverLoaded": the old list is always non-empty,   the new
		--		argument is a singleton list:
		appendOverLoaded:: [BIDO_Entry]-> [BIDO_Entry]-> [BIDO_Entry]
		appendOverLoaded   olds@(_:_)     [new]       =  new : olds
		appendOverLoaded   _ _ = error2 "attachBI" "Invalid internal call"


type		 BI_DictN  = FiniteMap  BINo  (String, AldorType, BI_Body)
builtIns_N:: BI_DictN

builtIns_N = foldFM attachBIs  emptyFM  builtIns_O
	where
	attachBIs:: String -> [BIDO_Entry] -> BI_DictN -> BI_DictN
	attachBIs   opcode    entries         curr      =
				foldl     attachBI  curr  entries
		where
		attachBI:: BI_DictN -> BIDO_Entry -> BI_DictN
		attachBI   curr'  (bi_no, bi_type, bi_body)
			| elemFM bi_no curr' =
			  error2 "builtIns_N"   ("Duplicate BINo."  ++ (show bi_no) )
			| otherwise	=  addToFM curr' bi_no (opcode, bi_type, bi_body)

--------------------
--	"isBuiltIn":  --
--	"biBinds"  :  --
--------------------
--	"isBuiltIn" checks whether a given name (with no type information!) is the
--		name of some built-in function:
--
isBuiltIn :: String -> Bool
isBuiltIn    name   =  elemFM  name  builtIns_O

--
--	"biBinds" returns the list of all BINo's for a given name (if it is indeed
--		a built-in function name; otherwise it returns the empty list):
--
biBinds   :: String -> [BINo]
biBinds      name   =
	case (lookupFM  builtIns_O name) of
		Just   bis  -> map  (\ (bino,_,_) -> bino)  bis
		_		    -> []

-----------------
--	"biInfo":  --
-----------------
--	Returns the opcode, proto and body for a given BINo:
--
biInfo :: BINo   -> (String, AldorType, BI_Body)
biInfo    bino   =
	case (lookupFM  builtIns_N bino) of
		Just res -> res
		_		 -> error2 "biInfo" ("BINo."++(show bino)++" not found")

-- =========== --
--  "error2":  --
-- =========== --
error2:: String -> String -> a
error2 = errorM modName

