-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--							   "DataTypes.hs":								 --
--    Definitions of principal data structures derived from the Aldor AST:   --
--						  L.A.Timochouk, 2000--2002						     --
-- ========================================================================= --
module DataTypes
	   (
			AldorType    (..),
			AldorExpr    (..),
			AldorValue   (..),
			AldorFBlock  (..),

			DefTableEntry(..),
			FlatDTE      (..),

			DefTable,
				--  "DefTable" is an abstract type (unlike "AldorFBlock"), so
				--		we define export all operations on "DefTable"s:
				emptyDT,    isEmptyDT,    foldDT,      attachFDTE,
				lookupDT,   getFParm,     singltnDT,   attachImport,
				getImports, replaceDTEs,  removeDTEs,  dt2list,
				dte2fdte,

			BINo,  DefNo,   UseNo, FBNo,  noNUMBER,    BI_Body,
			QName, getFBNo_U,

			Consolidatable  (consolidate), 	   consolidateList, 
			WithAttr		(..),
			ConsoList,      getCList,		   mkCList,
			crossVectors,   crossWithAttrs,

			ANDable ((-&&-)),      (=&&=),
			ORable  ((-||-)),	   (=||=),

			undef,  errorM
	   )
where

import Maybe
import Data.FiniteMap

modName :: String
modName = "DataTypes"

-----------------------------------------------
--	Enumeration types used in the following: --
-----------------------------------------------
type DefNo = Int
type UseNo = Int
type FBNo  = Int
type BINo  = Int

noNUMBER  :: Int
noNUMBER   =  -1


-------------------------------------------------------------------------------
--	Aldor Types (those we are interested in) are represented in the following -
--		way. NB: This is only a data type for Types themselves, not for their -
--		values!																  -
-------------------------------------------------------------------------------
data   AldorType =
	--	"Elementary" types:
	--
	TrivT     	-- A 1-element type (its only value is "TrivVal")
  |	ExitT	  	-- A 0-element type
  |	TypeT	  	-- The type of all types
  |	BooleanT  
  |	IntegerT  
  |	FloatT    
  |	StringT   

	--	Function ProtoType:  Parms and ReturnType.  Parms are "FParmFB"s (see
	--		below) whose DefTable contains a single entry with "VoidFB"  body
	--		it would be non-Void if default args were provided, but  they are
	--		NOT, as that would complicate type-checking very much). Parms are
	--		not just DefTables as "uplinks" are required in them: each points
	--		to the previous Parm,  so they form a hierarchical context).  The
	--		Return Type did not really need to be an FBlock (the Aldor syntax
	--		does not allow any nested contexts in it),  but it is made into a
	--		"SimpleFB"   containing the return type expression  to provide an
	--		"uplink" from it as well:
	--
  |	FunctionT [AldorFBlock]	AldorFBlock

	--	BuiltIn Proto: A simplified form of "FunctionT":
	--		The argument and the result types can only be ELEMENTARY AldorType
	--		ctors (those without any arguments):
	--
  |	BuiltInT  [AldorType]   AldorType

	--	Record Type:
	--	Record fields are organised similarly to function Parms  (as a seq of
	--		"FParmFB"s, see above):
	--
  |	RecordT   [AldorFBlock]

	--	Union  Type:
	--	Structure similar to that of "RecordT", just different semantics:
	--
  |	UnionT    [AldorFBlock]

	--	Abstract Data Type:
	--	Contains an FBlock yielding the internal representation type for this
	--		ADT and a DefTable specifying its namespace:
	--	To provide a comparable self-identification of an ADT, it has a built-
	--		in DefNo:
	--
  | ADT	DefNo AldorFBlock DefTable

	--	Category: an ADT signature:
	--	Implemented as a DefTable with entries only containing type exprs, and
	--		void bodies. As for ADT, it has a DefNo for self-identification:
	--
  | CategoryT DefNo DefTable


-------------------------------------------------------
--	Aldor Expressions: A limited functional sub-set: --
-------------------------------------------------------
--	Note that most sub-expressions of AldorExpr are not just AldorExprs  but
--		FBlocks (see below), so they may contain local DefTables:
--
type   QName      =  [String]   -- Fully-qualified name, outer-most first

data   AldorExpr  =
	VoidExpr				 	-- Nothing, evaluates to TrivVal (see below)

	--	Literal is an AldorValue (defined below):
	--
  |	Literal  	AldorValue

	--	Identifier is a symbolic name defined in the current evaluation context.
	--		The name can be fully-qualified (if found inside a domain which may
	--		be enclosed in another domain, ...), hense "QName".  The  UseNo  is
	--		used for resolving of (possibly overloaded) names:
	--
  |	Identifier	UseNo  QName

	--	"And" and "Or" are not just built-in functions because their arguments
	--		do not always need to be completely evaluated: evaluation stops as
	--		soon as "False" and "True", resp, are encountered. But: "Not" is a
	--		built-in function, not an expr, as it always takes 1 argument:
	--
  |	AndList		[AldorFBlock]
  |	OrList		[AldorFBlock]

	--	If-Then-Else construct: Condition, Then-Branch, Else-Branch,  where
	--		the Condition must evaluate to "Bool", and the Branches must be
	--		of the same type:
	--
  |	IfThenElse	 AldorFBlock  AldorFBlock  AldorFBlock

	--	Function Application, for all functions    (built-in, atomic domain
	--		exports, user-defined, ...):
	--	The 1st arg is an FBlock (in general) whose expression evaluates to
	--		an ident naming a function or to a LambdaVal.  The 2nd arg is a
	--		list of Actual Parms:
	--
  |	FunctionCall AldorFBlock [AldorFBlock]

	--	Record Ctor:
	--	Not a built-in function as it is inherently polymorphic.   The args 
	--		are values of the record fields. Returns a "RecordVal":
	--
  | RecordCtor  [AldorFBlock]

	--	Union Ctor:
	--	The 1st arg is the Identifier of the field to be written,   the 2nd
	--		one is the value of that field. Returns a "UnionVal":
	--
  | UnionCtor    AldorExpr    AldorFBlock

	--	UnionCase: the 1st arg must be a Union, the 2nd is a Selector Ident,
	--		the result is Boolean:
	--
  | UnionCase	 AldorFBlock  AldorExpr

	--	Restriction of the expr result (given by the 1st arg)   to the type
	--		given by the 2nd arg:
	--
  |	RestrictType AldorFBlock  AldorFBlock

	--	Type conversion expr (a controlled form of "pretend"): the result of
	--		the 1st arg is converted to the type given by the 2nd arg:
	--
  | ConvertTo    AldorFBlock  AldorFBlock

	--	ADT construction from a DT (from a NestedFB with Void Inner FB),  or
	--	ADT extension. DefNos are installed for self-identification:
	--
  | Add1  DefNo  AldorFBlock
  | Add2  DefNo  AldorFBlock  AldorFBlock

	--	Category construction or extension (similar to "Add*" above):
	--
  | With1 DefNo  AldorFBlock
  | With2 DefNo  AldorFBlock  AldorFBlock


---------------------------------------------------------------
--	Aldor Values (to which Expressions are to be evaluated): --
---------------------------------------------------------------
--	NB: "AldorValue"s can be embedded into "AldorExpr"s via the "Literal" ctor:

type   BI_Body    =  [AldorValue] -> AldorValue

data   AldorValue =

	--	"Elementary" values:
	--
	TrivVal			 	  	-- The only value of "TrivT"
  |	TypeVal    AldorType  
  |	BooleanVal Bool		  
  |	IntegerVal Integer    	-- Infinite-precision!
  |	FloatVal   Double     	-- IEEE-754, 53-bit mantissa
  |	StringVal  String	  

	--	"Lambda" is actually a Value, not an Expr:  as itself,  it  cannot be
	--		evaluated further. It consists of a FunctionT proto and  a  body:
	--
  |	LambdaVal  AldorType AldorFBlock

	--	Built-in operations are treated similarly, but the proto and the body
	--		are of simplified form (BuiltInT AldorType ctor and BI_Body resp)
	--		and the OpName is included:
	--
  |	BuiltInOp  String    AldorType   BI_Body

	--	"RecordVal" is a list of Field Values,  in the same order as declared
	--		in the corresp "RecordT":
	--
  |	RecordVal  [AldorValue]

	--	"UnionVal": only one value field is allowed at a time;  once created,
	--		it becomes immutable (in our functional model, that is).    Also
	--		specified is the Identifier (tag) from the corresp UnionCtor:
	--
  |	UnionVal   AldorExpr AldorValue

	--	"ADT_Val": The value of an Abstract Data Type is that of  its internal
	--		representation, i.e. any AldorValue. It also includes  a reference
	--		to the corresp ADT definition:
	--
  | ADT_Val    AldorType AldorValue  -- ADT ref, Actual Value


-- ================================================================= --
--	"DefTable" and "AldorFBlock": The major "high-level" structures: --
-- ================================================================= --
--	The DefTable is a dictionary whose keys are names being defined and values
--		are LISTS of entries    (not single entries -- to allow for overloaded
--		names in a DefTable).
--	The generic entries ("GenDTE"s) are triples (DefNo, Type, Body) where Type
--		can be any AldorExpr (but not an FBlock -- Aldor syntax does not allow
--		nested scopes here)  evaluating to AldorType, and the Body which is an
--		FBlock.
--	Special case of entries ("RenDTE"):  no Type, the RHS is an Identifier re-
--		naming the name in the LHS of the corresp definition.
--	The DefTable also contains a list of domains "opened" in the current scope
--		("imported" in the Aldor terminology).   These domains are (currently) 
--		specified by just their names (possibly fully-qualified);  these names
--		are currently not Idents as they are not to be resolved as themselves:
--
data    DefTableEntry =
	  	GenDTE DefNo AldorExpr AldorFBlock  -- General : DefNo, TypeExpr, Body
	  |	RenDTE DefNo AldorExpr  			-- Renaming: DefNo, Renaming_Ident

data    DefTable      =
		DefTable (FiniteMap String [DefTableEntry]) [AldorExpr]
		-- 		  DefTable itself				 	Names of imported domains

--
--	A functional Aldor program is represented as an AldorFBlock instance:
--
data    AldorFBlock   =
		--	"VoidFB" has no UpLink, but it still needs an FBNo for unique
		--		positional identification:
		VoidFB			FBNo

		--	"NestedFB": FBNo, DefTable, the inner FBlock and the UpLink:
		--
	  |	NestedFB   		FBNo  DefTable  AldorFBlock AldorFBlock

		--	"SimpleFB": FBNo, Expression and the UpLink:
		--
	  |	SimpleFB   		FBNo  AldorExpr AldorFBlock

		--	"FParmFB" : For function parms in FunctionT and other purposes:
		--	 FBNo, singltn DefTable and the UpLink:
		--  Although this is a particular case of "NestedFB", we cannot use
		--		the latter instead, as we sometimes need to tell whether  a
		--		given FB is a function parm:
		--
	  |	FParmFB			FBNo  DefTable  AldorFBlock
		

-- ==================================== --
--	Elementary operations on DefTables: --
-- ==================================== --
--	The DefTable ctor itself is not exported, so the type implementation is
--		hidden from the external view.      Instead we export the following
--		operations on DefTables:

emptyDT  :: DefTable
emptyDT  =  DefTable emptyFM []


isEmptyDT:: DefTable -> Bool
isEmptyDT  (DefTable fm imports) = (isEmptyFM fm) && (length imports == 0)


---------------------
--	"attachFDTE":  --
---------------------
--	Attaches a new Definition (given as a FlatDTE) to the  given DefTable. Here
--		"DefNo" is typically a "through-number" of the definition    within the
--		out-most FBlock ("noNUMBER" if not known at the time of "append"ing):
--
attachFDTE:: DefTable  -> FlatDTE -> DefTable

attachFDTE  (DefTable fm  imports)  fdte
	--
	--	If this "name" key already exists in the DefTable, we append one more
	--		overloaded definition for the "name" to the existing list and re-
	--		turn the modified DefTable:
	| elemFM  name   fm = DefTable
						  (addToFM_C appendOverLoaded fm name entry_lst) imports
	--
	--	Otherwise, create a new entry:
	| otherwise		    = DefTable  (addToFM		  fm name entry_lst) imports

	where
	--	"name" exists for both GenFDTE and RenFDTE:
	--
	name   = case fdte of
		GenFDTE _ name' _ _ -> name'
		RenFDTE _ name' _   -> name'

	--	"entry_lst" is the entry to be inserted as a singleton list:
	--
	entry_lst::  [DefTableEntry]
	entry_lst =  case fdte of
		GenFDTE def_no _ type_expr body_fb -> [GenDTE def_no type_expr body_fb]
		RenFDTE def_no _ ren_expr  		   -> [RenDTE def_no ren_expr]

	--	The following combining function is ALWAYS called with the 1st arg being
	--		a non-empty list, and the 2nd one being a singleton list:
	--
	appendOverLoaded:: [DefTableEntry]   -> [DefTableEntry] -> [DefTableEntry]
	appendOverLoaded   olds@(_:_)           [new_se]        =  new_se  : olds
	appendOverLoaded   _ _  =   error2 "attachFDTE" "Invalid internal call"


----------------------
--	"attachImport": --
----------------------
attachImport:: DefTable -> AldorExpr -> DefTable

attachImport  (DefTable fm imports) dom@(Identifier _ dom_name) =
	DefTable fm imports'
	where
	imports' = if dom_name `elem` old_names then imports else dom : imports
	old_names=   [old_name | Identifier _ old_name <- imports]
	--
	--	Here "elem" performs comparison of "QName"s

attachImport  _ _ = error2 "attachImport" "Not an Identifier in the RHS"


------------------
--	"lookupDT": --
------------------
--	Returns a list of over-loaded definitions for a given Name within the given
--		DefTable. An empty list is returned (no error!) if the look-up fails:
--
lookupDT :: DefTable  -> String -> [DefTableEntry]

lookupDT   (DefTable fm _) name  =  getDTEs (lookupFM fm name)
	where
	getDTEs :: (Maybe [DefTableEntry]) -> [DefTableEntry]
	getDTEs (Just dte_list)  =
		if  length dte_list /= 0 then  dte_list
								 else  error2 "lookupDT" "Empty DTE list"
	getDTEs _				 = []


-----------------
--	"dt2list": --
-----------------
--	Converts the DefTable contents into [(String, [DefTableEntry])].   Does NOT
--		return the Imports:
--
dt2list :: DefTable		  -> [(String, [DefTableEntry])]
dt2list   (DefTable fm _) =  fmToList  fm


-------------------
--	"dte2fdte":  --
-------------------
dte2fdte :: String -> DefTableEntry -> FlatDTE

dte2fdte    name      (GenDTE def_no type_expr body) =
	GenFDTE def_no name type_expr body

dte2fdte    name      (RenDTE def_no id_expr)        =
	RenFDTE def_no name id_expr


---------------------
--	"removeDTEs":  --
---------------------
--	Remove the list of entries corresp to a given name from the DefTable (or do
--		nothing if this name is not defined):
--
removeDTEs  :: DefTable -> String -> DefTable

removeDTEs    (DefTable fm imports)  name =
			   DefTable   (filterFM (inclBut name) fm)  imports
	where
	inclBut :: String   -> String -> [DefTableEntry] -> Bool
	inclBut    to_excl     curr_name  _  =  to_excl /=  curr_name


---------------------
--	"replaceDTEs": --
---------------------
--	Remove the original entry list for a given name (if present at all in the
--		DefTable) and replace it with a new list  (or just install the latter
--		if there was nothing to remove):
--
replaceDTEs  :: DefTable -> String -> [DefTableEntry] -> DefTable

replaceDTEs    dt  name  new_dtes =
	foldl   attachDTE   (removeDTEs dt name) new_dtes
	where
	attachDTE:: DefTable -> DefTableEntry -> DefTable

	attachDTE   curr       (GenDTE  def_no type_expr body_fb) =
		attachFDTE curr    (GenFDTE def_no name type_expr body_fb)

	attachDTE   curr       (RenDTE  def_no id_expr)			  =
		attachFDTE curr	   (RenFDTE def_no name id_expr)


--------------------
--	"getImports": --
--------------------
--	Returns the list of domains imported into the context of this DefTable:
--
getImports :: DefTable -> [AldorExpr]
getImports   (DefTable _  imports) = imports


-----------------
--	"foldDT":  --
-----------------
--	NB: This function performs actions on DTEs only, not Imports!
--
foldDT:: (FlatDTE-> a -> a) -> a -> DefTable  -> a

foldDT   (action :: FlatDTE -> a -> a)
		 init_a    (DefTable fm _)  =  foldFM action' init_a  fm
	where
	action' :: String -> [DefTableEntry] -> a -> a

	action' name  ovlst init'  =  foldl  actionX init' ovlst
		where
		actionX ::  a ->  DefTableEntry  -> a

		actionX initX ( GenDTE def_no      type_expr body_fb) =
			   action (GenFDTE def_no name type_expr body_fb) initX

		actionX initX ( RenDTE def_no 	   ren_expr) =
			   action (RenFDTE def_no name ren_expr)  		  initX

-------------------
--	"getFBNo_U": --
-------------------
--	Returns FBlock's FBNo without any integrity checks. I.e., it's OK if the
--		value to be returned is "noNUMBER":
--
getFBNo_U   ::AldorFBlock -> FBNo
getFBNo_U     (VoidFB   fb_no)        = fb_no
getFBNo_U     (SimpleFB fb_no _ _  )  = fb_no
getFBNo_U     (NestedFB fb_no _ _ _)  = fb_no
getFBNo_U     (FParmFB  fb_no _ _  )  = fb_no


-- ===================== --
--	Operations on DTEs:  --
-- ===================== --
--	"FlatDTE" is introduced just for ease of use in many situations.
--	It combines the Name and the contents of DefTableEntry into one
--		object:
--
data FlatDTE = 
		GenFDTE DefNo String AldorExpr AldorFBlock   -- DefNo, Name, Type, Body
	  | RenFDTE DefNo String AldorExpr  		     -- DefNo, Name, RenamingId

--
--	"getFParm":
--	Specific case: extracts a (single) Function Parm from its enclosing FBlock
--		(within a FunctionT) as a FlatDTE (more precisely, GenFDTE):
--
getFParm :: AldorFBlock -> FlatDTE

getFParm   (FParmFB _ (DefTable fm []) _) =
			GenFDTE def_no name type_expr void_body
	where
	[(name, [GenDTE def_no type_expr void_body])]   = fmToList fm
	--
	--	A program error will occur at run time if (fmToList fm) is not a 1-elmnt
	--		list, or there are more than 1 DTE in it, or it is not a GenDTE

getFParm  _ = error2 "getFParm" "Not a (valid) \"FParmFB\""

--
--	"singltnDT":
--	(a kind of) counter-part to "getFParm", useful in constructing "FunctionT"s:
--
singltnDT :: FlatDTE -> DefTable
singltnDT =  attachFDTE  emptyDT


-- ======================= --
--	Data with Attributes:  --
-- ======================= --
--	Consider a data type "whole" which contains some attribute "attr"; below we
--		define a class ("WithAttr") of such types.
------------------------
--	"Consolidatable": --
------------------------
--	First of all, to be really useful, the "attr" type must be "Consolidatable":
--		given two attr values, we can produce their "common" value (or fail):
--
class Consolidatable attr   where
	consolidate ::   attr-> attr  -> Maybe attr

--
--	We can then consolidate lists of such "attrs" as well:
--
consolidateList ::		 Consolidatable  attr =>  [attr] -> Maybe attr

consolidateList []	   = error2 "consolidateList" "Empty list"

consolidateList (a:as) = consolidateList_R as a
	where
	--	ConsolidateList_R :: tail-recursive, of type
	--	[attr] -> attr -> Maybe attr
	--
	consolidateList_R []     curr = Just curr -- Input finished, got the result!

	consolidateList_R (x:xs) curr =
		case consolidate   x curr of
			Just curr' ->  consolidateList_R xs curr'
			Nothing	   ->  Nothing			  -- Fail immediately!

------------------
--	"WithAttr": --
------------------
--	We can now define the "WithAttr" class itself.   It provides "getAttr" and
--		"setAttr" operations:
--
class Consolidatable attr  =>  WithAttr attr whole   where
	getAttr      ::  whole ->  attr
	setAttr      ::  attr  ->  whole -> whole

-------------------
--	"ConsoList": --
-------------------
--	A list of "WithAttr" data characterised by the same "attr" values. Formally
--		speaking, this type is unconstrained   (ie it is NOT required here that
--		"attr" and "whole" belong to some "WithAttr" instance),  but it will be
--		ABSTRACT; its instances can only be created via the "mkCList"  function
--		which IS constrained by "WithAttr" membership.
--	NB: (1) the common "attr" is stored explicitly;
--		(2) the "attr" type is NOT required to be an "Eq" member:
--
data     ConsoList   attr whole =   ConsoList attr [whole]

--
--	For suitable "attr" and "whole", we can make "ConsoList whole" a member of
--		the "WithAttr" class, ie define "getAttr" and "setAttr" on such "Conso-
--		List"s:
--
instance WithAttr attr whole	  => WithAttr attr  (ConsoList  attr whole)
	where
	getAttr (ConsoList attr _   ) =  attr
	setAttr attr (ConsoList _ ws) =  ConsoList attr (map (setAttr attr) ws)

--
--	Accessing the data stored in a "ConsoList":
--
getCList  :: ConsoList attr whole -> [whole]
getCList    (ConsoList _    ws  ) =  ws

--
--	Finally, we provide for making of "ConsoLists"  from ordinary lists of
--		"WithAttr" (not arbitrary!) data.  NB: don't confuse this with the
--		"consolidateList" function.     The return type is a "Maybe" since
--		attrs consolidation may fail. This is a CONTROLLED way of creating
--		"ConsoList" instances.
--	NB: This function is used in "CrossWithAttrs":
--
mkCList:: (WithAttr attr whole) => [whole] -> Maybe (ConsoList attr whole)

mkCList  ws  =
	case ma of
		Just ca  -> Just   (ConsoList ca (map (setAttr ca) ws))
					-- Set the common attr to all list elements...
		Nothing  -> Nothing
	where
	ma  = consolidateList  (map  getAttr  ws)   -- ma:: Maybe attr


------------------------
--	"crossVectors":   --
------------------------
--  A polymorphic function which constructs the cross-product of vectors from
--      a list [vec1, ..., vecN] where each vector "vecI" (a vector of "vers-
--		ions") is of type [a].   The result is a set  (actually again a list)
--		[x1, ..., xM],   where each "xJ" is of the form [a1, ..., aN],   "aI"
--		belonging to "vecI":
--
crossVectors :: [[a]] -> [[a]]
crossVectors    []    =  []         -- Or should it be an error?
crossVectors    (as : ass) =
	case ass  of
		--  When we have just one vector (e.g. vec1):  return it as a list of
		--      singltn lists corresponding to the arg versions:
		--
		[]  -> [[a] | a <- as]

		-- The General Case:
		_   -> [l:r | l <- as, r <- crossVectors ass]


------------------------
--	"crossWithAttrs": --
------------------------
--	Crosses over (using "crossVectors") several vectors of "WithAttr" elements.
--		Within each  version constructed,   the componets are consolidated, or
--		failed that, the version in question is dropped.  The result is a list
--		of "ConsoList"s:
--
crossWithAttrs ::  WithAttr attr whole =>  [[whole]] -> [ConsoList attr whole]
--
--  Arg: Inner dim:  "versions"
--  Res: Outer dim: X-versions retained

crossWithAttrs wss = catMaybes (map mkCList (crossVectors wss))


-- ===================== --
--  The "ANDable" Class: --
-- ===================== --
--	Similar to "Consolidatable":
--		(-&&-) <~=~> consolidate
--		(=&&=) <~=~> consolidateList
--	The difference is that (-&&-) cannot fail: its result is not a Maybe type:
--
infixr 3 -&&-  		 -- Similar to the (&&) operator from the standard Prelude

class ANDable a where
    (-&&-) :: a -> a -> a

--
--  Lists of "ANDable"s can be folded by the "=&&=" operator:
--
(=&&=) :: ANDable a => [a] -> a

(=&&=) []		= error "(=&&=): Empty list"
(=&&=) (x : xs) = foldl  (-&&-)  x  xs


-- ====================== --
--	The "ORable" Class:   --
-- ====================== --
--	Similar to "ANDable":
--
infixr 2 -||-

class ORable  a where
	(-||-) :: a -> a -> a

--
--  Lists of "ORable"s can be folded by the "=||=" operator:
--
(=||=) ::  ORable a => [a] -> a

(=||=) []		= error "(=||=): Empty list"
(=||=) (x : xs) = foldl  (-||-)  x  xs


-- ================ --
--	Hash functions: --
-- ================ --
--	By far, the simplest way to provide hash functions is to calculate MD5
--		checksums on "Show" instances, and convert the results into "Int"s:
--
-- hashInt ::  Show a => a -> Int


-- ========== --
--	"undef":  --
-- ========== --
--	"General-purpose" undefined value:
--
undef :: ()
undef =  undef


-- =========== --
--	"errorM":  --
--  "error2":  --
-- =========== --
errorM:: String -> String -> String -> a
errorM   thisMod   funcName  msg =
	error ('"':thisMod++"."++funcName++"\": "++msg)

error2:: String -> String -> a
error2 = errorM    modName

