-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--								  "EvalX.hs":								 --
-- 					   Evaluation of Aldor eXpressions						 --
--					      L.A.Timochouk, 2000--2002							 --
-- ========================================================================= --
module EvalX
	   (
		  evalExpr, evalFB, evalProto, evalDT, evalFBList, evalTypeExpr
	   )
where

import DataTypes
import BuiltIns
import ParseAST
import PrPrint
import FButils
import NameBinds
import AlphaDelta
import EvalLib

import IOExts
import Maybe

modName :: String
modName = "EvalX"


-- ============= --
--	"evalExpr":  --
-- ============= --
--  Evaluation engine with eval-time type checking.
--	NB:
--	(1) Evaluation is STRICT (function args are always evaluated  just before
--		the function is called),  except for components of AND and OR  exprs.
--		Untaken branches of conditionals are never evaluated;
--	(2) Evaluation can modify the ETC_Env (e.g. if evaluation fails with some
--		BindVec, the corresp bindings are removed);
--	(3) The result of an evaluation  is a list of EvalRes;  each list element
--		corresponds to a particular set of bindings under which that evaluat-
--		ion result was obtained:
--
evalExpr :: ETC_Env -> AldorExpr -> [EvalRes]

-----------------------
--	"FunctionCall":  --
-----------------------
--	(and also Record or Union field selection):
--
evalExpr ee (FunctionCall callable args)

	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise = 
		case eres of
			[] -> trace ("\nFAILED FUNCTION APPLICATION: "++
						(show (FunctionCall callable args))++"\n") []
			_  -> rres ++ eres
	where
	--	Decrement the Remaining Depth once this "evalExpr" commences:
	--
	ee0 :: ETC_Env
	ee0 =  decr_rem_depth ee

	--	Evaluate the "callable" expr first. Valid results include BuiltInOp,
	--		LambdaVal, and also evaluated and unevaluated Record objects:
	--
	ecals    :: [EvalRes]
	ecals    =  evalFB ee0 callable
--	ecals    =  trace ("\n+++ECALLABLE: "++(show ecals')++"\n") ecals'

	--	Process Record or Union field selectors first, if "args" is a singltn
	--		list (it MAY then be a Field Ident).    We need to remember which
	--		"ecals" elements were actually treated as Record objects -- they
	--		will be excluded from being interpreted as  Lambdas or Built-Ins
	--		below, otherwise multiple evaluation results would be possible for
	--		the same NameBinds -- an integrity violation:
	--	In "run_sels" below,  the first element of each pair  is the corresp
	--		callable's position in "ecals" which we need to memoise:
	--
	all_inds :: [Int]
	all_inds =  [0..(length ecals)-1]

	run_sels :: [(Int, [EvalRes])]

	run_sels =  case args of
		--	Single "Identifier" arg: MAY be a field selector:
		--
		[SimpleFB _  sel@(Identifier _ _) _] ->
			[(i, selectRUnField sel (ecals !!! i)) | i <- all_inds]

		--	Any other "args": certainly not a Record field selector:
		_ ->[]

	--	"rres" is the first part of the over-all result, built from "run_sels":
	--
	rres:: [EvalRes]
	rres = concat (map snd run_sels)

	--	Now: FUNCTIONS (Lambdas and Built-Ins):
	--	Only those "ecals" elements will be considered as callables here, for
	--		which there was no results in "run_sels" (i.e. either no such "i"
	--		at all in "run_sels", or [] in the snd pair component):
	--
	run_inds :: [Int]
	run_inds =  [i | (i,res) <- run_sels, length res > 0]

	fcals    :: [EvalRes]
	fcals    =  [ecals !!! i | i <- all_inds, i `notElem` run_inds]

	--	Evaluate the arguments:
	--	As each evaluation can produce a list of results, they are evaluated
	--		"independently from each other" in the "ee0" environment   (even
	--		though there may  be type dependencies between the corresponding
	--		formal parms -- they don't affect the evaluation of actual args):
	--
	eargs:: [[EvalRes]]				-- Outer dim: args;  inner dim: versions
	eargs=  map (evalFB ee0) args

	--	Cross the callables ("fcals" only!) and the arguments,   producing a
	--		global list of versions with unified NameBinds:
	--
	xvers:: [EvalRes_CL]			-- Outer dim: vers; inner dim: func:args
	xvers=  crossWithAttrs (fcals : eargs)


	--	For each callable object and args in "cvers", perform the call.  The
	--		env for the call is built from "ee0"  and the NameBinds for that
	--		version.  The results of all calls are then merged into a single
	--		list which is returned:
	--
	eres :: [EvalRes]
	eres =  concat  [fCall   (setBinds  (getAttr  ver)  ee0)
							 (head      (getCList ver))
							 (tail      (getCList ver))    |   ver <- xvers]

	------------------------
	--	"selectRUnField": --
	------------------------
	selectRUnField ::  AldorExpr -> EvalRes  ->  [EvalRes]

	selectRUnField     sel@(Identifier use_no _)  run =
		case run of
			--	(0)
			--	"Bottom"s are ignored here since they are indistinguishable
			--		from Lambda Bottoms (so they are treated as the latter):
			--
			Bottom_ER _ -> []

			--	(1)
			--	"run" is fully evaluated to a RecordVal:
			--  	just select the corresp field(s);  there can be multiple
			--		results here since field names can be overloaded (within
			--		different Record types, or even in the same one). We use
			--		type-checking (safely, -- without further evaluation) to
			--		filter out Record types to which the "run" cannot belong:
			--
			Full_ER	_ rval@(RecordVal fields) ->
				let
					--	Type-check "rval" against all "run_targs" found:
					--
					itcrs:: [([TC_Res], Int)]
					itcrs = [(tcExpr_T  (restrBinds ee1 targ)
										rect (Literal rval),
							 index) |
							(targ, rect, index) <- run_targs]

					--	Merge the type-checking results. Each successful result
					--		means that the "callable" "fb" is indeed  of a type
					--		to which the "sel" is bound,   so extraction of the
					--		corresp field is valid:
					--
					oks  :: [(TC_Res,   Int)]
					oks   = merge_itcrs itcrs
				in
					--	Get the corresp fields with new (type-checked) binds
					--		("Bottom_TCR"s are also accepted here  -- should
					--		it be so XXX?):
					--
					[getFieldVal (getAttr tcr) fields index |
								 (tcr,  index) <- oks]

			--	(2)
			--	"run" is fully evaluated to a "UnionVal". Unlike in "RecordVal",
			--		here we have an explicit Selector UseNo, so we return "val"
			--		if the selector is AD-equivalent to the "sel":
			--
			Full_ER _ (UnionVal usel val) ->
				case adeq (etc2ade ee1) usel sel of
					Yes_ADER binds' ->  [Full_ER   binds' val]
					Bottom_ADER     ->  [Bottom_ER binds1]
					No_ADER			->  []
	
			--	(3)
			--	"run" is fully-evaluated but not to a "{Record,Union}Val" --
			--		nothing:
			Full_ER _ _ -> []

			--	(4)
			--	"run" is partially evaluated to a "RecordCtor":   so select the
			--		corresp (partially- or fully-evaluated) field(s):
			--
			Partial_ER _ fb@(SimpleFB _ (RecordCtor  fields ) _) ->
				let
					--	Type-check "fb" against all "run_targs" found; "ee1"
					--		requires further sspecialisation for each target:
					--
					itcrs:: [([TC_Res], Int)]
					itcrs = [(tcFB_T  (restrBinds ee1 targ) rect fb, index)  |
							 (targ, rect, index) <- run_targs]

					--	Merge the type-checking results. Each successful result
					--		means that the "callable" "fb" is indeed  of a type
					--		to which the "sel" is bound,   so extraction of the
					--		corresp field is valid:
					oks  :: [(TC_Res,   Int)]
					oks   = merge_itcrs itcrs
				in
					--	Get the corresp fields, further evaluating them.  Note
					--		that they are already a result of (partial) evalu-
					--		ation, and now the binds could only shrink,  so we
					--		will not get any new results here,   but will drop
					--		those which are inconsistent with new binds:
					--
					concat [evalFieldFB (getAttr tcr) fields index |
										(tcr,  index) <- oks]

			--	(5)
			--	"run" is partially evaluated to a "UnionCtor".  Its 1st arg is
			--		a field selector, the 2nd is a (partially-evaluated) field
			--		value, so we return the 2nd arg if the 1st one is AD-equiv
			--		to the "sel":
			--
			Partial_ER _ (SimpleFB _ (UnionCtor utag pval) _) ->
				case adeq (etc2ade ee1) utag sel of
					Yes_ADER binds' ->  [Partial_ER binds' pval]
					Bottom_ADER     ->  [Bottom_ER  binds1]
					No_ADER			->  []
	
			--	(6)
			--	"run" is partially-evaluated but NOT to a "{Record,Union}Ctor"
			--		(e.g. it remains an Identifier which may still be of Record
			--		or Union type). In this case, we will not be able to return
			--		the selected field, UNLESS the callable is "RecordT" type;
			--		so return [], and this case will then be normalised as the
			--		general function application.  If, however, the "callable"
			--		can be proved to be of "RecordT" type,    and the "sel" is
			--		bound to one of the fields of that Record, return the "sel";
			--		this is to enable type-checking of partially-evaluated de-
			--		pendent records:
			--
			Partial_ER _ hmm ->
				let
					--	Type-check "hmm" against all the "run_targs" which are
					--		"RecordT"s:
					--
					itcrs:: [([TC_Res], Int)]
					itcrs = [(tcFB_T (restrBinds ee1 targ) rect hmm, index)  |
							 (targ, rect@(RecordT _), index) <- run_targs]

					--	Merge the type-checking results. Each successful result
					--		means that the "callable" is indeed of that Record
					--		type:
					oks  :: [(TC_Res,   Int)]
					oks   = merge_itcrs itcrs
				in
					--	Return the Selector itself, with restricted binds:
					--	XXX: UNDECOR:
					--
--					[Partial_ER (getAttr ok) (simpleFB sel) | (ok,_) <- oks]
					[]

		where
		--	Common to all cases above:
		--
		--	All "RecordT"s / "UnionT"s / "FunctionT"s with field indices to
		--		which the "sel" can be bound:
		--
		binds1   :: NameBinds
		binds1    = getAttr run

		run_targs:: [(BindTarget, AldorType, Int)]
		run_targs = findTypesEnclFPTargs binds1 use_no

		ee1      :: ETC_Env
		ee1		  = setBinds binds1  ee0

		restrBinds :: ETC_Env -> BindTarget -> ETC_Env
		restrBinds    eeX        targX      =
			setBinds (modifyBinds (getBinds eeX) use_no [targX]) eeX

		--	Field fetching functions:
		--
		--	(*) For fully-evaluated records:
		--
		getFieldVal:: NameBinds ->  [AldorValue] -> Int -> EvalRes
		getFieldVal   binds'        fields        index

			| index < 0 || index >= length fields =
				error2 "getFieldVal" ("Invalid index: "++(show index))

			| otherwise           = Full_ER binds' (fields !!! index)

		--
		--	(*) For partially-evaluated records:
		--
		evalFieldFB:: NameBinds ->  [AldorFBlock]-> Int -> [EvalRes]
		evalFieldFB   binds'        fields        index

			| index < 0 || index >= length fields =
				error2 "evalFieldFB" ("Invalid index: "++(show index))

			| otherwise           = evalFB ee' (fields !!! index)
			where
			ee' :: ETC_Env
			ee' =  setBinds binds' ee1

		--
		--	Merging record type-checking results:
		--
		merge_itcrs :: [([TC_Res], Int)] -> [(TC_Res, Int)]
		merge_itcrs    itcrs =
			concat [[(tcr, index) | tcr <- tcrs] | (tcrs, index) <- itcrs]


	selectRUnField _ _ = error2 "selectRecField"
								"Selector is not an Identifier"


	---------------
	--	"fCall": --
	---------------
	--	Applies a given callable to a given list of single-versioned args. They
	--		all have the same NameBinds as those in the ETC_Env:
	--
	fCall:: ETC_Env -> EvalRes ->	[EvalRes] -> [EvalRes]
	--		Env		   Callable     Args		 Result(s) or []

	------------------------------
	--	The Callable is Bottom: --
	------------------------------
	fCall   _   hmm@(Bottom_ER _) _  =  [hmm]  -- Already has correct binds!

	---------------------
	--	Built-In call: --
	---------------------
	--	If the function being applied is a  BUILT-IN OPERATION,  it will not
	--		perform any further calls, so it can be invoked directly without
	--		saving the args  on the call stack.
	--
	fCall   ee1   (Full_ER _ op@(BuiltInOp _ bi_proto bi_body))  er_args
		--
		--	Check the number of arguments. If it is wrong, don't raise an error
		--		here, just produce []:
		--
		|	length bi_argtypes /= length er_args     = []

		--
		--	Now type-check the args, and if that is successful, actually apply
		--		the built-in, otherwise return []:
		--
		|	and (map tchk (zip bi_argtypes er_args)) = evalBI

		|	otherwise = []
		where
		--
		--	Get the types of formal parms:
		--
		bi_argtypes:: [AldorType]
		BuiltInT	  bi_argtypes bi_restype =  bi_proto

		--	The type-checking function. NB: partially evaluated args and Bottoms
		--		are treated as TYPE CONFORMANT:
		--
		tchk:: (AldorType, EvalRes) -> Bool
		tchk   (_,		   Bottom_ER  _    ) = True
		tchk   (_,		   Partial_ER _ _  ) = True
		tchk   (arg_type,  Full_ER    _ val) = conforms ee1 val arg_type

		--	Compute the evaluation result and type-check it against the BI's
		--		result type (returning []  if type-checking fails):
		--
		evalBI :: [EvalRes]
		evalBI =
			if  tchk (bi_restype, hres)  then  [hres]  else []
			where
			v_binds:: NameBinds
			v_binds = getBinds    ee1

			hres :: EvalRes
			hres =  case homogenise_ER_List er_args of
				Bottom_ERL 		 ->
					--	There was at least one "Bottom" in "er_args"; the whole
					--		result is undefined:
					--
					Bottom_ER v_binds

				Partial_ERL  fbs ->
					--	There was at least one partially-evaluated argument; the
					--		result is a normalised application expr:
					--	XXX UNDECOR
					Partial_ER
						v_binds
						(simpleFB (FunctionCall (simpleFB(Literal op)) fbs))

				Full_ERL    vals ->
					--	All "er_args" are fully evaluated;   ACTUALLY APPLY the
					--		built-in function HERE:
					--
					Full_ER v_binds (bi_body vals)


	-------------------------
	--	LAMBDA evaluation: --
	-------------------------
	--	Since evaluation is STRICT, if there is at least one Bottom in the args,
	--		the result is Bottom,  even if that argument value  is not actually
	--		used in the function body:
	--
	fCall   ee1  (Full_ER _ (LambdaVal (FunctionT parms _) body))  er_args  =
		ress
		where
		--	Get DefNos for the parms:
		--
		def_nos = [def_no | GenFDTE def_no _ _ _ <- map getFParm parms]

		--	Now push "er_args" (a single version) on  the EvalStack;  "pushES*"
		--		verifies that the number of args is the same as that of
		--		"def_nos" (i.e. "parms"):   default args  are deliberately not
		--		supported.
		--	Due to the STRICTNESS of Aldor evaluation,  if at least one Bottom
		--		was encountered in the  "er_args",   the result is immediately
		--		Bottom. However, we do not check here for Partials in the args:
		--		this willbe done when the function body is actually evaluated.
		--	The vals being pushed on the stack already have the same NameBinds
		--		as in "ee1" and "ee2":
		--
		ee2 :: ETC_Env
		ee2  = pushES_EE  ee1   def_nos     er_args

		ress:: [EvalRes]
		ress=  case  homogenise_ER_List     er_args   of
			Bottom_ERL -> [Bottom_ER (getBinds ee2)]
			_		   -> evalFB    ee2     body

	-----------------------------------------------
	--	Partially-Evaluated Function Expression: --
	-----------------------------------------------
	--	Reconstruct the function application with fully or partially evaluated
	--		arguments. If at least one argument is Bottom, the whole application
	--		evaluates to Bottom due to the STRICTNESS of evaluation in Aldor:
	--
	fCall   ee1  (Partial_ER _ ffb)  er_args = [part]
		where
		binds:: NameBinds
		binds = getBinds ee1

		part :: EvalRes
		part  = case  homogenise_ER_List er_args  of
			Bottom_ERL  	->
				Bottom_ER	binds

			Partial_ERL fbs ->
				Partial_ER binds (simpleFB (FunctionCall ffb fbs))
			--	XXX UNDECOR

			Full_ERL    vals->
				Partial_ER  binds (simpleFB (FunctionCall ffb
											(map (simpleFB . Literal) vals)))
			--	XXX UNDECOR

	-----------------------
	--	All other cases: --
	-----------------------
	--	Empty result (not an error, as this may be OK in multi-evaluation):
	fCall _ _ _ = []


-----------------
--	"Literal": --
-----------------
--	"Complex" Literals (e.g. those containing "complex" "TypeVal"s) may require
--		non-trivial evaluation.  NB: we don't have separate functions for  Type
--		objects evaluation, it is all done via Literals here!
--
--	(1) TypeVal FunctionT:
--		The result is either Full  (which would contain a FunctionT with norma-
--			lised FParmFBs), or Bottom:
--
evalExpr ee  (Literal (TypeVal  proto@(FunctionT _ _)))

	| _remEvalDepth ee <= 0 =  [Bottom_ER  (getBinds  ee)]

	| otherwise		    =  map mapProto(evalProto ee0 proto)
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	mapProto:: Proto_Res-> EvalRes
	mapProto   (Bottom_PR  binds')        = Bottom_ER binds'
	mapProto   (OK_PR      binds' proto') = Full_ER binds' (TypeVal proto')

--
--	(2) TypeVal RecordT:
--		The result is either Full (containing a normalised TypeVal RecordT), or
--		Bottom:
--
evalExpr ee  (Literal (TypeVal (RecordT flds )))

	| _remEvalDepth ee <= 0 = [Bottom_ER  (getBinds ee)]

	| otherwise		= map mapRecord (evalFBList ee0 flds)
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	mapRecord:: FBList_Res-> EvalRes
	mapRecord   (Bottom_FBL  binds')	   = Bottom_ER binds'
	mapRecord   (OK_FBL      binds' flds') = 
					Full_ER  binds' (TypeVal (RecordT flds'))

--
--	(3) TypeVal UnionT:
--		Similar to TypeVal RecordT:
--
evalExpr ee  (Literal (TypeVal (UnionT cases )))

	| _remEvalDepth ee <= 0 = [Bottom_ER  (getBinds ee)]

	| otherwise     = map mapUnion  (evalFBList ee0  cases)
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	mapUnion :: FBList_Res-> EvalRes
	mapUnion    (Bottom_FBL  binds')		= Bottom_ER binds'
	mapUnion	(OK_FBL      binds' cases') =
					Full_ER  binds' (TypeVal (UnionT cases'))

--
--	(4) TypeVal ADT:
--		Here the evaluation result is again Full or Bottom:
--
evalExpr ee  (Literal (TypeVal (ADT adt_no repr_fb  dt) ))

	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise			= radts
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	--	Evaluate the "repr_fb" and "dt". To be able to cross them over using
	--		our "CrossResults" function, they must  be of the same type,  so
	--		we embed them into the FBorDT type.   If a Bottom is encountered
	--		in any of these evaluations, the result is just a Bottom:
	--
	rfbs :: [FBorDT]
	rfbs =  map mapFBRes (evalFB ee0 repr_fb)

	rdts :: [FBorDT]
	rdts =  map mapDTRes (evalDT ee0 dt)

	mapFBRes :: EvalRes ->  FBorDT
	mapFBRes   (Bottom_ER   binds    ) = Bottom_FBDT binds
	mapFBRes   (Partial_ER  binds fb ) = FB1   binds fb
	mapFBRes   (Full_ER     binds val) = FB1   binds (simpleFB (Literal val))

	mapDTRes :: DT_Res  ->  FBorDT
	mapDTRes   (Bottom_DTR  binds)	   = Bottom_FBDT binds
	mapDTRes   (OK_DTR      binds dtr) = DT1   binds dtr

	--	Cross the results over:
	--	The inner dim in "resX" is of length 2:
	--
	resX 	 :: [ConsoList NameBinds FBorDT]
	resX	 =  crossWithAttrs  [rfbs, rdts]

	--	Re-construct the ADTs and convert them into "Full_ER"s:
	--
	radts:: [EvalRes]
	radts=  [Full_ER (getAttr res)
					 (TypeVal (mkADT (getAttr res) (getCList res))) |
			 res <- resX]

	mkADT:: NameBinds -> [FBorDT] -> AldorType
	mkADT   binds        [FB1 _ rfb, DT1 _ rdt] =
			ADT (newEnum ee binds adt_no) rfb rdt
	mkADT   _  _ =  error2 "mkADT" "Invalid internal call"

--
--	(5) TypeVal CategoryT: Similar to ADT:
--
evalExpr ee  (Literal (TypeVal  (CategoryT cat_no dt)))

	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise			= rcats
	where
	ee0  :: ETC_Env
	ee0   = decr_rem_depth ee

	rcats:: [EvalRes]
	rcats = map mkCat (evalDT ee0 dt)

	mkCat:: DT_Res -> EvalRes
	mkCat   (Bottom_DTR binds)     = Bottom_ER binds
	mkCat   (OK_DTR     binds rdt) =
			Full_ER     binds (TypeVal
							  (CategoryT (newEnum ee binds cat_no) rdt))

--
--	(6) LambdaVal:
--
evalExpr ee  (Literal (LambdaVal proto body))

	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise	  = rlambds
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	--	Again, use a composite type (TypeOrFB) to cross the results of "proto"
	--		and "body" evaluation:
	--
	eprotos :: [TypeOrFB]
	eprotos =  map mapProto (evalProto ee0 proto)

	ebodies :: [TypeOrFB]
	ebodies =  map mapBody  (evalFB    ee0 body)

	mapProto::  Proto_Res ->	TypeOrFB
	mapProto   (Bottom_PR binds)		 = Bottom_TFB    binds
	mapProto   (OK_PR     binds proto')  = Type2   binds proto'

	mapBody ::  EvalRes  ->	 TypeOrFB
	mapBody    (Bottom_ER binds)		 = Bottom_TFB    binds
	mapBody     other					 = FB2 (getAttr  other) (er2fb other)

	--	The inner dim is "resX" of length 2, and its binds are consolidated:
	--
	resX    :: [ConsoList    NameBinds TypeOrFB]
	resX    =  crossWithAttrs [eprotos, ebodies]

	--	Re-construct and return "LambdaVal"s:
	--
	rlambds :: [EvalRes]
	rlambds =  map (mapLambda . getCList) resX

	mapLambda :: [TypeOrFB] -> EvalRes
	mapLambda    [Bottom_TFB binds, _]      = Bottom_ER binds
	mapLambda    [_, Bottom_TFB binds]      = Bottom_ER binds
	mapLambda    [Type2 binds rp, FB2 _ rb] = Full_ER   binds (LambdaVal rp rb)
	mapLambda    _ = error2 "mapLambda" "Invalid internal call"
	
--
--	(7) All other cases: "Simple" Literals, including (but not limited to):
--
--		TypeVal BuiltInT, BuiltInOp, RecordVal, UnionVal, ADT_Val:
--
--	The NameBinds associated with a Literal are those contained in
--		"ee0", and there is just 1 version of the result:
--
--	We still need a formal recursion control here:
--
evalExpr ee  (Literal val)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]
	| otherwise			= [Full_ER   (getBinds ee0) val]
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee


----------------------
--	"Identifier" :  --
----------------------
evalExpr ee  ident@(Identifier use_no _)
	--
	--	Recursion control first:
	--
	| _remEvalDepth ee <= 0 = [Bottom_ER  (getBinds ee)]

	--	OTHERWISE: how many bindings do we have for this "ident"?
	--	None -- error:
	--
	| length targs == 0 =
		error2 "evalExpr" ("Identifier: No binding(s) for "++(show ident))

	--	OTHERWISE: get the Indentifier values for each binding  target.  Note
	--		that a single target can produce multiple values   (as the target
	--		evaluation may itself involve other overloaded Idents), hence the
	--		"concat" below:
	--
	| otherwise  = concat (map restrAndEval targs)

	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	targs:: [BindTarget]
	targs = lookupBinds (getBinds ee0) use_no

	restrAndEval:: BindTarget -> [EvalRes]
	restrAndEval   targ       =  evalTarg eeT ident targ
		where
		--	In this evaluation, the bindings of "use_no" are restricted  to the
		--		single "targ":
		eeT = setBinds (modifyBinds (getBinds ee0) use_no [targ]) ee0


---------------------
--	"IfThenElse":  --
---------------------
--	Conditional evaluation is a partial "reduction point":  we proceed with
--		only those results of  "cond_fb"  evaluation which  contain Boolean
--		values. More than 2 vals are allowed (both True and False can occur
--		under different name bindings;  the latter are retained for further
--		processing).
--	It is also possible that "cond_fb" is only partially evaluated; in that
--		case the whole unevaluated IfThenElse expr is returned (as we don't
--		know which branch to evaluate, and must not evaluate both):
--
--	If "cond_fb" is of a special form:
--		(1) Identifier,  or
--		(2) Identifier == anyFB,
--		and the Identifier is left unevaluated in the current context, then
--		the Conditional yields new local definitions for it (in the form of
--		"CondTarg"s). In case (1), True and False values for the Identifier
--		are propagated to evaluation  of the Then and Else  branches, resp;
--		in case (2), the "anyFB" evaluation result is propagated to evalua-
--		tion of the Then branch:
--
evalExpr ee    (IfThenElse cond_fb then_fb else_fb)

	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise			= ress

	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	--	Evaluate the "cond_fb",   split the whole ("all_") result into 3 parts:
	--		Bottoms ("bot_"), Partially evaluated ("part_") and Fully evaluated
	--		("full_"):
	--
	all_conds  :: [EvalRes]
	all_conds   = evalFB ee0 cond_fb

	bot_conds  :: [EvalRes]
	bot_conds   = [res | res@(Bottom_ER _) <- all_conds]

	part_conds :: [(NameBinds, AldorFBlock)]
	part_conds  = [(binds, fb ) | Partial_ER binds fb  <- all_conds]

	full_conds :: [(NameBinds, AldorValue)]
	full_conds  = [(binds, val) | Full_ER    binds val <- all_conds]

	--	For "bot_conds", return Bottoms (they are the same as "bot_conds"):
	--
	bot_ress   :: [EvalRes]
	bot_ress    = bot_conds

	--	For "part_conds", return the whole expr unevaluated (but with normalised
	--		"cond_fb"). Note that we CANNOT normalise the Then and Else branches
	--		here, as the semantics of the conditional operator does not allow us
	--		to evaluate both of them together,   and we cannot evaluate just one
	--		branch since we don't know which!
	--
	part_ress   :: [EvalRes]
	part_ress    = [Partial_ER binds (simpleFB
							         (IfThenElse norm_cond then_fb else_fb)) |
				   (binds, norm_cond) <- part_conds]
	--  XXX UNDECOR

	--	For "full_conds", reduce them further to Booleans only (skip others) and
	--		do full evaluation of the resp branches. Possibly enlarge the binds:
	--
	bool_conds  :: [(NameBinds, Bool)]
	bool_conds  =  [(binds,   b_cond) | (binds, BooleanVal b_cond) <-full_conds]

	full_ress   :: [EvalRes]
	full_ress    = concat (map thenORelse bool_conds)

	thenORelse:: (NameBinds, Bool) -> [EvalRes]
	thenORelse   (binds, b_cond) =
		if   b_cond
		then evalFB eeT then_fb
		else evalFB eeF else_fb
		where
			--	NB: The binds are enlarged here:
			eeT, eeF::  ETC_Env
			eeT = setBinds (mkCondTargs binds cond_fb True ) ee0
			eeF = setBinds (mkCondTargs binds cond_fb False) ee0

	--	Concatenate and return all results:
	--
	ress     :: [EvalRes]
	ress	  = bot_ress ++ part_ress ++ full_ress


--------------------------------
--	Logical List Expressions: --
--------------------------------
--	AndList, OrList are implemented via "evalLogList":
--
evalExpr ee  (AndList args)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]
	| otherwise			= evalLogList ee0 args And
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee


evalExpr ee  (OrList  args)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]
	| otherwise         = evalLogList ee0 args Or
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee


----------------------
--	"RestrictType": --
----------------------
--	Evaluate an expression and its "must-be" type (both to a list of results!),
--		cross these lists and retain those pairs where the 1st element (value)
--		satisfies the type given by the 2nd one:
--
evalExpr ee  (RestrictType expr_fb type_fb)

	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise			= ressOK

	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	--	Evaluate the "expr_fb" and the "type_fb":
	--
	vals  :: [EvalRes]
	vals   = evalFB ee0 expr_fb

	types :: [EvalRes]
	types  = evalFB ee0 type_fb

	--	Cross the results over and retain those which have compatible binds;
	--		the inner dim in "ressX" is of length 2:
	--
	ressX :: [EvalRes_CL]
	ressX  = crossWithAttrs [vals, types]

	--	Retain those crossed results where the "val" satisfies the "ttype":
	--		only the "val" and "binds" are returned.     If a result is not
	--		fully evaluated,  a normalised "RestrictType" expr is returned:
	--
	ressOK:: [EvalRes]
	ressOK = mapMaybe filterConform ressX

	-----------------------
	--	"filterConform": --
	-----------------------
	filterConform:: EvalRes_CL -> Maybe EvalRes

	filterConform   ercl =
		case  getCList ercl of
			--
			--	(1) Bottoms in both LHS and RHS are allowed, and treated as
			--		conformant to everything   (in order to propagate them,
			--		not to drop them silently).   NB: if the type evaluates
			--		to "Bottom" but the value does not,   we still return a
			--		Bottom:
			--
			[Bottom_ER _, _] ->   Just (Bottom_ER (getAttr ercl))
			[_, Bottom_ER _] ->   Just (Bottom_ER (getAttr ercl))

			--
			--	(2) Both LHS and RHS  are fully  evaluated,  and the RHS is
			--		indeed a Type: check the weak type conformance:
			--
			[lhs@(Full_ER _ val),  Full_ER _ (TypeVal ttype)] ->
				let
					ee' :: ETC_Env
					ee' =  setBinds (getAttr  ercl) ee0
				in
					if   conforms  ee' val ttype
					then Just   lhs
					else Nothing

			--
			--	(3) Both LHS and RHS are fully-evaluated, but the RHS is not
			--		a Type: drop this combination!
			--
			[Full_ER _ _, Full_ER _ _] -> Nothing

			--
			--	(4) LHS and/or RHS is partially evaluated: re-construct the
			--		expr; NB: no Bottom can occcur here, as they are already
			--		accounted for above:
			--
			[lhs, rhs] ->
				Just (Partial_ER  (getAttr  ercl)
								  (simpleFB (RestrictType (er2fb lhs)
														  (er2fb rhs))))
			--	XXX UNDECOR

			--
			--	(5) The list is not of length 2: this cannot happen:
			--
			_ -> error2 "filterConform" "Invalid internal call"


--------------------
--	"ConvertTo":  --
--------------------
--	Type conversion is liberal but not uncontrolled. The following conversions
--		are allowed:
--		T1   <=> T2   iff T1        and T2        are compatible;
--		T    <=> ADT  iff T         and rep(ADT)  are compatible;
--		ADT1 <=> ADT2 iff rep(ADT1) and rep(ADT2) are compatible.
--	"Compatible" means: type equality test does not yeild "Just False":
--	NB: This function is in particular used to create "ADT_Val"s  from
--		the values of representation types:
--
evalExpr ee  (ConvertTo expr_fb ttype_fb)

	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise				= ressC

	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	--	Evaluate the expr and the type; both yield a list of results:
	--
	ress_e = evalFB ee0 expr_fb
	ress_t = evalFB ee0 ttype_fb

	--	Construct all combinations of them, as a list of tuples, and retain
	--		those combinations which have compatible binds. In "ressX", the
	--		inner dim is of length 2:
	--
	ressX :: [EvalRes_CL]
	ressX =  crossWithAttrs [ress_e, ress_t]

	--
	--	Now convert compatible types using "doConv":
	--
	ressC :: [EvalRes]
	ressC =  concat [doConv (getAttr resX) (getCList resX) | resX <- ressX]

	----------------
	--	"doConv": --
	----------------
	--	NB: Consolidated NameBinds are given (for convenience) by the 1st arg;
	--		there is no need to extract them from the 2nd arg elements:
	--
	doConv:: NameBinds -> [EvalRes] -> [EvalRes]

	--
	--	(1) Expr and/or Target Type evaluates to Bottom: produce Bottom:
	--
	doConv   binds        [Bottom_ER _, _]  = [Bottom_ER binds]
	doConv   binds		  [_, Bottom_ER _]  = [Bottom_ER binds]

	--
	--	(2) ADT CAST:	  "ADT_Val" into another ADT:
	--		"val_e" is an "ADT_Val". Conversion is possible  iff "val_t"
	--		is an ADT with compatible ultimate representation:
	--
	doConv   binds 		  [Full_ER _ (ADT_Val adt_e aval_e),
						   Full_ER _ (TypeVal adt_t@(ADT _ _ _))] = conv_res
		where
		--	Get the rep of "adt_e":
		--
		eeT   = setBinds binds  ee0

		rvs_e:: [EvalRes]
		rvs_e = getUltADTrep eeT adt_e

		rvs_t:: [EvalRes]
		rvs_t = getUltADTrep eeT adt_t

		--	Cross the evaluated representations and retain those merged results
		--		which have compatible binds:
		--
		repsX:: [EvalRes_CL]
		repsX = crossWithAttrs [rvs_e, rvs_t]

		--	Filter "repsX" further: retain only type-compatible combinations of
		--		reps AND those containing a Bottom:
		--
		repsC:: [EvalRes_CL]
		repsC = filter compaTypes repsX

		--	"compaTypes":
		--
		--	Verifies that the type given by the LHS  is convertible into one
		--		given by the RHS. Returns False only if the types are surely
		--		incompatible, or if one or both args are fully evaluated but
		--		not to "TypeVal"(s):
		--	NB: We know that vectors of values are of length 2:
		--
		compaTypes:: EvalRes_CL -> Bool

		compaTypes   repX =
			case getCList  repX of
				--
				--	Both LHS and RHS are fully evaluated to TypeVals:
				--
				[Full_ER _  (TypeVal lt),  Full_ER _ (TypeVal rt)] ->
					case  adeq (etc2ade eeT) lt rt of
						Yes_ADER   _  		-> True  -- XXX: Binds thrown away!
						No_ADER				-> False
						Bottom_ADER			-> True	 -- Undecided equality,
													 --	treated as OK

				--	LHS and/or RHS is fully evaluated but  NOT to a TypeVal:
				--		negative result no matter what the opposite side is:
				--
				[Full_ER _ _, _] -> False
				[_, Full_ER _ _] -> False

				--	The rest: LHS and RHS are not fully evaluated;   each of
				--		them is either Partial or Bottom.   We consider them
				--		to be compatible!
				--
				_   ->  True

		--	For each remaining (compatible) combination of reps,  compute the
		--		conversion result. Note that the evaluated reps themselves do
		--		not come into the result, only the resp binds do.   The value
		--		of the result is still formed by "aval_e" and "adt_t":
		--
		conv_res :: [EvalRes]
		conv_res =  [Full_ER (getAttr repC) (ADT_Val adt_t aval_e) |
				    		  repC <- repsC]

	--
	--	(3) ADT CTOR:    Rep val into an ADT:
	--		"val_e" is a rep value (NOT an ADT_Val, otherwise eqn (1) works),
	--		"val_t" is an ADT: this is an ADT CTOR:    it succeeds if the rep
	--		value is indeed compatible with the rep type declared in the ADT:
	--
	doConv   binds        [Full_ER _ val_e,
						   Full_ER _ (TypeVal adt_t@(ADT _ _ _))] = conv_res
		where
		eeT  :: ETC_Env
		eeT   = setBinds binds ee0

		--	Get the rep(s) of the RHS:
		--
		rvs_t:: [EvalRes]
		rvs_t = getUltADTrep   eeT  adt_t

		--	Retain those rep values (must be Type Literals) to which "val_e"
		--		is weakly conformant.  The reps which are not Type values at
		--		all are skipped without signalling an error:
		--
		rvsC :: [EvalRes]
		rvsC  = filter valConforms rvs_t

		valConforms::  EvalRes -> Bool
		valConforms   (Full_ER _ (TypeVal t)) = conforms  eeT val_e  t
		valConforms    _              	      = False		-- Not a TypeVal

		--	For each retained rep, generate a conversion result. The rep value
		--		itself is not used, only its binds:
		--
		conv_res :: [EvalRes]
		conv_res =  [Full_ER (getAttr rvC) (ADT_Val adt_t val_e) | rvC<- rvsC]

	--
	--	(4) ADT DTOR:   ADT_Val into the Rep type:
	--		"rv_t" is a type compatible with the ultimate rep of the LHS (rather
	--		than an ADT -- in the latter case eqn (1) works):
	--
	doConv  binds		  [Full_ER _ (ADT_Val adt_e aval_e),
				   		   Full_ER _ (TypeVal rv_t)]	   = conv_res
		where
		--	Get the ultimate rep of the LHS:
		--
		eeT  :: ETC_Env
		eeT   = setBinds binds ee0

		rvs_e:: [EvalRes]
		rvs_e = getUltADTrep eeT adt_e

		--	Retain those "rvs_e" results which are compatible with "rv_t".
		--		Those results for which the compatibility test hits Bottom
		--		are retained as being OK:
		--
		rvsC :: [EvalRes]
		rvsC  = filter compaTypes rvs_e

		compaTypes :: EvalRes -> Bool

		compaTypes   (Full_ER _  (TypeVal e)) =
			case adeq (etc2ade eeT)   e  rv_t  of
				Yes_ADER _    -> True   -- XXX: Binds thrown away!
				No_ADER		  -> False
				Bottom_ADER   -> True   -- Undecided, considered to be OK

		compaTypes   _        =  False  -- Not a TypeVal

		--
		--	In each case (of "rvsC"), the returned value is still "aval_e",
		--		but with the new resp binds:
		--
		conv_res :: [EvalRes]
		conv_res =  [Full_ER (getAttr  rvC) aval_e | rvC <- rvsC]

	--
	--	(5) GENERAL TYPE CAST:
	--		Does not modify "val_e" as long as it weakly conforms to the type
	--		given by the RHS.  "val_e" is not an ADT_Val and "rv_t" is not an
	--		ADT:
	--
	doConv  binds		  [lhs@(Full_ER _ val_e), Full_ER _ (TypeVal rv_t)] =

		if   conforms  eeT  val_e  rv_t
		then [setAttr binds  lhs]
		else []
		where
		eeT	:: ETC_Env
		eeT	=  setBinds binds ee0

	--
	--	(6) IMPROPER:
	--		Both LHS and RHS are fully evaluated, but the RHS is not a TypeVal:
	--		skip this combination:
	--
	doConv  _		[Full_ER _ _,  Full_ER _ _] = []

	--
	--	(7) UNEVALUATED LHS and/or RHS: re-construct the expr:
	--
	doConv  binds	[lhs, rhs] =
		[Partial_ER binds (simpleFB (ConvertTo (er2fb lhs) (er2fb rhs)))]
	--  XXX UNDECOR

	--
	--	(8) ERROR: The list is not of length 2: this cannot happen:
	--
	doConv  _ hmm = error2 "doConv" ("Invalid internal call on "++(show hmm))


--------------------
--	"RecordCtor": --
--------------------
--	Multi-evaluate each component,  cross the results and construct the record
--		for each valid combination. Each result version may be fully evaluated
--		(to a RecordVal) or partially evaluated:
--
evalExpr ee  (RecordCtor fields)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise			= recsR

	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	--	NB: components are evaluated independently, in the same "ee0"; the
	--		results are merged:
	--
	ressX::  [EvalRes_CL]
	ressX =  crossWithAttrs (map (evalFB ee0) fields)

	--	Each component of "ressX" is a vector which can be transformed into the
	--		resulting record (or Bottom).   Depending on whether all fields are
	--		fully evaluated, we can produce "RecordVal" or a "RecordCtor":
	--
	recsR::  [EvalRes]
	recsR =  map mkRec ressX

	mkRec::  EvalRes_CL -> EvalRes
	mkRec    ercl =
		case homogenise_ER_List (getCList ercl)   of
			Bottom_ERL 			->  Bottom_ER  binds
			Partial_ERL  pfields->  Partial_ER binds	-- XXX UNDECOR
										  	   (simpleFB (RecordCtor pfields))
			Full_ERL  vals_only ->  Full_ER    binds	 (RecordVal vals_only)
		where
		binds :: NameBinds
		binds =  getAttr   ercl


--------------------
--	"UnionCtor":  --
--------------------
--	We just verify that the "tag" is an Identifier (may be fully-qualified),
--		and evaluate the "body_fb" to construct a "UnionVal"  (if "body_fb"
--		is fully evaluated) or re-construct a "UnionCtor".
--	"UnionVal", if constructed, inherits the UseNo of the ctor's Identifier,
--		so the Val's type can later be re-constructed:
--
evalExpr ee  (UnionCtor tag_expr body_fb)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]

	| otherwise			= ress

	where
	ee0   :: ETC_Env
	ee0   =  decr_rem_depth ee

	ress ::  [EvalRes]
	ress =   map mkUnion (evalFB ee0 body_fb)

	mkUnion::EvalRes  -> EvalRes

	mkUnion  res@(Bottom_ER    _ ) = res
	mkUnion (Partial_ER binds fb ) = Partial_ER	-- XXX UNDECOR
									 binds  (simpleFB (UnionCtor tag_expr fb))
	mkUnion (Full_ER    binds val) = Full_ER    binds (UnionVal  tag_expr val)


--------------------
--	"UnionCase":  --
--------------------
evalExpr ee (UnionCase union tag@(Identifier _ _)) = fress ++ press ++ bots
	where
	--	Evaluate  the "union" first:
	--
	uress :: [EvalRes]
	uress =  evalFB ee union

	--	Split "uress" into 3 groups: "UnionVal"s, Partial results and Bottoms.
	--		Note that Full results which are not "UnionVal"s are DISCARDED (ie
	--		we do NOT produce any result for them, even the Boolean False). In
	--		"UnionVal"s, we only need the corresp "UseNo"s and Binds:
	--
	uuns  :: [(UseNo, NameBinds)]
	uuns  =  [(use_no,binds) | Full_ER binds
							   (UnionVal (Identifier use_no _) _) <- uress]

	uparts:: [EvalRes]
	uparts=  [upart  | upart@(Partial_ER _ _) <- uress]

	--	"bots" are propagated directly to the result:
	--
	bots  :: [EvalRes]
	bots  =  [bot    | bot@(Bottom_ER _) <- uress]

	--	Now: search for "uuns" whose "UseNo"s are AD-equivalent  to the "UseNo"
	--		in the Selector. AD-equivalence is computed wrt "UseNo"s only,  not
	--		names, so we can use any name in the LHS (e.g. ""):
	--	XXX: we currently don't generate an error if the "tag"  does not belong
	--		to the set of permissible tags for the type of the "union"; in that
	--		case, we just produce a negative result.
	--
	fress :: [EvalRes]
	fress =  map eqUseNos uuns

	eqUseNos :: (UseNo, NameBinds) -> EvalRes
	eqUseNos    (fun,   binds)  =
		case adeq (etc2ade (setBinds binds ee)) (Identifier fun [""]) tag of
			Yes_ADER binds' ->
				Full_ER   binds' (BooleanVal True)

			Bottom_ADER     ->
				Bottom_ER binds

			No_ADER			->
				Full_ER   binds  (BooleanVal False)
	
	--	"uparts": normalise the "UnionSel" (XXX: UNDECOR):
	--
	press :: [EvalRes]
	press =  [Partial_ER binds' (simpleFB (UnionCase union' tag)) |
			  Partial_ER binds' union' <- uparts]


evalExpr _  (UnionCase _ hmm) =
	error2 "evalExpr--UnionCase" ("RHS is not an Identifier: "++(show hmm))

---------------
--	"Add1":  --
---------------
--	Creates an ADT Literal if possible, or just a normalised "Add1" expr:
--
evalExpr ee  (Add1 add_no fb)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]
	| otherwise         = ress
	where
	ee0   :: ETC_Env
	ee0   =  decr_rem_depth ee

	fbrs  :: [EvalRes]
	fbrs  =  evalFB   ee0   fb

	ress  :: [EvalRes]
	ress  =  foldl  mkADT1  []  fbrs

	-----------------
	--	"mkADT1":  --
	-----------------
	mkADT1:: [EvalRes]-> EvalRes -> [EvalRes]

	mkADT1   curr   bot@(Bottom_ER  _)     					= bot: curr

	--	If "fb" evaluated to a Full Value, it must be an ADT Literal to be
	--		included in the result. Otherwise, the result is dropped:
	--
	mkADT1   curr    ok@(Full_ER  _  (TypeVal (ADT _ _ _))) = ok : curr
	mkADT1   curr       (Full_ER  _  _)						= curr

	--	If "fb" evaluated to a Partial, the entire result of evaluating the
	--		"Add1" expr is either Full (if an ADT Literal is obtained),  or
	--		a normalised "Add1" otherwise:
	mkADT1   curr		(Partial_ER binds rfb) =
		case mkADT_fromFB rfb of
			Just adt ->
				(Full_ER    binds (TypeVal  (setPerCentDefNo_T adt nw_no)))  :
				curr
			Nothing  ->
				(Partial_ER binds (simpleFB (Add1 nw_no rfb))) :
				curr
		where
		nw_no :: DefNo
		nw_no =  newEnum ee binds add_no


---------------
--	"Add2":  --
---------------
--	Dynamic ADT extension:
--
evalExpr ee  (Add2 add_no fb1 fb2)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]
	| otherwise         = ress
	where
	ee0   :: ETC_Env
	ee0   =  decr_rem_depth ee

	--	Evaluate both FBs and cross the results:
	fbr1s :: [EvalRes]
	fbr1s =  evalFB   ee0  fb1

	fbr2s :: [EvalRes]
	fbr2s =  evalFB   ee0  fb2

	xfbrs :: [EvalRes_CL]
	xfbrs =  crossWithAttrs [fbr1s, fbr2s]
	--	In "xfbrs": Outer dim: versions; inner dim: [resOf1, resOf2]

	ress  :: [EvalRes]
	ress  =  foldl  mkADT2  []  xfbrs

	-----------------
	--	"mkADT2":  --
	-----------------
	mkADT2:: [EvalRes] -> EvalRes_CL -> [EvalRes]
	mkADT2   curr     clres =
		case getCList clres of
			--	At least one operand evaluated to Bottom -> result is Bottom:
			[Bottom_ER _, _] -> (Bottom_ER binds) : curr
			[_, Bottom_ER _] -> (Bottom_ER binds) : curr

			--	Now, the final result is an ADT object iff:
			--	(*) the 1st operand evaluated to an ADT object, AND
			--	(*) the 2nd operand evaluated either  to  a  DefTable without
			--		"Rep", or to an ADT with "Rep" equivalent to that  of the
			--		1st result:
			[res1,     res2] ->
				case er2adtr res1 of
					Just (OK_ADT (ADT _ rep1 dt1)) ->
						--	The 1st result is an ADT (its DefNo is discarded):
						--
						case   er2adtr res2 of
							Just (OK_ADT (ADT _ rep2 dt2)) ->
								--	OK, both are ADTs: check that their "Rep"s
								--		are AD-compatible, and if so, just re-
								--		tain the 1st "Rep" in the combined res.
								--		Otherwise, drop the result:
								case adeq  (etc2ade ee0) rep1 rep2 of
									Bottom_ADER    ->
										--	Hmm: equivalence undecided. What to
										--		do: return a Partial or Bottom?
										--	XXX Produce Bottom for now:
										(Bottom_ER binds) : curr

									Yes_ADER nbinds ->
										--	OK, they are equivalent:
										--		do the extension:
										(Full_ER nbinds
											(TypeVal (mkADT3 rep1 dt1 dt2))):
										curr

									No_ADER ->
										--  "Rep"s Not equivalent -- drop the
										--	result:
										curr

							Just (OK_ADT _) ->
								-- 1st res is marked as an ADT but is not such:
								--	   This does not happen:
								error2 "mkADT2" "Invalid OK_ADT"

							Just (OK_DT dt2) ->
								--	The 2nd result is just a DT:  do direct
								--		extension:
								(Full_ER binds
									(TypeVal (mkADT3 rep1 dt1 dt2))) :
								curr

							Nothing ->
								--	2nd result is neither ADT nor DT:  Make
								--		a normalised "Add2" expr in an UNDECORd
								--		SimpleFB:
								(Partial_ER binds
								    (simpleFB
									(Add2  nw_no
									(er2fb res1) (er2fb res2)))) : curr
					_ ->
						--	The 1st result is not an ADT object:
						-- 	Make a normalised "Add2" expr in an UNDECORd
						--		SimpleFB:
						(Partial_ER binds
							(simpleFB
							(Add2 nw_no (er2fb res1) (er2fb res2)))) : curr
			_ ->
				--	The X-list is not of length 2: this does not happen:
				error2 "mkADT2" "Invalid internal call"
		where
		binds :: NameBinds
		binds =  getAttr clres

		nw_no :: DefNo
		nw_no =  newEnum ee binds add_no

		-----------------
		--	"mkADT3":  --
		-----------------
		mkADT3 :: AldorFBlock -> DefTable -> DefTable -> AldorType
		mkADT3    rep_fb         dt1         dt2      =  new_adt
			where
			new_adt :: AldorType
			new_adt =  ADT nw_no rep_fb  sum_dt

			sum_dt :: DefTable
			sum_dt  =  mergeDTs  dt1  dt2


---------------
--	"With1": --
---------------
--	Creates a Full Category object if possible, or a normalised Partial result:
--
evalExpr ee (With1 with_no fb)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]
	| otherwise         = ress
	where
	ee0   :: ETC_Env
	ee0   =  decr_rem_depth ee

	fbrs  :: [EvalRes]
	fbrs  =  evalFB     ee0 fb

	ress  :: [EvalRes]
	ress  =  foldl  mkCat1  []  fbrs

	----------------
	--	"mkCat1": --
	----------------
	mkCat1:: [EvalRes] -> EvalRes -> [EvalRes]

	--	"Bottom" is just a Bottom:
	mkCat1   curr  bot@(Bottom_ER _) = bot : curr

	--	"Full" result of evaluating the "fb" can be converted into a Category
	--		only if it is already a Category object:
	mkCat1   curr  cat@(Full_ER _  (TypeVal (CategoryT _ _))) = cat : curr
	mkCat1   curr      (Full_ER _   _) 						  = curr

	--	"Partial" "fb" result yields a "Full" or UNDECORd "Partial" final res:
	mkCat1   curr   (Partial_ER binds  rfb) =
		case getJustDT rfb of
			Just dt ->
				(Full_ER    binds (TypeVal  (CategoryT nw_no dt))) : curr
			Nothing ->
				(Partial_ER binds (simpleFB (With1     nw_no rfb))): curr
		where
		nw_no:: DefNo
		nw_no = newEnum ee binds with_no


---------------
--	"With2": --
---------------
--	Dynamic Category extension. Note that we must be careful with evaluating
--		"fb1" and "fb2" --  if any of them yields just a normalised "VoidFB"
--		(because it contained "VoidFB" at the inner-most level, rather than
--		a "SimpleFB"), then we must evaluate its DefTable. Nested DefTables
--		are NOT allowed:
--
evalExpr ee (With2  with_no fb1 fb2)
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee)]
	| otherwise         = ress
	where
	ee0   :: ETC_Env
	ee0   =  decr_rem_depth ee

	--	Evaluate both FBs and cross the results:
	fbdts1:: [FBorDT]
	fbdts1=  evalFB_2DT  ee0  fb1

	fbdts2:: [FBorDT]
	fbdts2=  evalFB_2DT  ee0  fb2

	xfbdts:: [ConsoList  NameBinds FBorDT]
	xfbdts = crossWithAttrs   [fbdts1, fbdts2]
	--	In "xfbdts": Outer dim: versions; Inner dim: [resOf1, resOf2]

	ress  :: [EvalRes]
	ress  =  foldl  mkCat2  []  xfbdts

	-----------------
	--	"mkCat2":  --
	-----------------
	mkCat2:: [EvalRes] -> (ConsoList NameBinds FBorDT) -> [EvalRes]
	mkCat2   curr     clres   =
		case getCList clres  of
			--	At least one operand evaluated to Bottom -> result is Bottom:
			[Bottom_FBDT _, _] -> (Bottom_ER binds) : curr
			[_, Bottom_FBDT _] -> (Bottom_ER binds) : curr

			--	If got 2 DefTables: a Category object can directly be produced
			--		by merging them:
			[DT1 _ dt1, DT1 _ dt2]  ->
				let
					sum_dt :: DefTable
					sum_dt =  mergeDTs  dt1  dt2

					new_cat:: AldorType
					new_cat=  CategoryT nw_no sum_dt
				in
					(Full_ER binds (TypeVal new_cat)) : curr

			--	At least one FB: re-construct the "With2" operator:
			--
			[r1, r2] ->
				(Partial_ER  binds
				(simpleFB (With2 nw_no (fbdt2fb r1)(fbdt2fb r2)))) : curr

			_		 ->
				--	Not a two-element list: Cannot hapen:
				error2 "mkCat2" "Invalid internal call"
		where
		binds :: NameBinds
		binds =  getAttr clres

		nw_no :: DefNo
		nw_no =  newEnum ee binds with_no


--------------------
--	"VoidExpr":   --
--------------------
--	Evaluation returns the current NameBinds and TrivVal:
--	XXX: is it the right thing to do?
--
evalExpr ee  VoidExpr
	| _remEvalDepth ee <= 0 = [Bottom_ER (getBinds ee )]
	| otherwise			= [Full_ER   (getBinds ee0) TrivVal]
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee


-- =========== --
--	"evalFB":  --
-- =========== --
--	SimpleFB and NestedFB can be completely or partially evaluated. If the value
--		returning inner FB of a NestedFB cannot be completely evaluated, do we
--		have to evaluate its DefTable as well to preserve the context  for the
--		partially-evaluated component? -- The current solution is NO: all these
--		dependencies move into the normalised expression anyway!
--	VoidFB and FParmFB can only be partially evaluated, so for FParmFBs, the
--		(singleton) DefTable is always explicitly evaluated.
--	FBs resulting from partial evaluation are left UNDECORATED at this stage:
--
evalFB   :: ETC_Env-> AldorFBlock  -> [EvalRes]

--	(0) Cached version of "EvalFB":
--
evalFB   ee fb
	--	If caching is off, or the "fb" is undecorated (can it occur here???) so
	--		that it's result cannot be cached, we evaluate "fb" directly. Also,
	--		if evaluation is requested with non-[] stack, cache cannot be used:
	--
	| not    (_useCaching ee) || getFBNo_U fb == noNUMBER ||
	  length (_evalStack  ee) /= 0 =
	  evalFB_R ee fb

	--	Otherwise: check for the cached results:
	--
	| otherwise    =
		case fromCache  fb_key of
			Nothing   ->
				--	Evaluate this FB and save the results in the cache:
				toCache fb_key (evalFB_R ee fb)
			Just ress ->
				--	Return the cached value:
				ress
	where
	fb_key :: EvalCacheKey
	fb_key =  FB_Key (getFBNo fb) (getBinds ee)

--
--	"EvalFB_R":
--	The actual (uncached) evaluation:
--
evalFB_R :: ETC_Env -> AldorFBlock -> [EvalRes]


--	(1) "SimpleFB":
--	If it is partially evaluated (to another SimpleFB), the FBNo of the original
--		FB is installed in the result:
--
evalFB_R ee (SimpleFB fb_no expr _)
	| _remEvalDepth ee <= 0 = [Bottom_ER  (getBinds ee)]
	| otherwise		    = map (setPartialFBNo fb_no) (evalExpr ee0 expr)
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth  ee

--
--	(2) "NestedFB": Similar to "SimpleFB" above:
--	XXX: Should we normalise and retain the DefTable   if "nfb" evaluates only
--		Partially?  --  Currently the DefTable is just dropped, but "nfb" gets
--		evaluated in its context:
--
evalFB_R ee (NestedFB fb_no _ nfb _)
	| _remEvalDepth ee <= 0 = [Bottom_ER     (getBinds ee)]
	| otherwise         = map (setPartialFBNo fb_no) (evalFB   ee0 nfb)
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth  ee

--
--	(3) "VoidFB":
--	Always evaluates to Partial (the same VoidFB): XXX why not to "VoidVal"???
--
evalFB_R ee hmm@(VoidFB _)
	| _remEvalDepth ee <= 0 = [Bottom_ER   (getBinds ee)]
	| otherwise         = [Partial_ER  (getBinds ee) hmm]

--
--	(4) "FParmFB":
--	Always evaluates to Partial (FParmFB), or Bottom. The uplink is retained
--		(XXX is it OK?), but FBNo modified:
--
evalFB_R ee (FParmFB  fb_no dt uplink)
	| _remEvalDepth ee <= 0 = [Bottom_ER   (getBinds ee)]
	| otherwise         = map  mapFParm (evalDT   ee0 dt)
	where
	ee0:: ETC_Env
	ee0 = decr_rem_depth ee

	mapFParm:: DT_Res  ->   EvalRes
	mapFParm   (Bottom_DTR  binds)	   = Bottom_ER    binds
	mapFParm   (OK_DTR      binds dt') =
		Partial_ER   binds (FParmFB  (newEnum ee binds fb_no) dt' uplink)


-- ===================== --
--	Auxiliary "eval*"s:  --
-- ===================== --
-----------------
--	"evalDT":  --
-----------------
evalDT   :: ETC_Env -> DefTable -> [DT_Res]

evalDT   ee dt = rdts
	where
	--	Take the DT apart: construct the lists of Entries and Imports:
	--
	entries :: [FlatDTE]
	entries =  foldDT (:) [] dt

	imports :: [AldorExpr]
	imports =  getImports    dt

	--	Evaluate all entries in the same env. For each entry, a list of
	--		results is obtained:
	--
	eentrs  :: [[FDTE_Res]]   -- Outer dim: entries; inner dim: vers
	eentrs  =  map evalFDTE entries

	--	Cross over all versions in "eentrs" and unify the binds:
	--
	xentrs :: [ConsoList NameBinds FDTE_Res]
	xentrs =  crossWithAttrs eentrs

	--	Finally: Construct a resulting DefTable from each element of "xentrs":
	--
	rdts   :: [DT_Res]
	rdts   =  map (mkDT . getCList)   xentrs

	------------------
	--	"evalFDTE": --
	------------------
	--	Evaluates an FDTE into a list of result versions:
	--
	evalFDTE:: FlatDTE -> [FDTE_Res]
	--
	--	(1) "GetFDTE"s require full evaluation:
	--
	evalFDTE   (GenFDTE def_no name type_expr body_fb) = ent_ress
		where
		--	Evaluate the Type and Body and cross the results over:
		--
		etypes  :: [EvalRes]
		etypes  =  evalTypeExpr ee def_no type_expr

		ebodies :: [EvalRes]
		ebodies =  evalFB   ee body_fb

		epairs  :: [EvalRes_CL]
		epairs  =  crossWithAttrs [etypes, ebodies]
		--
		--	Convert "epairs" into a list of resulting FDTE_Res. The resulting
		--		Partial FDTEs have the same DefNos  as the  corresp  original
		--		ones IFF that FDTE actually appears in FParmFB (this is a TRICK
		--		to make re-constructed function applications working),  and are
		--		UNDECORATED in all other cases.
		--	If either Type or Body in a def evaluated to Bottom, Bottom_FDTER
		--		is produced:
		--
		ent_ress:: [FDTE_Res]
		ent_ress=  map mapFDTER epairs

		--
		--	"mapFDTER":
		--
		mapFDTER:: EvalRes_CL -> FDTE_Res

		mapFDTER   ercl =
			case getCList ercl of
				-- Type expr evaluated to Bottom:
				[Bottom_ER _,   _] -> Bottom_FDTER binds

				-- Body FB   evaluated to Bottom:
				[_,   Bottom_ER _] -> Bottom_FDTER binds

				-- Generic case:
				[type_er, body_er] ->
					OK_FDTER     binds
						(GenFDTE def_no' name (er2expr type_er) (er2fb body_er))

				hmm	->
					error2 "mapFDTER" ("Invalid internal call on "++ (show hmm))
			where
			binds   :: NameBinds
			binds   =  getAttr ercl

			def_no' :: DefNo
			def_no' =  newEnum ee binds def_no

	--
	--	(2) "RenFDTE"s do not contain any evaluatable parts, so they are just
	--		undecorated:
	--
	evalFDTE (RenFDTE _  name ident_expr) =
			 [OK_FDTER   (getBinds  ee) (RenFDTE noNUMBER name ident_expr)]

	--------------
	--	"mkDT":
	--------------
	--	Within the argument list, all elements have the same  NameBinds,  as a
	--		result of "crossWithAttrs". If the list is empty, the binds of the
	--		result will be taken from the env:
	--
	mkDT   :: [FDTE_Res]  -> DT_Res

	mkDT	  fdtrs
		| any isBottom fdtrs  = Bottom_DTR  dt_binds
		| otherwise   		  = OK_DTR      dt_binds dt2
		where
		--	When there are no Bottoms:
		--	Attach all "fdtrs" to an initially empty DT:
		--
		dt1:: DefTable
		dt1 = foldl  attachFDTER emptyDT fdtrs

		attachFDTER:: DefTable -> FDTE_Res -> DefTable
		attachFDTER   curr  (OK_FDTER      _  fdte) = attachFDTE curr fdte
		attachFDTER   _     (Bottom_FDTER  _      ) =
			error2 "attachFDTER" "Bottom_FDTER encountered"

		--	Copy all imports (undecorating them) from the original DT:
		--
		dt2:: DefTable
		dt2 = foldl  attachUnDecImport dt1 imports

		attachUnDecImport :: DefTable -> AldorExpr -> DefTable

		attachUnDecImport    curr       (Identifier _  qname)  =
			attachImport     curr (Identifier noNUMBER qname)

		attachUnDecImport    _ hmm =
			error2 "attachUnDecImport" ("Invalid import: "++(show hmm))

		--	Construct "dt_binds" for the result: See the comment at the head
		--		of this function:
		--
		dt_binds:: NameBinds
		dt_binds = case fdtrs of
			fdtr : _ -> getAttr  fdtr
			[]		 -> getBinds ee


-------------------
--	"evalProto": --
-------------------
--	Returns a list of "Proto_Res" which are either Bottom or OK. This function
--		does NOT decrement the "_remEvalDepth" in "ee":
--
evalProto  :: ETC_Env -> AldorType  -> [Proto_Res]

evalProto ee (FunctionT parms retype) = ress
	where
	--	Evaluate the "parms" and the "retype", the latter as a singleton list
	--		for compatibility (for crossing) with the former:
	--
	eparms :: [FBList_Res]
	eparms =  evalFBList   ee parms

	ert    :: [FBList_Res]
	ert    =  evalFBList   ee [retype]

	--	Cross the results over:
	--	The outer dim is X-versions, the inner dim is [parms, retype]:
	--
	eprotos:: [ConsoList NameBinds FBList_Res]
	eprotos=  crossWithAttrs     [eparms, ert]

	--	Re-construct the protos or return "Bottom" if one is encountered:
	--
	ress   :: [Proto_Res]
	ress    = map (fbl2proto . getCList) eprotos

	--
	--	"fbl2proto":
	--
	fbl2proto:: [FBList_Res] -> Proto_Res
	--	The arg list is always of length 2, with consolidated binds

	fbl2proto   [Bottom_FBL binds, _] = Bottom_PR  binds   -- Parms  is Bottom
	fbl2proto   [_, Bottom_FBL binds] = Bottom_PR  binds   -- ReType is Bottom

	fbl2proto   [OK_FBL   binds  parms',  OK_FBL _  [retype']] =
		OK_PR   binds (FunctionT parms'   retype')	   -- Generic case

	fbl2proto   hmm  =
		error2  "fbl2proto" ("Invalid args: "++(show hmm))


evalProto _ hmm  =  error2   "evalProto" ("Not a FunctionT: "++(show hmm))


---------------------
--	"evalFBList":  --
---------------------
--	Evaluates all list elements, crosses over the results  and returns them as
--		lists of Bottom or Partial elements. There are no Full elements in the
--		result since all values (if encountered) are lifted to FBs.
--	This function is used e.g. for evaluation of RecordT / UnionT Literals and
--		function Protos. NB: it does NOT decrement the "_remEvalDepth" in "ee".
--	The result always contains CONSOLIDATED NameBinds. The FBs in the "OK_FBL"
--		come in the same order as the original FBs:
--
evalFBList:: ETC_Env -> [AldorFBlock] ->  [FBList_Res]
										  -- All result versions
evalFBList   ee fbs  =  ress2
	where
	ress0 :: [[EvalRes]]
	ress0 =  map (evalFB ee) fbs

	ress1 :: [EvalRes_CL]
	ress1 =  case crossWithAttrs ress0 of
--				[] ->
--					--	Alarm! Unification failed!
--					trace ("\"evalFBList: ALARM:\n"++(show ress0)) []
				ok -> ok
	--
	--	NameBinds are consolidated here, so they are then passed as the 1st arg
	--		to "erl2fbl",  and that function does not need to examine the binds
	--		of its 2nd arg.

	ress2 :: [FBList_Res]
	ress2 =  [erl2fbl (getAttr ercl) (getCList ercl) [] | ercl <- ress1]

	--	"erl2fbl":
	--	The 2nd arg accumulates the result. If the 1st arg is all spent and no
	--		Bottoms are encountered,    "Partial" with the 2nd arg contents is
	--		returned;  if any Bottom is encountered, the Bottom is immediately
	--		returned:
	--
	erl2fbl :: NameBinds -> [EvalRes] ->   [AldorFBlock]   -> FBList_Res
	erl2fbl    binds  []		curr  =    OK_FBL  binds curr
	erl2fbl    binds  (er:ers)  curr  =
		case er of
			Bottom_ER  _     -> Bottom_FBL binds
			other			 -> erl2fbl    binds ers (curr++[er2fb other])
			--	XXX: BEWARE of the order here:  use (++), not (:)!


---------------------
--	"evalLogList": --
---------------------
--	The common implementation of "AndList" and "OrList" evaluation.
--		Evaluation is still STRICT:  all operands are evaluated  before the
--		logical expr itself can be evaluated  (otherwise it would be diffi-
--		cult to normalise logical exprs with partially-evaluated args).  If
--		any of the args evaluates to Bottom, the result is always Bottom.
--	NB: although "evalAnd" and "evalOr" are perfectly defined  on empty arg
--		lists, we disallow that case  (in fact, it should already be caught
--		during the FBlocks decoration).
--	Evaluation of Boolean expressions is a partial "reduction point":  non-
--		Boolean vals are thrown away (but multiple Boolean results are OK).
--	This function does NPT decrement "_remEvalDepth" in "ee0":
--
data   LogOp = And | Or   deriving Eq

evalLogList :: ETC_Env -> [AldorFBlock] -> LogOp -> [EvalRes]

evalLogList  ee0 args log_op  = ress
	where
	--	Evaluate each of "args" and cross the results over. In "e_args",
	--		the outer dim is versions, the inner dim is args:
	--	NB: "args"    is assumed to be always of length >= 2:
	--
	e_args :: [EvalRes_CL]
	e_args  = crossWithAttrs   [evalFB ee0 arg | arg <- args]

	--	Evaluate each version of "e_args":
	--
	ress   :: [EvalRes]
	ress    = mapMaybe    evalLogVector  e_args

	evalLogVector:: EvalRes_CL -> Maybe EvalRes

	evalLogVector   ercl

		| length b_vec /= length e_vec = Nothing
			--
			-- Skip this "vec": it contains fully-evaluated non-Boolean values

		| any isBottom vec			   = Just (Bottom_ER binds)
			-- Bottom encountered

		| log_op == And && (not b_res) = Just (Full_ER binds (BooleanVal False))
			-- Certainly "False" according to fully-evaluated parts

		| log_op == Or  &&  b_res      = Just (Full_ER binds (BooleanVal True ))
			-- Certainly "True"  according to fully-evaluated parts

		| otherwise					   = Just (Partial_ER     binds   u_res)
			-- Re-construct a normalised expr from partially-evaluated parts
			-- XXX UNDECOR

		where
		vec   :: [EvalRes]
		vec   =  getCList ercl

		binds :: NameBinds
		binds =  getAttr  ercl

		--	Get fully-evaluated components of "vec":
		--
		e_vec :: [AldorValue]
		e_vec =  [val | Full_ER _ val <- getCList ercl]

		--	Restrict these components further to Booleans only:
		--
		b_vec :: [Bool]
		b_vec =  [bool | BooleanVal bool <- e_vec]

		--	If "e_vec" and "b_vec" are of different lengths,      the former
		--		contains some fully-evaluated elemnts which are nevertheless
		--		not Booleans; then "vec" is not valid at all,  and "curr" is
		--		unmodified (see above).
		--	Otherwise, apply the corresp "log_op" on "b_vec".  Empty "b_vec"
		--		is allowed:
		--
		b_res :: Bool
		b_res =  case log_op of
			And -> foldl (&&) True  b_vec
			Or  -> foldl (||) False b_vec

		--	Depending on "b_res",   we may or may not  need to re-construct the
		--		unevaluated parts: "u_vec" is actually computed only if "b_res"
		--		is not sufficient to decide the result.  Note that both "b_vec"
		--		and "u_bec" cannot be empty unless there was also a Bottom:
		--
		u_vec :: [AldorFBlock]
		u_vec =  [fb  | Partial_ER _ fb <- vec]

		u_res :: AldorFBlock
		u_res =  case log_op of
			And -> simpleFB (AndList u_vec)
			Or  -> simpleFB (OrList  u_vec)


------------------
--	"evalTarg": --
------------------
--	Evaluate a "BindTarget" for a given "Identifier". The latter needs to be
--		supplied explicitly,  since it will be required construct the result
--		of a partial evaluation.
--	NB: when this function is called, the bindings for the Identifier  given
--		are ALREADY restricted to the given BindTarget in "eeT":

evalTarg  :: ETC_Env-> AldorExpr->  BindTarget -> [EvalRes]

--
--	(1) "DefTarg" with (MUST BE) "GenFDTE":
--		Evaluate the Def Body.   There is a SPECIAL CASE here: For a Function
--		Definition (which is short-cut for a Lambda), we need to re-construct
--		the "LambdaVal" properly, so it could be used directly in a "Function
--		Call". Also, if the DefBody is Void (as for DECLARATIONS),  a Partial
--		result is produced:
--
evalTarg eeT  ident@(Identifier _ _)
					(DefTarg (GenFDTE _ _ type_expr body_fb)) =
	case body_fb of
		VoidFB _ ->
			--	No body:
			--
			[Partial_ER (getBinds eeT) (simpleFB ident)]

		_ ->
			--	The body exists:
			--
			case type_expr of
				Literal (TypeVal proto@(FunctionT _ _)) ->
					--
				 	-- It's a function def, so re-construct  and evaluate the
					--		Lambda (as it may need further normalisation):
					--
					evalExpr eeT (Literal (LambdaVal proto body_fb))

			 	_ ->
					--	Generic case: evaluate the body	(even if the Ident is
					--	"%" -- the target is a DefTarg, not a "PerCentTarg"):
					evalFB eeT body_fb

--
--	(2) "FP_Targ" with (MUST BE) "FParmFB":
--		We extract a pre-evaluated arg from the EvalStack, NOT from the DTE
--		(as it has VoidFB body!)
--	NB:
--	(*) It's not enough to scan just the stack top, -- the Identifier in
--		question can be bound to a deeper-laying arg;
--	(*) Use "foldr" rather than "foldl" for traversing the stack frames,
--		then the resulting "arg_val" lists  come in the right order:
--
evalTarg  eeT  ident@(Identifier use_no _) (FP_Targ fb@(FParmFB _ _ _)) =

	case lookupES_EE eeT def_no of

		--	Found, returning the result (which may itself be anything -- Full,
		--		Partial or Bottom!):
		--
		Just res -> [res]

		--	Not found: but maybe the required value exists in extra bindings
		--		("CondTarg"s):
		--
		Nothing  ->
			case getCondTarg (getBinds eeT) use_no of
				Just ct_fb ->
					evalFB eeT ct_fb

				Nothing    ->
					--	The search completes with no results: partial evaluation
					--		is to be performed, so return the wrapped Identifier
					--		itself. HERE we retain its "use_no" as it will still
					--		point to an FParmFB   (no matter that the latter may
					--		be normalised,- we are not using its contents during
					--		evaluation, only the fact that it's an FParmFB):
					--
					[Partial_ER (getBinds eeT) (simpleFB ident)] -- XXX UNDECOR
	where
	GenFDTE def_no _ _ _ = getFParm fb

--
--	(3) "PerCentTarg":
--		Returns unevaluated "%" Identifier:
--
evalTarg   eeT  ident@(Identifier _ name)   targ@(PerCentTarg  _)

	| name == ["%"] = [Partial_ER (getBinds eeT) (simpleFB ident)]
	| otherwise     = error2 "evalTarg"
							 ((dotcat name)++" bound to "++(show targ))

--
--	(4) "BI_Targ":
--		Construct an (always unique) "BuiltInOp" callable object:
--
evalTarg     eeT   (Identifier _ _)  (BI_Targ bi_no)  =
	[Full_ER (getBinds eeT) opval]
	where
	(opname, bi_proto,    bi_body)  = biInfo    bi_no
	opval  = BuiltInOp      opname    bi_proto  bi_body 

--
--	(5) "RepTarg":
--		Evaluate the rep body and return the result and all updated binds. The
--		rep body is not a def body, so use just "evalFB", not "getDefVal"s:
--
evalTarg    eeT    (Identifier _ _)  (RepTarg rep_fb)  = evalFB  eeT rep_fb

--
--	(6) "DynTarg":
--		Find its ULTIMATE non-dynamic targs, and evaluate all of them:
--
evalTarg    eeT    ident@ (Identifier use_no _) dtarg@(DynTarg _ _) = ress
	where
	odts   :: [OpenedDynTarg]
	odts'  =  openDynTarget   eeT use_no dtarg
	odts   =  -- trace ("\nOpenedDynTargs1 for "++(show ident)++":\n\n"++
			  -- (show [otarg | OK_ODT _ otarg <- odts']))
			  odts'

	--	NB: Binds returned by "openDynTarget" are propagated:
	ress   :: [EvalRes]
	ress'  =  concat [evalODT odt | odt <- odts]
	ress   =  -- trace ("OpenedDynTargs2 for "++(show ident)++":\n\n"++
			  -- (show ress')) 
			  ress'

	evalODT:: OpenedDynTarg  -> [EvalRes]
	evalODT   (Bottom_ODT  ubinds)		  =  [Bottom_ER ubinds]
	evalODT   (OK_ODT      ubinds  otarg) =
		evalTarg (setBinds ubinds  eeT  ) ident otarg
		--	NB: If "otarg" has no body, this will generate the "ident" again!

--
--	(7) "CondTarg":
--		Should not occur on its own:
--
evalTarg    _      (Identifier _ _)  ct@(CondTarg _) =
	error2 "evalTarg" ("Stand-alone "++(show ct)++" encountered")

--
--	(8) "AddedTarg":
--		Get its contents, BUT preserve the "AddedTarg" in the final NameBinds,
--		otherwise it would be silently eliminated!
--
evalTarg eeT ident@(Identifier use_no _) added@(AddedTarg t) =
	map restoreAdded (evalTarg eeT ident t)
	where
	restoreAdded :: EvalRes -> EvalRes
	restoreAdded    er      =
		setAttr (modifyBinds (getAttr er) use_no [added]) er

--
--	(9) All other cases: ERROR:
--
evalTarg   _ expr targ =
	error2 "evalTarg" ("Invalid call on "++(show expr)++",\n"++(show targ))


-----------------------
--	"evalTypeExpr":  --
-----------------------
--	Similar to "evalExpr", but uses caching, since each Type Expr is a part
--		of a DTE, so the corresp DefNo serves as a caching key:
--
evalTypeExpr :: ETC_Env -> DefNo -> AldorExpr -> [EvalRes]
evalTypeExpr    ee         def_no   type_expr
	--	If caching is OFF, or "def_no" is "noNUMBER"  so the eval result for
	--		this "type_expr" cannot be cached: evaluate "type_expr" directly.
	--		The same happens if there is a non-[] stack:
	--
	| not    (_useCaching ee) || def_no == noNUMBER ||
	  length (_evalStack  ee) /= 0  =
		evalExpr ee type_expr

	| otherwise =
		--	Do cache look-up:
		case fromCache dkey of
			Nothing ->
				--	Not found: evaluate the expr now and store the results back
				--		in the cache:
				toCache dkey (evalExpr ee type_expr)
			Just rs ->
				--	Cached result found!
				rs
	where
	dkey    :: EvalCacheKey
	dkey    =  DefKey def_no (getBinds ee)


-- =========== --
--	"error2":  --
-- =========== --
error2 ::String -> String -> a
error2   func_name msg    =  errorM modName func_name msg

