-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--                               "NameBinds.hs":                             --
--	    Construction and Management of Name Bindings for Aldor FBlocks		 --
--                          L.A.Timochouk, 2000--2002	                     --
-- ========================================================================= --
module NameBinds
	   (
			BindTarget(..),
			NameBinds,

			mkBinds,         lookupBinds,   sizeBinds,		 modifyBinds,
			appendBinds,     foldBinds,		mkAddedTarg,
			TargFBE,		 getTargFBE,

			getEnumsUBounds, mkCondTargs,   getCondTarg,
			WithBinds (..)
	   )
where

import DataTypes
import FButils
import PrPrint
import BuiltIns
import ParseAST
import DecorFB

import Maybe
import List
import Data.FiniteMap
import IOExts

modName :: String
modName = "NameBinds"

--	XXX: this looks very poor:
bigUseNo_OffSet :: UseNo
bigUseNo_OffSet =  1048576


-- =============== --
--	Name Bindings: --
-- =============== --
--	In the presence of names overloading, a UseNo for a name can potentially be
--		bound to several Targets  (but they  may be  reduced to just one during
--		the type-checking). The Target may be:
--	(*) a user-specified definition (DefTarg),  in that case its  DefTableEntry
--		(renaming resolved to a GenDTE, never a RenDTE) is specified; also spe-
--		cified is the enclosing FBlock  (to provide some context if necessary);
--	(*) a particular case of the above: a Function parm or  Record/Union field,
--		ie a Def found in an FParmFB; we store the whole FParmFB;
--	(*) a built-in function, in that case its BINo is given;
--  (*) a dynamic binding: contains a relative QName to be resolved at run-time
--		inside the (ADT- or Category-valued) FBlock; if it has a type-constrain-
--		ing expr, it is stored as well for run-time use. The corresp binding is
--		to be resolved after these entities are fully evaluated;
--	(*) a reference to ADTs representation clause; contains the FBlock with that
--		repr:
--
data    BindTarget  = DefTarg  FlatDTE
					  -- User-specified definition, the most common case

					| FP_Targ  AldorFBlock
					  -- A special case of the above: an "FParmFB"

					| PerCentTarg AldorExpr
					  -- Target for the "%" identifier: ADT- or Category-valued
					  --	expr, normally NOT to be evaluated,  as "%"s inside
					  --	it are bound back to it

					| BI_Targ  BINo  -- Built-in function

					| DynTarg  QName BindTarget
					  -- For names defined inside unevaluated ADTs  or Cats:
					  -- QName (the part yet to be resilved) and its Anchor

					| RepTarg  AldorFBlock
					  -- FBlock which is an ADT's representation type

					| CondTarg AldorFBlock
					  -- "Synthetic def" propagated from a Conditional down the
					  --  branch(es) in some cases: value only  (for the Ident
					  --  bound to "CondTarg"), no type

					| AddedTarg BindTarget
					  -- Typicaly used for dynamically-added bindings for "%"
					  -- Idents corresponding to Category instances; this ctor
					  -- encapsulates the actual binding.  "AddedTarg"s are to
					  -- be MERGED, not intersected, while intersecting binds,
					  -- as they would otherwise be elimited prematurely


--	The "NameBinds" type consists of a map UseNo => [BindTarget] and an Enums3
--		vector containing upper bounds (not necesserily the least) INCREMENTED
--		BY 1 for FBNos, DefNos and UseNos occurring in the FBlock (or collectn
--		of FBlocks) for which these NameBinds are constructed:
--
data    NameBinds   = NameBinds (FiniteMap UseNo [BindTarget]) Enums3

----------------------------------------------
--	Equality and Ordering of "BindTarget"s: --
----------------------------------------------
--	In most cases it's easy to decide, since all Defs and FBlocks are numbered
--		through:
--
instance Eq BindTarget
	where
	--
	--	"DefTarg"s:
	--
	(==) (DefTarg lfdte) (DefTarg rfdte) = ckDefNo ldef_no == ckDefNo rdef_no
		where
		ldef_no = getGenDefNo     lfdte
		rdef_no = getGenDefNo     rfdte

	--
	--	"FP_Targ"s:
	--	For the reliability, we check not only "FBNo"s but also inner "DefNo"s:
	--
	(==) (FP_Targ lfb@(FParmFB lfb_no _ _)) (FP_Targ rfb@(FParmFB rfb_no _ _)) =

		if  ckFBNo lfb_no   == ckFBNo rfb_no
		then
			if   ckDefNo ldef_no == ckDefNo rdef_no
			then True
			else error2 "(==)--FP_Targ" "Integrity error"
		else
			False
		where
		GenFDTE ldef_no _ _ _ = getFParm lfb
		GenFDTE rdef_no _ _ _ = getFParm rfb

	(==) (FP_Targ hmm1) (FP_Targ hmm2)	=
		 error2 "(==)"
			   ("Not an FParmFB in FP_Targ: "++(show hmm1)++",\n"++(show hmm2))

	--
	--	"PerCentTarg":
	--	All ADT- and Category-valued exprs contain "DefNo"s which we compare:
	--
	(==) (PerCentTarg expr1) (PerCentTarg expr2) =
		 (ckDefNo (getPerCentDefNo expr1)) ==
		 (ckDefNo (getPerCentDefNo expr2))

	--
	--	"BI_Targ"s:
	--
	(==) (BI_Targ lbi_no) (BI_Targ rbi_no) = ckBINo lbi_no == ckBINo  rbi_no


	--	"DynTarg"s:
	--	Compared by "FBNo"s of their FBs; even if the latter are "VoidFB"s,
	--		they still have valid positional "FBNo"s. Type constraint exprs
	--		are ignored (they are assumed to be the same if QNames and FBs
	--		are the same):
	(==) (DynTarg lqname ltarg) (DynTarg rqname rtarg)  =
		 (lqname == rqname)  && (ltarg == rtarg)

	--
	--	"Rep_Targ"s:
	--
	(==) (RepTarg lfb)  (RepTarg  rfb) =
		 ckFBNo (getFBNo lfb) == ckFBNo (getFBNo rfb)

	--
	--  "CondTarg"s:
	--
	(==) (CondTarg lfb) (CondTarg rfb) =
		 ckFBNo (getFBNo lfb) == ckFBNo (getFBNo rfb)

	--
	--	"AddedTarg"s:
	--
	(==) (AddedTarg lt) (AddedTarg rt) = lt == rt

	--
	--	Mixed Targs: Always False:
	(==) _ _ = False


instance Ord BindTarget where
	--	Comparison of Targets:
	--	DefTarg  < FP_Targ < PerCentTarg < BI_Targ < DynTarg < RepTarg <
	--	CondTarg < AddedTarg
	--
	--	Inside each group, they are compared by embedded enumerators;
	--		"noNUMBER"s are NOT allowed:
	--
	--	(1) LHS is "DefTarg":
	--
	compare (DefTarg fdte1) (DefTarg fdte2) =
		compare (ckDefNo def_no1) (ckDefNo def_no2)
		where
		def_no1 = getGenDefNo fdte1
		def_no2 = getGenDefNo fdte2

	compare (DefTarg _) _ = LT

	--
	--	(2) LHS is "FP_Targ":
	--
	compare (FP_Targ _)   (DefTarg   _) = GT
	compare (FP_Targ fb1) (FP_Targ fb2) = compare (getFBNo fb1) (getFBNo fb2)
	compare (FP_Targ _)   _				= LT

	--
	--	(3) LHS is "PerCentTarg":
	--
	compare (PerCentTarg _)  (DefTarg  _)	  = GT
	compare (PerCentTarg _)  (FP_Targ  _)	  = GT
	compare (PerCentTarg e1) (PerCentTarg e2) = 
		compare (ckDefNo (getPerCentDefNo e1))
				(ckDefNo (getPerCentDefNo e2))

	compare (PerCentTarg _)  _				  = LT

	--
	--	(4) LHS is "BI_Targ":
	--
	compare (BI_Targ _)   (DefTarg   _)   = GT
	compare (BI_Targ _)   (FP_Targ   _)   = GT
	compare (BI_Targ _)   (PerCentTarg _) = GT
	compare (BI_Targ bn1) (BI_Targ bn2)   = compare bn1 bn2
	compare (BI_Targ _)   _				  = LT

	--
	--	(5) LHS is "DynTarg":
	--
	compare (DynTarg _ _) (DefTarg _)     = GT
	compare (DynTarg _ _) (FP_Targ _)     = GT
	compare (DynTarg _ _) (PerCentTarg _) = GT
	compare (DynTarg _ _) (BI_Targ _)     = GT

	compare (DynTarg q1 t1) (DynTarg q2 t2) =
		case compare t1 t2 of
			EQ    -> compare q1 q2
			other -> other

	compare (DynTarg _ _) _				  = LT

	--
	--	(6) LHS is "RepTarg":
	--
	compare (RepTarg _)     (DefTarg _)		= GT
	compare (RepTarg _)     (FP_Targ _)		= GT
	compare (RepTarg _)		(PerCentTarg _) = GT
	compare (RepTarg _)		(BI_Targ _)		= GT
	compare (RepTarg _)	    (DynTarg _ _)   = GT
	compare (RepTarg fb1)   (RepTarg   fb2) =
		compare (getFBNo fb1) (getFBNo fb2)
	compare (RepTarg _)     _				= LT

	--
	--	(7) LHS is "CondTarg":
	--
	compare (CondTarg _)    (DefTarg _)		= GT
	compare (CondTarg _)    (FP_Targ _)		= GT
	compare (CondTarg _)	(PerCentTarg _) = GT
	compare (CondTarg _)    (BI_Targ _)		= GT
	compare (CondTarg _)    (DynTarg _ _)   = GT
	compare (CondTarg _)    (RepTarg _)		= GT
	compare (CondTarg fb1)  (CondTarg  fb2) =
		compare (getFBNo fb1) (getFBNo fb2)
	compare (CondTarg _)    _				= LT

	--
	--	(8) LHS is "AddedTarg":
	--
	compare (AddedTarg _)   (DefTarg _)		= GT
	compare (AddedTarg _)   (FP_Targ _)		= GT
	compare (AddedTarg _)	(PerCentTarg _) = GT
	compare (AddedTarg _)   (BI_Targ _)		= GT
	compare (AddedTarg _)   (DynTarg _ _)   = GT
	compare (AddedTarg _)   (RepTarg _)		= GT
	compare (AddedTarg _)   (CondTarg _)	= GT
	compare (AddedTarg t1)  (AddedTarg t2)  = compare t1 t2


--------------------------------------
--	"Eq" and "Ord" on "NameBinds":  --
--------------------------------------
--	To decide that, we unload their contents into a list:
--
instance Eq  NameBinds where
	(==)    (NameBinds fm1 _) (NameBinds fm2 _) = (==) fm1 fm2
	--	"Eq" on "FiniteMap" is already defined

instance Ord NameBinds where
	--	Hmm... This is more complex. We first need to equalise domains of
	--		"fm1" and "fm2",    then compare lists of (key, value)  pairs
	--		lexicographically:
	--
	compare (NameBinds fm1  _) (NameBinds fm2 _) =
		compare (fmToList fm1) (fmToList  fm2)


-- ============ --
--	"mkBinds":  --
-- ============ --
--	Constructs all possible bindings for all names used within a given FBlock.
--		The FBlock must be decorated. The function requires a vector of total
--		"Enums3" counts (returned by "decorFB") as an additional argument, as
--		this vector is stored inside NameBinds constructed.
--	We need to extract UseNos from a given Expr ("Identifier" only), and call
--		"findDefs" for each UseNo encountered.  We also need to record the FB
--		in which the UseNo in question is located   (defs for that UseNo will
--		be searched from that FB up). This is done by using "foldFB_Gen" with
--		actions on Exprs only:
--
emptyBinds ::  NameBinds
emptyBinds  =  NameBinds  emptyFM   (0,0,0)

mkBinds    ::  AldorFBlock -> Enums3     -> NameBinds
mkBinds		   fb			  enums_count =
	NameBinds  new_fm		  enums_count
	where
	new_binds :: NameBinds
	new_binds =  snd (foldFB_Gen passFB scanUseNos (undef, emptyBinds) fb)
	NameBinds new_fm _ = new_binds


---------------------
--	"scanUseNos":  --
---------------------
scanUseNos ::  AldorFBlock -> AldorExpr -> NameBinds  -> NameBinds

scanUseNos     encl_fb ident@(Identifier   use_no qname)    binds0
	| use_no == noNUMBER =
		error2 "scanUseNos" ("UnDecorated Identifier: "++(show ident))

	| qname  == ["%"]    =
	  --  Special case:
	  --  No overloading, renaming, ..., no search in DefTables at all!
	  findPerCentTarg  binds0 use_no encl_fb

	| otherwise	= binds3
	where
	--	First find all user-specified definitions for this "qname", or possibly
	--		a special definition (e.g. for "Rep"):
	--
	info1 :: FindInfo
	info1 =  findDefsR FindInfo
							{_use_nos = [use_no],
							 _name    = qname,
							 _fromFB  = encl_fb,
							 _mode    = initMode qname,
							 _binds   = binds0}

	--	Then see whether the "qname" also corresponds to a built-in function:
	--	NB: If it does not, the "targets" list below will be [], so  "install
	--		Binds" will simply return a FindInfo with original "binds1"
	--
	info2 :: FindInfo
	info2 =  if length qname == 1
			 then
				installBinds info1 [BI_Targ bi_no | bi_no <-biBinds(head qname)]
			 else
				--	A fully-qalified name can never be that of a Built-In:
				info1

	binds2:: NameBinds
	binds2 = _binds info2

	--	Now see whether the name in question has actually been resolved, and
	--		produce an error if not.
	--	The "%" identifier is a special case; it is bound to the DefNo of the
	--		inner-most Category- or ADT-valued object,
	--
	binds3:: NameBinds
	binds3 = case lookupBinds binds2 use_no of
		[]-> error2 "scanUseNos" ("Failed to resolve "++(show ident))
		_ -> binds2

scanUseNos _ _ curr_binds = curr_binds


--------------------------
--	"findPerCentTarg":  --
--------------------------
--	Search for the inner-most ADT- or Category-valued expr:
--
findPerCentTarg :: NameBinds -> UseNo -> AldorFBlock -> NameBinds

findPerCentTarg    binds use_no fb =
	case fb of
		--
		--	"Found" cases:
		--
		SimpleFB  _ expr@(Literal (TypeVal (ADT _ _ _)))     _ ->
			append_pc_targ expr

		SimpleFB  _ expr@(Literal (TypeVal (CategoryT _ _))) _ ->
			append_pc_targ expr

		SimpleFB  _ expr@(Add1    _ _) _ ->
			append_pc_targ expr

		SimpleFB  _ expr@(Add2  _ _ _) _ ->
			append_pc_targ expr

		SimpleFB  _ expr@(With1   _ _) _ ->
			append_pc_targ expr

		SimpleFB  _ expr@(With2 _ _ _) _ ->
			append_pc_targ expr

		--
		--	"Go up and try again" cases:
		--
		SimpleFB  _ _   uplink -> findPerCentTarg binds use_no uplink
		NestedFB  _ _ _ uplink -> findPerCentTarg binds use_no uplink
		FParmFB   _ _   uplink -> findPerCentTarg binds use_no uplink

		--
		--	"Not found" case: hitting the ceiling:
		--
		VoidFB _ -> error2 "findPerCentTarg"
						  ("Failed to resolve "++(show use_no))
	where
	append_pc_targ :: AldorExpr-> NameBinds
	append_pc_targ    expr =	  modifyBinds binds use_no [PerCentTarg expr]


-------------------
--	"findDefsR": --
-------------------
--	Search context to be carried along by "findDefs*" is quite extensive, so a
--		record type is provided for it:
--
data FindInfo  	= FindInfo
	{
		--	Renaming chain: THE MOST "ADVANCED" use_no IS AT THE HEAD (at 0):
		--		use_nos[1] points to (via renames, ==) use_nos[0],
		--		use_nos[2] points to (via renames, ==) use_nos[1],  ...,
		--		the LAST use_no in the chain corresponds to the ORIGINAL Ident
		--		being resolved, before all renamins were unfolded.
		--	The renaming chain is used to prevent  cyclic renamings:
		--
		_use_nos :: [UseNo],

		--	The name or a trailing part of the name corresp to the head use_no
		--		in the above chain. It it is a trailing part (of a fully-quali-
		--		fied name), the leading part of it is already resolved:
		--
		_name    :: QName,

		--	The FBlock where the head use_no (and its corresp name) were found:
		--		the current resolution process starts here and follows uplinks:
		--
		_fromFB  :: AldorFBlock,

		--	The resolution mode for the current search (see "FindMode" below):
		--
		_mode	 :: FindMode,

		--	The Name Bindings: modifying them is the whole objective of the
		--		resolving process. When the target becomes known, the binds
		--		are modified so that all use_nos in the renaming chain  are
		--		bound to that target:
		--
		_binds   :: NameBinds
	}

--
--	"FindMode": There are currently 3 name resolution modes:
--
data FindMode   = 
	--
	--	(1)	Resolution of an original UNQUALIFIED name: it is searched directly
	--		in DefTables and also in all explicitly imported domains (this means
	--		that the imported domains  themselves  are not searched recursively
	--		inside each other,  so if nested domains are to be imported,  their
	--		names must be fully-qualified):
	Mode1

	--	(2)	This mode is invoked  from within Mode-1 search  when an originally
	--		unqualified name is searched within a given imported domain  (which
	--		may itself be given by an unqualified or fully-qualified name). The
	--		name is searched for its explicit occurrences only (i.e. not inside
	--		any other imported domains), and no dynamic bindings  are installed
	--		if only some leading part of the name is found:
   | Mode2

	--	(3) Resolution of an originally FULLY-QUALIFIED name. Imported  domains
	--		are not followed,  but if only a leading name component is found, a
	--		dynamic binding is produced (as the domain where this name is to be
	--		found might not be fully evaluated at the name resolution time).
	--		Thus, to allow a name  from a dynamically-generated domain to be
	--		used, we must fully qualify its Idents:
   | Mode3
   deriving (Eq)


--
--	Now the "findDefsR" itself:
--	The aim is to resolve the leading UseNo in the chain (even if it is already
--		resolved in "info", new bindings for it can be found):
--
findDefsR:: FindInfo -> FindInfo

findDefsR   info =
	case fromFB of
		--	(1):
		--	"NestedFB" with its DefTable:
		--	Scan the DefTable and then go up;  "fromFB" is the enclosing
		--		FBlock for this "dt":
		--
		NestedFB _ dt _ uplink ->
			let
				--	It may or may not be found in this "dt":
				info' :: FindInfo
				info'  = findInDT info dt
			in
			--	In any case, go up -- there could be more bindings there:
			findDefsR  info'{_fromFB = uplink}

		--	(2):
		--	"SimpleFB" encapsulating an ADT with its DefTable:
		--	Similar to "NestedFB". This case occurs when we search an ADT on
		--		the way UP, e.g.  if we have started from inside of the same
		--		ADT, as opposed to "descend", in which an ADT is searched on
		--		the way DOWN.
		--	Also, if the name to be resolved is "Rep",   we take the repr of
		--		this ADT and look no further:
		--
		SimpleFB _ (Literal (TypeVal (ADT _ rep_fb dt))) uplink ->
			if  qname == ["Rep"]
			then
				--	Special case: "Rep":
				installBinds info  [RepTarg rep_fb]
			else
				--	The generic case:
				let
					--  It may or may not be found in this "dt":
					info' :: FindInfo
					info' =  findInDT info dt

					--	It may also be found inside the "Rep" clause:
					info'':: FindInfo
					info''=  findInFB info' rep_fb
				in
					--	Still go up -- there could be more bindings there:
					findDefsR  info'' {_fromFB = uplink}

		--	(3):
		--	"SimpleFB" encapsulating a "CategoryT":
		--	Similar to "ADT" above, but it has no "Rep":
		--
		SimpleFB _ (Literal (TypeVal (CategoryT _ dt))) uplink ->
			let
				--  It may or may not be found in this "dt":
				info':: FindInfo
				info' = findInDT info dt
			in
				--  Still go up -- there could be more bindings there:
				findDefsR  info' {_fromFB = uplink}

		--	(4):
		--	All other "SimpleFB"s: nothing can be found at this level (there
		--		is no DT), so just follow the "uplink":
		--
		SimpleFB _ _ uplink	   -> findDefsR info {_fromFB = uplink}

		--	(5) "FParmFB":
		--	Scan the singltn DT  and stop recursion if the name in question
		--		is found there (this is an example of NAME HIDING in Aldor):
		--
		FParmFB _ fdt uplink ->
			let
				info'  :: FindInfo
				info'  =  findInDT info fdt

				use_no :: UseNo
				use_no =  head (_use_nos info)

				before,   after:: Int
				before =  length (lookupBinds (_binds info ) use_no)
				after  =  length (lookupBinds (_binds info') use_no)

				found  :: Bool
				found  =  after > before
			in
				if    found
				then  info'			  					  -- Stop here!
				else  findDefsR info' {_fromFB = uplink}  -- Go further up

		--  (6) "VoidFB": Top level reached -- all done:
		--
		VoidFB _   -> info
	where
	qname  :: QName
	qname  = _name    info

	fromFB :: AldorFBlock
	fromFB = _fromFB  info


------------------
--	"findInDT": --
------------------
--	This part of the resolution process is quite mode-dependent:
--
findInDT :: FindInfo -> DefTable -> FindInfo

findInDT    info0       dt		 =  info4
	where
	--
	--	(1) In any mode, we begin with looking-up the first name component
	--		("nameH") in the DefTable:
	--
	mode :: FindMode
	mode		  = _mode   info0

	qname:: QName
	qname		  = _name   info0

	--	NB: "qname" MUST NOT be []; by construction of QNames, this was a pre-
	--		condition before entering "findDefsR", and we will check that the
	--		tail is not [] before going into the "descend"  (the latter calls
	--		"findInDT" recursively):
	nameH:: String
	nameT:: QName
	nameH : nameT = qname

	dtes :: [DefTableEntry]
	dtes =  lookupDT dt nameH

	--	What happens then depends on the mode and on whether "dtes" is empty:
	--
	info2:: FindInfo
	info2 =
		if  length dtes == 0
		then
			--	No success here; as we are already inside a DefTable and new
			--		entries cannot be created at run time, dynamic binds are
			--		not to be installed here, even in Mode2:
			info0
		else
			--	"dtes" is non-empty:
			let
				--	Traverse "dtes", resolve possible renamings in them, and
				--		construct the list of def bodies; "info" can potenti-
				--		ally be modified in this process. The 2nd element of
				--		the tuple returned by "actualTargs" is  [BindTarget]
				--		after renamings are resolved:
				--
				(info1, targs) = foldl actualTargs (info0, []) dtes
			in
				if  length nameT == 0
				then
					--	If "nameT" is [], we are done: install all "targs":
					--
					installBinds info1 targs
				else
					--	Otherwise, we still need to resolve "nameT" recursively
					--		for all resolutions of "nameH".    The search is to
					--		be done inside the DefTables  (eg of ADTs and Cate-
					--		gories) associated with found "nameH" targets, NOT
					--		going up anymore.   If a targ body has no DefTable,
					--		a dynamic link is installed in Mode3    (but not in
					--		Mode2):
					foldl  descend (info1 {_name = nameT}) targs

	--
	--	(2) Also, in Mode1 only, search for this name in all explicitly impor-
	--		ted domains:
	--
	--		Get the list of all domains imported into the current DefTable,
	--		as Identifiers:
	--
	imports:: [AldorExpr]
	imports = getImports dt

	--	Resolve all these Identifiers (they may already be resolved, in that
	--		case "findDefsR" returns immediately). IMPORTANT:   Domain names
	--		themselves are considered to be always fully-qualified, ie  they
	--		are never searched inside other domains (even if they are 1-word
	--		names), ie we resolve them in Mode3:
	--
	info3  :: FindInfo
	info3   = if   mode == Mode1
			  then foldl resImport info2 imports
			  else info2

	resImport:: FindInfo -> AldorExpr -> FindInfo

	resImport   cinfo imp@(Identifier imp_use_no imp_qname)	= res3
		where
		--	Resolve the domain:
		--
		res1 =  findDefsR  cinfo {_use_nos = [imp_use_no],
								  _name    = imp_qname,
								  _mode    = Mode3}
		--
		--	Get the target(s) found,   and "descend" into its body searching
		--		for the original "qname". We also verify here that there are
		--		no unresolvable domain names:
		--
		imp_targs::  [BindTarget]
		imp_targs =  case lookupBinds (_binds res1)  imp_use_no of
			[]    -> error2 "resImport"
							("Domain "++(show imp) ++" cannot be resolved")
			targs'-> targs' 

		--	Now finally  "descend" into each of the "imp_targs":
		--	We start with "cinfo" (not "res1")  as it contains the original
		--		"use_nos" and "qname"), but modify the Mode and install new
		--		"binds"; "descend" can create "DynTarg"s if appropriate:
		--
		res2 =  foldl
				descend (cinfo {_mode = Mode2, _binds = _binds res1}) imp_targs

		--	Don't forget to restore the original "Mode" before returning:
		res3 =  res2 {_mode = mode}

	resImport   _ expr =
		error2 "resImport" ("Import is not an Identifier:\n"++ (show expr))

	--
	--	(3)	In addition, the name in question (if simple -- Mode1 only) can be
	--		bound to Parm or Field declarations INSIDE some type exprs of this
	--		DefTable   (not at the top level, but considered to be in the same
	--		scope), e.g. "F1"  and "F2" in
	--
	--		F (R: Record (F1: Type, F2: F1)): R.F1 == R.F2;
	--
	--		So go recursively through ALL type exprs, and also through all Def
	--			Bodies, since the Body of one definition can be referenced  in
	--			the type expr of another:
	--
	info4  :: FindInfo
	info4   = if   mode == Mode1
			  then foldDT  findInDTE info3 dt
			  else info3

	findInDTE :: FlatDTE  -> FindInfo -> FindInfo

	findInDTE   (GenFDTE _ _ type_expr body_fb)  info = info''
		where
		info', info'' :: FindInfo
		info'  = findInTypeExpr info   type_expr
		info'' = findInFB		info'  body_fb

	findInDTE _ info =  info


------------------
--	"findInFB": --
------------------
findInFB :: FindInfo -> AldorFBlock -> FindInfo

findInFB    info  fb =
	case fb of
		SimpleFB _ body_expr  _  -> findInTypeExpr info'  body_expr
		NestedFB _ _ inner_fb _  -> findInFB       info'  inner_fb
		FParmFB  _ dt _			 -> findInDT       info'  dt
		VoidFB   _				 -> info -- Nothing here!
	where
	--	Change "_fromFB" here, or we may not be able to distingusish between
	--		a DefTarg and an FP_Targ when that targ is found!
	info' :: FindInfo
	info' =  info {_fromFB = fb}


------------------------
--	"findInTypeExpr": --
------------------------
--	Search for "indirect" bindings of a given UseNo, i.e. definitions contained
--		within "complex" Type Exprs of an FParmFB:
--
findInTypeExpr ::    FindInfo -> AldorExpr -> FindInfo

findInTypeExpr       info     (Literal (TypeVal (FunctionT parms _))) =
	foldl scanFParm  info     parms

findInTypeExpr       info     (Literal (TypeVal (RecordT   fields ))) =
	foldl scanFParm  info     fields

findInTypeExpr       info     (Literal (TypeVal (UnionT    fields ))) =
	foldl scanFParm  info     fields

findInTypeExpr       info _ = info

--
--	"scanFParm":
--	A wrapper around "findInFParm" to make it suitable for "foldl":
--
scanFParm :: FindInfo -> AldorFBlock -> FindInfo

scanFParm    info  fb@(FParmFB _ fdt _) =  findInDT  info' fdt
	where
	--	Change "_fromFB" here, or we may not be able to distingusish between
	--		a DefTarg and an FP_Targ when that targ is found!
	info' :: FindInfo
	info' =  info {_fromFB = fb}

scanFParm    _     _ = error2 "scanFParm" "Not an FParmFB"


----------------------
--	"actualTargs":  --
----------------------
--	Invoked on DTEs (from "FindInDT" only) in order to convert them into proper
--		Targets, resolving renamings:
--
actualTargs :: (FindInfo, [BindTarget]) -> DefTableEntry ->
			   (FindInfo, [BindTarget])

--
--	"GenDTE"s are to be converted into "DefTarg"s or "FP_Targ"s, depending on
--		the enclosing FBlock.    Here we KNOW that the def_name of the GenDTE
--		found is determined by the "info":
--
actualTargs    (info, targs)  (GenDTE def_no type_expr body)  =
	case _fromFB info of
		fp@(FParmFB _ _ _) -> (info, (FP_Targ fp ) : targs)
		_				   -> (info, (DefTarg gen) : targs)
	where
	gen	  :: FlatDTE
	gen	  =  GenFDTE  def_no (head (_name info)) type_expr body


--
--	"RenDTE"s are to be resolved:
--
actualTargs    (info0,targs)  (RenDTE _ (Identifier ren_use_no ren_name)) =
	--
	--	It's a renaming Identifier, but its "ren_use_no" is to be followed ONLY
	--		if it was not already encountered in "use_nos", to prevent infinite
	--		recursion:
	if  not (ren_use_no `elem` use_nos)
	then
		(info2, r_targs ++ targs)
	else
		--	We have already seen that "ren_use_no"!
		--	Print the encountered cyclic chain of UseNos:
		--
		error2 "actualTargs"
			   ("Cyclic renaming encountered: UseNos loop: "++
			   (show (reverse (ren_use_no : use_nos))))
	where
		use_nos  = _use_nos info0

		--	Carry out a new search for "ren_use_no", but specify the complete
		--		"renaming chain" in order to prevent cyclic renamings. If the
		--		target(s) for "ren_use_no" are found, they will be  installed
		--		in "info1".
		--	NB: Possible mode change here compared to  "info0":
		--
		info1 = findDefsR
				info0
				{
					_use_nos = ren_use_no : use_nos,
					_name    = ren_name,
					_mode    = initMode  (ren_name)
				}

		--	Get the (newly-obtained) target list for  "ren_use_no".  It will
		--		serve as the  ACTUAL target list for the  ORIGINAL  "use_no"
		--		which appeared to be bound to a RenDTE.  This target list is
		--		returned to the caller ("findInDT") and installed into Binds
		--		for the original "use_no" there.
		--	If the target list is empty (renaming has not been resolved), it
		--		is an error:
		--
		r_binds  = _binds info1
		r_targs :: [BindTarget]
		r_targs  =
			case lookupBinds  (r_binds) ren_use_no of
				r_targs'@(_:_)-> r_targs'
				_			  -> error2 "actualTargs"
										("Failed to resolve UseNo."++
										(show ren_use_no))

		--	Construct "info2": just insert "r_binds" into "info0" (so HERE we
		--		permanently install binds for "ren_use_no" itself, originally
		--		coming from "Info1") and retain "info0"'s other fields, so eg
		--		mode change would not propagate	back:
		--
		info2   :: FindInfo
		info2    = info0 {_binds = r_binds}


actualTargs _  (RenDTE _ hmm) =
		error2 "actualTargs" ("Invalid target expr:\n" ++ (show hmm))


-------------------
--	"descend*":  --
-------------------
--	Search for "qname" (given via the "info") in the specified "BindTarget":
--
descend :: FindInfo -> BindTarget -> FindInfo

descend    info        targ		  =
	case getTargFBE targ of
		Just fbe ->
			descend3 info targ fbe
		Nothing  ->
			info

descend3:: FindInfo -> BindTarget -> TargFBE -> FindInfo

--	(1)
--	If there is a Body, but no Type-Constraining Expr:
--
descend3   info  targ (VoidExpr, body) =
	case   body  of
		--	(*)  ADT Literal in a SimpleFB: search its DT:
		--
		SimpleFB _ (Literal (TypeVal (ADT _ _ dt))) _     ->
			findInDT  info dt

		--	(*) Category Literal is SimpleFB:    as above:
		--
		SimpleFB _ (Literal (TypeVal (CategoryT _ dt))) _ ->
			findInDT  info dt

		--	(*) Any other Simple FB: install a dynamic binding (Mode3 only):
		--
		SimpleFB _ _ _ ->
			mkDynTarg info targ

		--	(*)	"NestedFB":
		--		Can it occur at all? We have no Packages, but can have "where"
		--		clauses. However, tha latter's DefTables are not used to reso-
		--		lve fully-qualified names -- we need to go further down:
		--
		NestedFB _ _ innerFB _	 ->
			descend3 info targ (VoidExpr, innerFB)

		--	(*) Other FBs ("FParmFB", "VoidFB"): Should not occur here at all:
		--
		other_fb ->
			error2 "descend" ("Improper FB No."++(show (getFBNo_U other_fb))++
							  "encountered")

--	(2)
--	If there is a Type-Constraining Expr, but no Body:
--
descend3   info  targ (type_constr, (VoidFB _)) =
	case type_constr of
		--	If the Type-Constraining Expr is a Category Literal, we go into it:
		--
		Literal (TypeVal (CategoryT _ dt)) ->
			findInDT info dt

		--	Can it also be an ADT? -- Probably not, but this case will not harm:
		--
		Literal (TypeVal (ADT _ _ dt)) ->
			findInDT info dt

		--	Any other exprs: assume them to be Category-valued (and so install
		--		a DynTarg), or simply ignore them?
		--	Category-valued exprs can only be Identifiers, With*, and Function
		--		Calls. For them, we produce a DynTarg:
		--
		Identifier   _ _   -> mkDynTarg  info  targ
		With1        _ _   -> mkDynTarg  info  targ
		With2        _ _ _ -> mkDynTarg  info  targ
		FunctionCall _ _   -> mkDynTarg  info  targ

		_ -> info -- Nothing at all here

--	(3)
--	We have BOTH a Type-Constraining Expr and a Body (i.e., both non-Void).
--		The problem is to decide whether the type constraint makes the name
--		visible in the body (if it exists there at all).  This is currently
--		done by first "descend"ing into the type constriant, and if it pro-
--		duced any new resolutions, "descend"ing into the body:
--
descend3  info  targ  (type_constr, body)
	| isTypeT    type_constr =
	  --  Special case: if the "type_constr" is just "Type", it's no constraint
	  --	at all:
	  descend3  info  targ  (VoidExpr, body)

	| length targsC > length targsI =
	  -- Type constraint satisfied (maybe in more than one way!), so make
	  -- 	targ(s) in the body:
	  descend3  info  targ  (VoidExpr, body)

	| otherwise =
	  -- NB: "targsI" is either a proper subset of "targsC", or they are equal,
	  --	as in this case. The constraint is NOT satisfied:
	  info
	where
	use_no :: UseNo
	use_no =  head (_use_nos info)

	infoC  :: FindInfo
	infoC  =  descend3 info targ (type_constr, voidFB)

	--	Binds for "use_no" BEFORE descending into the constarint:
	targsI :: [BindTarget]
	targsI =  lookupBinds  (_binds info)  use_no

	--	Binds for "use_no" AFTER  descending into the constraint:
	targsC :: [BindTarget]
	targsC =  lookupBinds  (_binds infoC) use_no

	isTypeT:: AldorExpr -> Bool
	isTypeT   (Literal (TypeVal TypeT)) = True
	isTypeT   _							= False


--------------------
--	"mkDynTarg":  --
--------------------
--	Installs a DynTarg in the "info" (Mode3 only), if the "anchor_targ" is
--		appropriate:
--
mkDynTarg:: FindInfo -> BindTarget-> FindInfo
mkDynTarg   info        anchor_targ  =
	if  (_mode info) == Mode3
	then
		let
			qname = _name info
		in
			case getTargFBE anchor_targ of
				Nothing ->
					--	This "anchor_targ" will NOT resolve successfully at
					--		run time -- drop it with a warning:
					--
					trace ("\n\"mkDynTarg\": Improper Anchor Targ: "++
						  (show anchor_targ)++"\n")   info
				_ ->
					--	"anchor_targ" seems to be OK:
					installBinds  info [DynTarg qname anchor_targ]
	else
		--	Not "Mode3" -- no "DynTarg"s at all:
		info


---------------------
--	"getTargFBE":  --
---------------------
--	Returns (ConstrainingTypeExpr, TargetBodyFB) if possible (either, but not
--		both, of the components can be Void), or Nothing if the Target is not
--		suitable at all. The ConstrainingTypeExpr can be later used in "desc-
--		end" as a Category interface to an ADT:
--
type TargFBE = (AldorExpr, AldorFBlock)

getTargFBE:: BindTarget -> Maybe TargFBE

--	"DefTarg": Normally yields (TypeExpr, Body):
--
getTargFBE	(DefTarg (GenFDTE _ _ type_expr body)) = Just (type_expr, body)

getTargFBE  (DefTarg (RenFDTE def_no _ _)) =
		error2 "getDefTargs" ("Got RenFDTE DefNo."++(show def_no))

--	"FP_Targ": Similar to "RefTarg", but the body is always Void:
--
getTargFBE  (FP_Targ  fp_fb)  = Just (type_expr, void_body)
	where
	GenFDTE _  _ type_expr void_body  = getFParm fp_fb

--	"RepTarg":  Yields (VoidExpr, RepFB) -- there is no type constraint here:
--
getTargFBE  (RepTarg  rep_fb) = Just (VoidExpr, rep_fb)

--	"CondTarg": Like "RepTarg":
--
getTargFBE  (CondTarg def_fb) = Just (VoidExpr, def_fb)

--	All others: ("PerCentTarg", "BI_Targ", DynTarg"): Nothing at all:
--
getTargFBE	_				  = Nothing


------------------
--	"initMode": --
------------------
--	The initial mode for a name resolution is Mode3 or Mode1, depending on
--		whether it is fully-qulified or not, resp (Mode2 is activated only
--		from within Mode1 when we search for a name within imported domains):
--
initMode :: QName    -> FindMode
initMode    qname    =
	case length qname of
		0 -> error2 "initMode" "Empty fully-qualified name!"
		1 -> Mode1
		_ -> Mode3


----------------------
--	"installBinds": --
----------------------
--	Attaches new entri(es) to the NameBinds map. Calls "attachTarget" for each
--		target specified in "info":
--
installBinds::  FindInfo -> [BindTarget] -> FindInfo

installBinds    info        targets		=  foldl attachTarget info targets
	where
	--	The most recent "use_no" actually resolved  is the HEAD of the sequence
	--		in "info". We install a binding for it; earlier (renamed) "use_no"s
	--		(if any) were satisfied when a sub-search was invoked from  "actual
	--		Targs" -- in that sub-search, THEY were at the head:
	--
	use_no :: UseNo
	use_no =  head (_use_nos info)

	qname  :: QName
	qname  = _name    info

	--	"attachTarget":
	--
	attachTarget :: FindInfo -> BindTarget -> FindInfo

	attachTarget    cinfo       target
		--	If this pair (UseNo, Target) is already in the NameBinds, just skip
		--		it without signalling an error (XXX can this condition occur at
		--		all?..)
		--	NB: HERE we need "Eq" on "BindTarget"s!
		--
		| target `elem` ctargs = cinfo

		--	Some overloading control is implemented here:
		--	E.g. the "%" symbols are never overloaded, so only the first (inner-
		--		most) binding for each of them is retained:
		--	XXX: this case cannot occur at all, since we make binds for "%" se-
		--		parately in a quite different manner ("findPerCentTarg"):
		--
		| length ctargs /= 0 && qname == ["%"] =
			trace "\nWARNING: Multiple targets for \"%\", discarded\n" cinfo

		--  In all other cases: Just install the binding:
		--
		| otherwise = new_info

		where
		binds :: NameBinds
		binds  = _binds  cinfo

		--	Currently installed targets for this "use_no" (if [],  the "use_no"
		--		has not been installed at all yet):
		--
		ctargs ::   [BindTarget]
		ctargs =    lookupBinds binds use_no

		--	"new_info": "cinfo" with the new bidings:
		--
		new_info :: FindInfo
		new_info =  cinfo {_binds  = modifyBinds binds use_no (target:ctargs)}


-- ========================================== --
--	"show" on "BindTarget"s and "NameBinds":  --
-- ========================================== --
instance Show  BindTarget where

	show (DefTarg  (GenFDTE def_no _ _ _)) = "Def#"++ (show def_no)

	show (DefTarg  hmm) =
		error2 "show--DefTarg"  ("Invalid FDTE: "  ++ (show hmm))

	show (FP_Targ  fb@(FParmFB  _ _ _)) = "FP#"    ++ (show def_no)
		where
		GenFDTE def_no _ _ _ =  getFParm fb

	show (FP_Targ  hmm) =
		error2 "show--FP_Targ"  ("Not an FParmFB: "++ (show hmm))

	show (PerCentTarg expr) = "Def%#"  ++(show (getPerCentDefNo expr))

	show (BI_Targ bi_no)	= "BI#"    ++(show bi_no )

	show (DynTarg qname anchor) = "Dyn(\""++ (dotcat qname) ++
								  "\", "  ++ (show  anchor) ++ ")"

	show (RepTarg  fb)		= "Rep{"  ++(show (ckFBNo (getFBNo fb)))++"}"

	show (CondTarg fb)		= "Cond{" ++(show (ckFBNo (getFBNo fb)))++"}"

	show (AddedTarg t)      = "Added("++(show t)++")"


instance Show  NameBinds  where
	show (NameBinds fm _) = "{\n"++(foldFM showNBE "" fm)++"}"
		where
		showNBE ::  UseNo -> [BindTarget]  -> String -> String
		showNBE     use_no   dtes   curr   =
			curr++ ("    @" ++ (show use_no))++
					closeLst (" -> ["++ (foldl showDNs "" dtes))
			where
			showDNs :: String  -> BindTarget -> String
			showDNs    curr'  targ = curr'++(show targ)++","

			closeLst:: String -> String
			closeLst   s
				| length s >= 1 && (last s)==',' = (init s)++"]\n"
				| otherwise						 = s       ++"]\n"
				-- The latter does not actually happen as we currently don't
				--		keep empty entries in NameBinds...


-- ================================= --
--	Other Functions on "NameBinds":  --
-- ================================= --
----------------------
--	"lookupBinds":  --
--	"getCondTarg":  --
--	  "sizeBinds":  --
--	"modifyBinds":  --
--	"appendBinds":  --
--	"mkAddedTarg":  --
----------------------
--	NB: "lookupBinds" returns [] if the key in question is not found in the
--		Bindings.    It does NOT return any "CondTarg"s; the latter must be
--		requested separately by "getCondTargs":
--
lookupBinds :: NameBinds  -> UseNo  -> [BindTarget]

lookupBinds   (NameBinds fm _) use_no =
	case (lookupFM fm  use_no) of
		Just targs -> filter (not . isCondTarg) targs
		_          -> []

--	For each UseNo, there can be no more than one "CondTarg". The following
--		function returns directly the FB content of CondTarg, if available:
--
getCondTarg :: NameBinds  -> UseNo -> Maybe AldorFBlock
getCondTarg   (NameBinds fm _)  use_no =
	case (lookupFM fm  use_no) of
		Just targs ->
			case filter isCondTarg targs of
				[CondTarg fb]  ->  Just fb
				[]			   ->  Nothing
				_			   ->  error2 "getCondTarg"
										  ("More than 1 CondTarg for UseNo."++
										  (show use_no))
		Nothing   ->   Nothing


sizeBinds   :: NameBinds  -> Int
sizeBinds     (NameBinds fm _) =  sizeFM fm


modifyBinds :: NameBinds  -> UseNo  -> [BindTarget]  -> NameBinds
modifyBinds   (NameBinds fm  enums3)   use_no    targs  =
	NameBinds (addToFM   fm  use_no    targs)    enums3
	--  The old bindings, if any, are thrown away


appendBinds :: NameBinds  -> UseNo  -> [BindTarget]  -> NameBinds
appendBinds        binds     use_no    targs		 =
	modifyBinds    binds     use_no    new_targs
	where
	--	The old bindings, if any, are merged with the new ones.
	--  Don't install the same targets again, or if there is already
	--		an "AddedTarg t", don't install "t":
	--
	curr :: [BindTarget]
	curr =  lookupBinds  binds use_no

	added:: [BindTarget]
	added=  [t | AddedTarg t <- curr]

	new_targs :: [BindTarget]
	new_targs =  curr ++ [t | t <- nub targs, notElem t curr, notElem t added]


--	Don't create nested "AddedTarg"s!
--	Strip them if they are already nested:
--
mkAddedTarg :: BindTarget ->  BindTarget
mkAddedTarg   (AddedTarg t')= mkAddedTarg t'
mkAddedTarg    t		    = AddedTarg t


--------------------
--	"foldBinds":  --
--------------------
--	The general iterator over NameBinds:
--
foldBinds :: (UseNo -> [BindTarget] -> a -> a) -> a -> NameBinds   -> a
foldBinds     action   init_a   (NameBinds  fm _)   =  foldFM action init_a fm


-------------------------
--	"getEnumsUBounds": --
-------------------------
--	See the definition of the NameBinds data type:
--
getEnumsUBounds :: NameBinds -> Enums3
getEnumsUBounds   (NameBinds  _ enums3) = enums3


-- ============================ --
--  "Consolidatable" instance:  --
-- ============================ --
--	XXX: For some reason, this would not type-check if we had no "Eq" on
--		 "BindTarget"s (but we need to have it for other reasons anyway):
--
instance Consolidatable NameBinds
	where
	consolidate  lbinds rbinds  = mbinds3
	--
	--  Constructs an INTERSECTION of two NameBinds. The operation fails and
	--		produces  Nothing if the intersection of bindings is found to be
	--		empty for some UseNo:
	--
		where
		NameBinds fmL enums3L   = lbinds
		NameBinds fmR enums3R	= rbinds

		binds1			 :: NameBinds
		mbinds2, mbinds3 :: Maybe NameBinds

		--	"binds1":  First of all,   install "AddedTarg"s from both sides --
		--		they are always to be retained, and an "AddedTarg t" overrides
		--		just "t" if it is to be installed later:
		--
		addedL, addedR, addedAll :: [(UseNo, BindTarget)]
		addedL   = getAddedTargs fmL
		addedR   = getAddedTargs fmR
		addedAll = union  addedL addedR

		getAddedTargs ::  (FiniteMap UseNo [BindTarget])-> [(UseNo,BindTarget)]
		getAddedTargs fm= concat  [[(use_no, at) | at@(AddedTarg _) <- ts] |
									(use_no, ts)  <- fmToList fm]

		binds1   = (foldl (\ cbinds (use_no, added_targ) ->
							appendBinds cbinds use_no [added_targ])
							emptyBinds  addedAll)

		--	"mbinds2": Now add intersection of the LHS with "rbinds".
		--	"appendInter" inserts, if possible, an entry corresponding to a LHS
		--		UseNo, into that intersection:
		--
		mbinds2 = foldFM appendInter (Just binds1) fmL

		appendInter:: UseNo -> [BindTarget]-> Maybe NameBinds-> Maybe NameBinds
		appendInter   _		   _			  Nothing        =  Nothing
		appendInter   use_no   ltargs		  (Just currb)   =
			case   lookupBinds rbinds use_no of
				[]    -> Nothing

				-- If the UseNo is in the RHS: intersect the target lists and
				--	  install the intersection under the corresp "use_no". If
				--	  the result is empty (so there was no "AddedTarg"s there
				--	  either), we get "Nothing":
				rtargs->
					let
						binds2 :: NameBinds
						binds2 =  appendBinds currb use_no
								  (intersect ltargs rtargs)
					in
						case lookupBinds binds2 use_no of
							[] -> -- Still nothing!
								Nothing
							_  ->
								Just binds2

		--	"mbinds3": "mbinds2" with Enums3 set to the corresp intersection
		--		values:
		mbinds3 = case mbinds2 of
			Nothing				   -> Nothing
			Just (NameBinds fm2 _) -> Just (NameBinds fm2
										   (mergeEnums enums3L enums3R))


---------------------
--	"mergeEnums":  --
---------------------
--  To be on safe side, we set all counter bounds to the MAX of their resp LHS
--		and RHS values, although for UseNos,  it would always be sufficient to
--		use the MIN:
mergeEnums :: Enums3 -> Enums3 -> Enums3
mergeEnums    enums3L   enums3R =
	(max fBoundL fBoundR, max dBoundL dBoundR, max uBoundL uBoundR)
	where
	(fBoundL, dBoundL, uBoundL) = enums3L
	(fBoundR, dBoundR, uBoundR) = enums3R


-- =================== --
--	"ORable" instance: --
-- =================== --
--	While "consolidate" intersects value lists for each key of two NameBinds,
--		(-||-) merges those lists.  Again, the MAXes of the resp enumerators
--		are installed in the result:
--
instance ORable NameBinds
	where
	(-||-) (NameBinds fmL enums3L) (NameBinds fmR enums3R) = bindsLR
		where
		--	Get the union set of keys. Normally this is not necessary since the
		--		domain of all NameBinds is the same:
		--
		keysLR :: [UseNo]
		keysLR  = union (keysFM fmL) (keysFM fmR)

		--	For each key, get the union of values:
		--
		pairsLR:: [(UseNo,  [BindTarget])]
		pairsLR = [(use_no, union (getVals fmL use_no) (getVals fmR use_no))
				  | use_no <- keysLR]

		--	Use low-level "lookupFM" here, NOT "lookupBinds", as the latter does
		--		not return "CondTarg"s:
		--
		getVals:: (FiniteMap UseNo [BindTarget])-> UseNo -> [BindTarget]
		getVals   fm         use_no =
			case  lookupFM    fm   use_no  of
				Just targs -> targs
				Nothing	   -> []

		bindsLR  :: NameBinds
		bindsLR  =  NameBinds (listToFM pairsLR) enums3LR

		enums3LR :: Enums3
		enums3LR =  mergeEnums enums3L enums3R


-- ================ --
--	"mkCondTargs":  --
-- ================ --
--	Examines the "cond_fb" and possibly installs extra "CondTarg"s  in the
--		"binds", also depending on the conditional's branch for which this
--		operation is requested ("True" -> "then", "False" -> "else").
--	New binds are installed ONLY if the Identifier in question is not  al-
--		ready bound to a CondTarg:
--
mkCondTargs :: NameBinds -> AldorFBlock -> Bool -> NameBinds

--
--	(1) "cond_fb" is actually an Identifier:
--		Its "True"/"False" values are then propagated to the resp branches:
--
mkCondTargs       binds    (SimpleFB _ (Identifier use_no _) _)  branch
	| ct_bound    binds    use_no  = binds
	| otherwise = appendBinds  binds use_no   [CondTarg val_fb]
	where
	val_fb:: AldorFBlock
	val_fb = SimpleFB (use_no + bigUseNo_OffSet)
					  (Literal (BooleanVal branch)) voidFB
	--	NB: We need to decorate the "SimpleFB" here by an FBNo, as the latter
	--		is used in "Eq" and "Ord" on Targets.   We use (XXX -- is it OK?)
	--		the UseNo of the Identifier + a large number...

--
--	(2) "cond_fb" is an equality of the form "Identifier == anything":
--		For the "then" branch only, the corresp binding is installed  for
--		that Identifier:
--
mkCondTargs	 binds  (SimpleFB _  (FunctionCall
								 (SimpleFB   _ (Identifier _ ["="]) _)
								[(SimpleFB   _ (Identifier luse_no _) _), rhs])
					_)	    True
	| ct_bound   binds   luse_no = binds
	| otherwise  =   appendBinds   binds luse_no [CondTarg rhs]

--
--	(3) "cond_fb" is a "NestedFB": process it recursively:
--
mkCondTargs		 binds  (NestedFB _ _ inner_fb _) branch =
	mkCondTargs  binds  inner_fb  branch

--
--	(4) All other cases: leave "binds" unmodified:
--
mkCondTargs		 binds  _ _  =  binds

--
--	"ct_bound":
--	Checks whether a given UseNo is already bound to a "CondTarg":
--
ct_bound :: NameBinds   -> UseNo    -> Bool
ct_bound    binds		   use_no   =  any  isCondTarg targs
	where
	targs:: [BindTarget]
	targs=  lookupBinds    binds   use_no

--
--	"isCondTarg":
--
isCondTarg:: BindTarget  -> Bool
isCondTarg   (CondTarg _)=  True
isCondTarg   _			 =  False
 
----------------------
--	"getGenDefNo":  --
----------------------
--	Returns a DefNo of a GenDTE:
--
getGenDefNo :: FlatDTE              -> DefNo
getGenDefNo   (GenFDTE def_no _ _ _) = def_no
getGenDefNo   hmm = error2 "getGenDefNo" ("Not a GenFDTE: "++(show hmm))


-- ============== --
--	"WithBinds":  --
-- ============== --
--	A class of any types containing "NameBinds" (as opposed to more general
--		"WithAttr" class):
--
class WithBinds a
	where
	getBinds :: a -> NameBinds
	setBinds :: NameBinds -> a -> a


-- =========== --
--  "error2":  --
-- =========== --
error2:: String-> String -> a
error2 = errorM   modName

