-- vim:ts=4
-- ========================================================================= --
--							    "TypeCheckN.hs":							 --
-- 			Type-Checking of Aldor-- FBlocks when No Target is given		 --
--						   L.A.Timochouk, 2001--2002						 --
-- ========================================================================= --
--	OVERVIEW:
--	Type-check the RHS (any AldorExpr or FBlock) against the LHS.   The LHS may
--		be absent, otherwise it must be one of the following valid type-forming
--		exprs:
--		(*) a  Type Literal;
--  	(*) an Identifier;
--  	(*) a  Function Call;
--  	(*) an If-Then-Else conditional.
--	Type-checking acts as a "knowledge-based system";  there are always some
--		well-typed exprs which cannot be effectively type-checked,  and will
--		thus be rejected. At the top level, the implementation is structured
--		wrt the LHS:
--		(*) "_N" functions: LHS is not given 				(THIS MODULE)
--		(*) "_T" functions: LHS is given as an AldorType; |
--		(*) "_E" functions: LHS is an AldorExpr;		  > (ELSEWHERE)
--		(*) "_R" functions: LHS is an AldorExpr or Bottom.|
--	During type-checking, the NameBinds are possibly reduced. For a definition
--		to type-check successfully, all overloaded names within the definition
--		must be resolved to exactly 1 binding.
--	In "_N" functions, the LHS (type-checking target) is absent altogether, and
--		the RHS CAN NOT be assumed to be already evaluated -- it can be any
--		Aldor expr!

module TypeCheckN
	   (
			typeCheckExpr_N,  typeCheckFB_N,  typeCheckDT, typeCheckAppl
	   )
where

import Maybe
import List

import DataTypes
import ParseAST
import FButils
import NameBinds
import AlphaDelta

--  We must know precisely where calls to the Evaluation Engine occur
--      (to prevent such calls if further evaluation is not allowed),
--      so use explicit qualification of Eval* functions:
import EvalLib
import qualified EvalX

import TypeCheckLib

import IOExts

modName :: String
modName = "TypeCheckN"


-- ==================== --
--	"typeCheckExpr_N":  --
-- ==================== --
--	Type-checks "AldorExpr"s for internal consistency only,  without checking
--		it against a target type. However, the target type can be inferred in
--		this process (if the "_inferTypes" flag is set in the env):

typeCheckExpr_N :: ETC_Env -> AldorExpr -> [TC_Res]

-----------------------
--	(N1) "VoidExpr": --
-----------------------
--	Always evaluates to "TrivVal" of type "TrivT":
--
typeCheckExpr_N  tcenv  VoidExpr =
	okWith tcenv TrivT

-----------------------
--	(N2) "Literal":  --
-----------------------
--
--	(a) "Simple" values (excl "TypeVal"s) which always succeed:
--
typeCheckExpr_N  tcenv  (Literal  TrivVal)			=
	okWith tcenv TrivT

typeCheckExpr_N  tcenv  (Literal (BooleanVal _))	=
	okWith tcenv BooleanT

typeCheckExpr_N  tcenv  (Literal (IntegerVal _))	=
	okWith tcenv IntegerT

typeCheckExpr_N  tcenv  (Literal (FloatVal   _))	=
	okWith tcenv FloatT

typeCheckExpr_N  tcenv  (Literal (StringVal  _))	=
	okWith tcenv StringT

typeCheckExpr_N  tcenv  (Literal (BuiltInOp _ bi_type _)) =
	okWith tcenv bi_type
	--	NB: checks on "BuiltInOp"s are always OK as these objects are INTERNAL

--
--	(b) "Simple" "TypeVal"s which always succeed as well:
--
typeCheckExpr_N  tcenv  (Literal (TypeVal TrivT))   =
	okWith tcenv TypeT

typeCheckExpr_N  tcenv  (Literal (TypeVal ExitT))   =
	okWith tcenv TypeT

typeCheckExpr_N  tcenv  (Literal (TypeVal TypeT))   =
	okWith tcenv TypeT

typeCheckExpr_N  tcenv  (Literal (TypeVal BooleanT))=
	okWith tcenv TypeT

typeCheckExpr_N  tcenv  (Literal (TypeVal IntegerT))=
	okWith tcenv TypeT

typeCheckExpr_N  tcenv  (Literal (TypeVal FloatT))  =
	okWith tcenv TypeT

typeCheckExpr_N  tcenv  (Literal (TypeVal StringT)) =
	okWith tcenv TypeT

typeCheckExpr_N  tcenv  (Literal (TypeVal (BuiltInT _ _))) =
	okWith tcenv TypeT
	--	NB: "BuiltInT" is always OK, like "BuiltInOp"

--
--	(c) "Complex" "TypeVal"s:
--
--		"FunctionT":
--		NB: it's a "FunctionT" itself being type-checked, not an application,
--			so the parms and the return type can be checked independently:
--
typeCheckExpr_N     tcenv (Literal (TypeVal (FunctionT parms retype))) = resN
	where
	resPR, resN ::  [TC_Res]
	resPR = (=&&=)  (map  (typeCheckFB_N tcenv) (parms++[retype]))
	resN  = setType (maybeType tcenv TypeT) resPR

--	(c1) "RecordT":
--		Similar to "FunctionT":
--
typeCheckExpr_N     tcenv (Literal (TypeVal (RecordT field_fbs)))	   = resN
	where
	resF,  resN ::  [TC_Res]
	resF  = (=&&=)  (map  (typeCheckFB_N tcenv) field_fbs)
	resN  = setType (maybeType tcenv TypeT) resF

--	(c2) "UnionT":
--		Similar to "FunctionT" and "RecordT":
--
typeCheckExpr_N     tcenv (Literal (TypeVal (UnionT  fields)))	= resN
	where
	resF,  resN ::  [TC_Res]
	resF  = (=&&=)  (map  (typeCheckFB_N tcenv) fields)
	resN  = setType (maybeType tcenv TypeT) resF

--	(c3) "ADT":
--
typeCheckExpr_N		tcenv (Literal (TypeVal (ADT _ rep_fb dt))) = resN
	where
	--	Type-check both "dt" and "rep_fb",  no need for type inference in the
	--		latter (since (-&&-) below will discard the inferred type anyway):
	--
	resDT, resFB, resN ::   [TC_Res]
	tcenv'			   ::   ETC_Env
	tcenv' =  tcenv {_inferTypes = False}

	resDT  =  typeCheckDT   tcenv' dt  -- Here there is no inference anyway...
	resFB  =  typeCheckFB_N tcenv' rep_fb

	--	Produce the result and possibly set the ADT's inferred type.
	--	XXX: if we infer the type of this ADT, should we construct the appropr
	--		Category type? -- For now, it is just "TypeT":
	--
	resN   =  setType (maybeType tcenv TypeT) (resDT -&&- resFB)

--	(c4) "CategoryT":
--
typeCheckExpr_N     tcenv (Literal (TypeVal (CategoryT _ dt)))  = resN
	where
	--	Type-check the "dt":
	resDT, resN ::  [TC_Res]
	resDT  = typeCheckDT  (tcenv {_inferTypes=False})  dt
	resN   = setType      (maybeType    tcenv  TypeT)  resDT

--
--	(d) "LambdaVal":
--
typeCheckExpr_N		tcenv (Literal (LambdaVal proto body))      = resN
	where
	--	Check the "proto" for consistency as a Type Literal expr:
	--
	resP, resB, resPB, resN :: [TC_Res]

	resP = typeCheckExpr_N   tcenv (Literal (TypeVal proto))

	--	Check the body  against the declared return type (an expr). The latter
	--		is explicitly evaluated first  (unless the "_noEval" flag is set),
	--		since there is no guarantee that it has already been evaluated
	--		(because we may just have a stand-alone "LambdaVal", for example):
	retype :: AldorExpr
	retype  = case proto of
		FunctionT  _ (SimpleFB _ expr _) -> expr
		_  -> error2 "typeCheckExpr_N" ("Invalid Lambda Proto: "++(show proto))

	ertypes:: [TC_Res]
	ertypes = if _noEval  tcenv
			  then
				--	No evaluation:
				[expr2tcr tcenv  retype]
			  else
				--	Evaluate "retype" further:
				map er2tcr (EvalX.evalExpr tcenv retype)

	resB = concat [typeCheckFB_R tcenv er body | er <- ertypes]

	--	"resPB" consists of positive and undecided results of both checks:
	--
	resPB= resP -&&- resB

	--	For each of the "resPB" results, possibly set the evaluated inferred
	--		type. Since we have muti-evaluation, 	 there could be multiple
	--		inferred types, and all results are concatenated:
	--
	resN = concat [setInferredProto tcr | tcr <- resPB]

	--
	--	"setInferredProto":
	--
	setInferredProto:: TC_Res -> [TC_Res]
	setInferredProto   tcr    =	 ress
		where
		eprotos:: [Proto_Res]
		eprotos = if _noEval tcenv
				  then
					--	No evaluation: take the "proto" as is:
					[OK_PR  (getBinds tcenv) proto]
				  else
					--  Evaluate "proto":
					EvalX.evalProto (updateTCE tcr tcenv) proto

		ress   :: [TC_Res]
		ress    = map setThisProto eprotos

		--	In "setThisProto", the NameBinds carried by the "Proto_Res" are
		--		always a proper subset of those for the "tcr" (according to
		--		the way how the "eprotos" were computed),  so we don't need
		--		to intersect binds here:
		--
		setThisProto:: Proto_Res -> TC_Res

		setThisProto   (Bottom_PR  pbinds) =
			--	Hmm... For this "tcr",   we could not evaluate the inferred
			--		type, -- so put  "noTYPE" if the type inference is off,
			--		or convert the whole result into Bottom if it is on:
			--
			if   _inferTypes tcenv
			then Bottom_TCR  pbinds
			else setTypeTCR  noTYPE  (tcr {_bindsR = pbinds})

		setThisProto   (OK_PR   pbinds proto')  =
			setTypeTCR (maybeType tcenv proto') (tcr {_bindsR = pbinds})

--
--	(e) "RecordVal":
--		In the absense of a target type, we can only check that the fields
--		themselves are well-formed.   The inferred type can be too general
--		(e.g. we cannot infer a dependent type):
--
typeCheckExpr_N		tcenv  (Literal (RecordVal fields)) = resN
	where
	resF,  resN  :: [TC_Res]

	--	Type-check all fields and (=&&=) the results. Type inference for fields
	--		can be done at this point (by "typeCheckExpr_N").  In that case the
	--		inferred types are assumed to be already fully evaluated:
	--
	resF   = (=&&=)  [typeCheckExpr_N tcenv (Literal field) | field <- fields]

	--	Construct the inferred types  for all versions of "resF"  (that is, if
	--		"resF" is OK and the type inference in on):
	--
	resN   = map (inferRecordT tcenv (length fields)) resF

--
--	(f) "UnionVal":
--		"UnionVal"s are INTERNAL objects,  so no type-checking is required for
--		them (the check always succeeds).  The Union type CAN sometimes be in-
--		ferred from the UnionVal's selector:
--
typeCheckExpr_N		tcenv  (Literal (UnionVal sel _)) = resN
	where
	--	Inferred Union type is encapsulated as a "TC_Res".   For a given UseNo,
	--		there could be multiple inferred types. In each "consolidated list"
	--		contained in "u_types", there are always 2 elemets: the Union type
	--		itself and the type of the field corresponding to "sel" -- we don't
	--		need the latter here:
	--
	u_types :: [ConsoList NameBinds TC_Res]
	u_types =  inferUnionT    tcenv sel

	resN    :: [TC_Res]
	resN    =  if   _inferTypes tcenv
			   then map    mapUType  u_types
			   else [OK_TCR (getBinds tcenv) noTYPE]

	--	"mapUType":
	--	Installs an inferred Union Type in a TC_Res, if type inference is on:
	--
	mapUType:: ConsoList NameBinds TC_Res -> TC_Res
	mapUType   utcl  =
		case   getCList utcl of
			[OK_TCR _ u_type@[Just _], _] -> OK_TCR     binds u_type
			--	Type inference successful

			[OK_TCR _ _,   _]			  -> OK_TCR     binds noTYPE
			--  Type inference failed but the result is still "OK"

			[Bottom_TCR _, _]			  -> Bottom_TCR binds
			--	Type inference failed, the result is Bottom

			_ -> error2 "typeCheckExpr_N--UnionVal" "Integrity violation"
			--	All other cases: the list is not of length 2: impossible!
		where
		binds :: NameBinds
		binds =  getAttr utcl

--
--	(g) "ADT_Val":
--		"ADT_Val"s are INTERNAL objects;   they are not directly created by the
--		user;   rather, they are created by evaluation of type conversion exprs.
--		So we type-check those exprs; "ADT_Val" always succeeds:
--
typeCheckExpr_N     tcenv  (Literal (ADT_Val adt _)) = resN
	where
	--	The "adt" is already supposed to be normalised (when this "ADT_Val" was
	--		constructed), so there is no need to evaluate it once again:
	--
	resN  :: [TC_Res]
	resN  =  [OK_TCR (getBinds tcenv) (maybeType tcenv adt)]


-------------------------
--	(N3) "Identifier": --
-------------------------
--	Without a target, it always succeeds  (unless a type expr evaluates to Bot-
--		tom), but we  need to  restrict the binds in the result,  and possibly
--		infer the Identifier's type:
--
typeCheckExpr_N  tcenv  (Identifier use_no _) = resN
	where
	--	Get all targets for this "use_no":
	--
	targs :: [BindTarget]
	targs  = lookupBinds (getBinds tcenv) use_no

	--	For each target, restrict the binds for "use_no" to that target,   get
	--		its EVALUATED type expr (as [TC_Res]) via the "getTargEType" func:
	--
	resN  :: [TC_Res]
	resN   = concat (map (getTargEType tcenv use_no) targs)


----------------------
--	(N4) "AndList": --
--	(N5)  "OrList": --
----------------------
--	Components must have the "Boolean" type under compatible binds:
--
typeCheckExpr_N		tcenv (AndList fbs) = resN
	where
	resA,  resN ::  [TC_Res]
	resA = (=&&=)   [typeCheckFB_T tcenv   BooleanT fb | fb <- fbs]
	resN = setType  (maybeType     tcenv   BooleanT)   resA


typeCheckExpr_N		tcenv (OrList  fbs) = resN
	where
	resA,  resN ::  [TC_Res]
	resA = (=&&=)   [typeCheckFB_T tcenv   BooleanT fb | fb <- fbs]
	resN = setType  (maybeType     tcenv   BooleanT)   resA


-------------------------
--	(N6) "IfThenElse": --
-------------------------
--	The condition must type-check against "Boolean", and Then and Else branches
--		must yield the same inferred type;  if the latter cannot be determined,
--		type-checking fails    (so it is more restrictive than in case when the
--		target type is known):
--
typeCheckExpr_N     tcenv (IfThenElse  cond_fb  then_fb  else_fb) = resN
	where
	resC,  resT,  resE,  resX,  resN :: [TC_Res]

	resC = typeCheckFB_T  tcenv BooleanT cond_fb
	resT = typeCheckFB_N  tcenvT      	 then_fb
	resE = typeCheckFB_N  tcenvF		 else_fb

	--	"tcenv{T|F}": type inference is required for branches, and we may also
	--		need to insert "synthetic" "CondTarg"s  into NameBinds.  There may
	--		be different  new binds  propagated  to the Then and Else branches
	--		("bindsT" and "bindsF", resp)!
	--
	bindsT, bindsF :: NameBinds
	bindsT  = mkCondTargs (getBinds tcenv) cond_fb True
	bindsF  = mkCondTargs (getBinds tcenv) cond_fb False

	tcenvT, tcenvF :: ETC_Env
	tcenvT  = setBinds bindsT (tcenv {_inferTypes  = True})
	tcenvF  = setBinds bindsF (tcenv {_inferTypes  = True})

	resX = (=&&=) [resC, resT, resE]

	--	Go through "resX" and retain those list components which have the same
	--		2nd and 3rd inferred type components ("checkBranches" function):
	--
	resN =  [tcr | Just tcr <- map checkBranches resX]

	-----------------------
	--	"checkBranches": --
	-----------------------
	--	Checks that the branches are of the same type, and infers the type of
	--		the whole expr:
	--
	checkBranches:: TC_Res  -> Maybe TC_Res

	checkBranches  (OK_TCR binds [_, Just then_type, Just else_type]) =
		--
		--	Type inference succeeded, so check the type exprs for equivalence:
		--
		case  adeq     adenv  then_type  else_type  of
			Yes_ADER   binds'-> Just (OK_TCR  binds' inf_type)
			No_ADER          -> Nothing
			Bottom_ADER		 -> Just (Bottom_TCR binds)
		where
		adenv   :: ADE_Env
		adenv   =  mkADEnv  binds (_maxADEpth tcenv) (_debugMode tcenv)

		--	Inferred type: no need to evaluate it again, we should get "then_"
		--		and "else_type"s already fully evaluated at this point,   and
		--		they are already verified to be equivalent:
		--
		inf_type:: InferredType
		inf_type=  maybeExpr tcenv then_type

	checkBranches  (OK_TCR _ _)		  = Nothing
		--	Type inference failed for one or both branches

	checkBranches  hmm@(Bottom_TCR _) = Just hmm
		--	"Bottom"s are to be preserved


---------------------------
--	(N7) "FunctionCall": --
---------------------------
--	Infer the type(s) of the callable object  (which must  be "FunctionT" or
--		"BuiltInT") and check the args against those types.  The return type
--		is not checked here but  inferred if necessary;  this is the type of
--		the whole expression.
--	"FunctionCall" can also stand for reading a Record's Field; in that case
--		the "callable"'s type must be "RecordT", and there must be exactly 1
--		"arg" -- the Field Identifier:
--
typeCheckExpr_N    tcenv fc@(FunctionCall callable args) = resN++ botN
	where
	--	Type-check the callable and infer its type:
	--
	resC      :: [TC_Res]
	resC  	  =  typeCheckFB_N (tcenv {_inferTypes = True}) callable

	--	Valid "callable" types: "BuiltInT", "FunctionT", "RecordT" and
	--		"UnionT". Find these types among "resC" and type-check the
	--		args against their parms. Bottoms in "resC" are preserved,
	--		all other "resC"s are dropped:
	--
	cprotos   :: [(NameBinds, AldorType)]
	cprotos   =  [(binds,     hmm) |
				  OK_TCR binds [Just (Literal (TypeVal hmm))] <- resC,
				  isCallable  hmm]
--	cprotos   =  trace ("\nCALLABLE: "++(show callable)++"\n") cprotos'

	isCallable:: AldorType		-> Bool
	isCallable   (BuiltInT  _ _) = True
	isCallable   (FunctionT _ _) = True
	isCallable   (RecordT     _) = True
	isCallable   (UnionT      _) = True
	isCallable   _				 = False

	--	Check the "args" against each of the "cprotos" using
	--		"checkAgainstCallable":
	--	
	resN	  :: [TC_Res]
	resN 	  =  concat (map checkAgainstCallable cprotos)
--	resN = trace ("\nTYPE-CHECKING FUNCTION CALL: "++(show fc)++
--				  "\nRESULT: "++(show resN')++"\n") resN'


	botN	  :: [TC_Res]
	botN	  =  [bot | bot@(Bottom_TCR _) <- resC]

	------------------------------
	--	"checkAgainstCallable": --
	------------------------------
	checkAgainstCallable ::  (NameBinds, AldorType) -> [TC_Res]

	--
	--	(1) Check against a "BuiltInT".
	--		Currently (XXX) "BuiltInT" cannot be dependent, so a simplified
	--		type-checking function is applied here (not "typeCheckAppl"):
	--
	checkAgainstCallable     (binds,     BuiltInT      parms   retype)
		| length parms /= length args = []
		| otherwise     = res2
		where
		tcenv1 ::  ETC_Env
		tcenv1 =  setBinds binds tcenv

		--	Type-check the "args" against "parms":
		res1   :: [TC_Res]
		res1   =  (=&&=) (zipWith    (typeCheckFB_T    tcenv1) parms args)

		--	Infer the application type:
		res2   :: [TC_Res]
		res2   =  map    (setTypeTCR (maybeType tcenv retype)) res1

	--
	--	(2) Check against a "FunctionT".
	--		As "FunctionT" can be dependent,  "typeCheckAppl" is used here:
	--
	checkAgainstCallable     (binds,     FunctionT     parms   retype_fb) =
		res2
		where
		tcenv1 :: ETC_Env
		tcenv1 =  setBinds binds tcenv

		--	Type-check the "args" against "parms" (taking "retype_fb" into
		--		account).  The "args" on which dependent types may depend,
		--		are NOT yet evaluated (hence "False"):
		--
		res1   :: [(EvalStack,  TC_Res)]
		res1   =  typeCheckAppl  tcenv1 False (parms++[retype_fb]) args

		--	Infer the application types ("typeCheckAppl" could not do it, as it
		--		processes generic sequences of parms, not just function protos)
		--		using the context returned by that function:
		--
		res2   :: [TC_Res]
		res2   =  concat  (map inferApplType res1)

		--
		--	"inferApplType":
		--	Infers the type of the function application, i.e. the return type
		--		of the corresp callable (if requested).    The return type is
		--		explicitly EVALUATED unless the "_noEval" flag is set:
		--
		inferApplType :: (EvalStack, TC_Res) -> [TC_Res]

		inferApplType    (es,        tcr)
			| _noEval tcenv =
				--	No evaluation of "retype_fb":
				--
				[setTypeTCR  (maybeExpr tcenv (getExpr retype_fb)) tcr]

			| otherwise     = erts
			where
			--	Make the type inference environment through "tcr" and "es":
			--
			tcenv2:: ETC_Env
			tcenv2 = (updateTCE tcr tcenv1) {_evalStack = es}

			--	Evaluate "retype_fb" in the dependent environment:
			--
			erts  :: [TC_Res]
			erts  =  [setTypeTCR  (mkApplType  er)  tcr  |
							  er<- EvalX.evalFB tcenv2 retype_fb]

			--	"mkApplType":
			--	The argument is an evaluated "retype" -- the inferred type of
			--		the function application:
			--
			mkApplType :: EvalRes -> InferredType

			mkApplType   (Full_ER    _    type_val) =
				maybeExpr tcenv2 (Literal type_val)

			mkApplType   (Partial_ER _ (SimpleFB _  type_expr _)) =
				maybeExpr tcenv2 type_expr

			mkApplType   _			 = noTYPE

	--
	--	(3) Check against a "RecordT":
	--		For a "RecordT", there must be a singleton "args" containing an
	--		Identifier which is the Field selector:
	--
	checkAgainstCallable     (binds, RecordT fields) =
		case args of
			[SimpleFB _ (Identifier use_no _) _] ->
				let
					--	(1)
					--  Find the field(s)  to which this  Selector is bound.
					--		If multiple  fields are found,  the Selector is
					--		overloaded, -- that's still OK.
					--
					--	Possible targets for "use_no" among the "fields".
					--		For each target field found, we also remember
					--		its index:
					ftargs :: [(Int, AldorFBlock)]
					ftargs =  findTargFields binds use_no fields

					--  (2)
					--	OK: the type-checking results are, generally speaking,
					--		inferred from the type exprs of "ftargs". However,
					--		if the Record type is dependent,  these types may
					--		need to be evaluated in the environment depending
					--		on the actual "callable".

					--	Get the values of the "callable". It is EVALUATED un-
					--		less the "_noEval" flag is set,   and unless its
					--		type is non-dependent.   Separate evaluations are
					--		produced for all "ftargs",   as the binds for the
					--		Selector are restricted  to a particulat "ftargs"
					--		element:
					--
					cVals :: [(Int, [EvalRes])]
					cVals =  map evalCallable ftargs

					evalCallable :: (Int, AldorFBlock) -> (Int, [EvalRes])
					evalCallable    (idx, ftarg)	   =
						if _noEval tcenv || getDepParms tcenv fields == []
						then
							--	No evaluation for "callable":
							(idx, [Partial_ER binds' callable])
						else
							--	Evaluate "callable" (e.g. a Record-valued
							--		expr):
							(idx, EvalX.evalFB (setBinds binds' tcenv) callable)
						where
						binds' :: NameBinds
						binds' =  modifyBinds binds use_no [FP_Targ ftarg]

					--
					--	We now need to make stacks of field values from "cVals"
					--		in order to evaluate the type of the selected field.
					--	The results are again grouped per-"ftarg":
					--
					dEnvs  :: [(Int, [ETC_Env],  Bool)]
					dEnvs  =  concat [[mkDepEnvs idx cRes | cRes <- cRess] |
									 (idx, cRess) <- cVals]

					--	In "mkFldStacks", we type-check the (possibly evaluated)
					--		"callable" against the RecordT fields. This is done
					--		NOT for the sake of type-checking itself,  as we've
					--		already inferred this RecordT  from  the "callable"
					--		before its evaluation, but just to build  dependent
					--		stacks in which the TYPE of the selected field  can
					--		then be evaluated. The "Bool" res is a Bottom flag:
					--
					mkDepEnvs :: Int -> EvalRes -> (Int, [ETC_Env], Bool)

					--	(a)
					--	Fully evaluated "callable" ("RecordVal"):  Put field
					--		values on the stack.   We just constructed these
					--		field values -- they are fully evaluated already,
					--		so indicate this by "True" in "TypeCheckAppl".
					--	For each particular target field "idx", we don't need
					--		to put "f_vals" with indices >=idx  on the stack,
					--		as the type of the field selected by that  "idx"
					--		cannot depend on the values of such "f_vals":
					--
					mkDepEnvs    idx (Full_ER  binds' (RecordVal f_vals)) =
						let
							i_fields :: [AldorFBlock]
							i_fields =  take idx fields

							i_vals   :: [AldorValue]
							i_vals   =  take idx f_vals

							dss		 :: [EvalStack]
							dss		 =  mkDepStacks
											(setBinds binds' tcenv)
											True 0
											i_fields
											[simpleFB (Literal val) |
															   val <- i_vals]
						in
							(idx,
							[setBinds binds' (tcenv{_evalStack=ds}) | ds <-dss],
							False)

					--	(b)
					--	All other fully-evalauated "callable"s: improper:
					--
					mkDepEnvs  idx  (Full_ER _ _) = (idx, [], False)

					--	(c)
					--	Partially-evaluated "callable", to a "RecordCtor": can
					--		still put (fully- or partially-evaluated, but any-
					--		way already normalisd-- hence "True" below) fields
					--		on the stack. Again, do full type-checking  of the
					--		first "idx" "f_args" against the corresp "fields":
					--
					mkDepEnvs  idx  (Partial_ER  binds'
									(SimpleFB _ (RecordCtor f_args) _)) =
						let
							i_fields :: [AldorFBlock]
							i_fields =  take idx fields

							i_args   :: [AldorFBlock]
							i_args   =  take idx f_args

							dss		 :: [EvalStack]
							dss		 =  mkDepStacks
											(setBinds binds' tcenv) True 0
											i_fields i_args
						in
							(idx,
							[setBinds binds' (tcenv{_evalStack=ds}) | ds <-dss],
							False)

					--	(d)
					--	Partially-evaluated "callable", to any other object,
					--		e.g. an Identifier:    Nothing to be put on the
					--		stack, BUT WE CAN STILL GO AHEAD:
					--
					mkDepEnvs  idx  (Partial_ER binds' _) =
						(idx,  [setBinds binds' tcenv],  False)

					--	(e)
					--	"Callable" evaluated to "Bottom" -- set the flag!
					--
					mkDepEnvs  idx  (Bottom_ER  binds')   =
						(idx,  [setBinds binds' tcenv], True)


					--	(3)
					--	Now, we extract the type of the selected field,  and
					--		evaluate that type in all of the dependent  envs
					--		constructed ("dEnvs").
					--	XXX: if no evaluation is allowed, or if the type  of
					--		the field is not to be inferred, we could produce
					--		the result almost straight from "ftargs", except
					--		for find settings of binds. OPTIMISATION?
					--
					res :: [TC_Res]
					res =  concat [concat [evalFldType   idx env bot_flg |
										   env  <- envs]				 |
								  (idx, envs, bot_flg) <-dEnvs]


					evalFldType :: Int -> ETC_Env -> Bool -> [TC_Res]

					evalFldType    _ env True = [Bottom_TCR (getBinds env)]

					evalFldType    idx env _  =
						if   _inferTypes   env
						then getFParmEType env  (fields !!! idx)
						else [OK_TCR binds noTYPE]
				in
					--	Deliver it at last!
					res

			_  ->
				--	Not a Field Selector (as a single Identifier):
				[]

	--
	--	(4): Check against a "UnionT":
	--		 Similar to the "RecordT" case: there must a single arg which is a
	--		 Union field selector:
	--
	checkAgainstCallable     (binds, UnionT fields) =
		case args of
			[SimpleFB  _ (Identifier use_no _) _] ->
				let
					--
					--	If "use_no" is bound to any of the "fields", the check
					--		succeeds and we infer the type of that field. Un-
					--		like "Record"T, a "UnionT" cannot be dependent, so
					--		we don't need to evaluate the "callable":
					--
					ftargs :: [AldorFBlock]
					ftargs =  map snd (findTargFields binds use_no fields)

					utcrs  :: [TC_Res]
					utcrs  =  map getFldType ftargs

					getFldType :: AldorFBlock -> TC_Res

					getFldType    fp@(FParmFB _ _ _) = OK_TCR binds [Just te]
						where
						GenFDTE _ _ te _ = getFParm fp

					getFldType    hmm    =
						error2 "getFldType" ("Not an FParmFB: "++(show hmm))
				in
					utcrs

			_ ->
				--  Not a Field Selector :
				[]

	--
	--	(5): All other cases: they should not occur:
	--
	checkAgainstCallable (_, hmm) =
		error2 "checkAgainstCallable"
			   ("Internal error: Invalid callable: "++ (show hmm))


-------------------------
--	(N8) "RecordCtor": --
-------------------------
--	Since the type of the record being constructed is unknown to the ctor, we
--		can only check the "fields" for internal consistency. The record type
--		can be inferred the same way as for "RecordVal"s:
--
typeCheckExpr_N    tcenv (RecordCtor field_fbs) = resN
	where
	resF,  resN  :: [TC_Res]

	resF   = (=&&=) (map (typeCheckFB_N tcenv)  field_fbs)
	resN   = map    (inferRecordT tcenv (length field_fbs))  resF


-------------------------
--	(N9) "UnionCtor" : --
-------------------------
--	Unlike "UnionVal", here we also need to type-check  the "tag_expr" (which
--		must be an Identifier) and the "field_fb". The Union type is inferred
--		(if possible -- this can fail!) and the "field_fb" is checked against
--		that type:
--
typeCheckExpr_N    tcenv (UnionCtor tag_expr field_fb) = resN
	where
	--	Infer the Union type(s); see "UnionVal" type-checking for the
	--		description of the "inferUnuinType" result format:
	--
	u_types :: [ConsoList NameBinds TC_Res]
	u_types =  inferUnionT    tcenv tag_expr

	--	"field_fb" is to be checked against inferred field types. If any
	--		the latter is "OK" but not inferred properly,   the  corresp
	--		combination is skipped;  if the target is "Bottom", "Bottom"
	--		is returned in that case:
	--
	resN :: [TC_Res]
	resN =  concat (map mapUType u_types)

	mapUType :: ConsoList NameBinds TC_Res  -> [TC_Res]
	mapUType	utcl  =
		case getCList utcl of
			--
			--	The Field type is properly inferred:
			--
			[hmm_ut, OK_TCR _ [Just ft_expr]] ->
				let
					--	Type-check "field_fb" against "ft_expr":
					--
					resU :: [TC_Res]
					resU =  typeCheckFB_E (setBinds (getAttr utcl) tcenv)
										  ft_expr field_fb
				in
					--	Now, the action depends on "hmm_ut":
					case hmm_ut of
						--	(i)
						--	The Union type is also properly inferred:
						--
						OK_TCR _ [Just ut_expr] ->
							setType (maybeExpr tcenv ut_expr) resU

						--	(ii)
						--	The Union type is nominally "OK"  but not
						--		properly inferred: set it to "noTYPE":
						--
						OK_TCR _ _ ->    setType noTYPE resU

						--	(iii)
						--	The Union type is Bottom: the action then
						--		depends on the "_inferTypes" flag:
						--
						Bottom_TCR _ ->
							if   _inferTypes tcenv
							then [Bottom_TCR (_bindsR ru) | ru <- resU]
							else setType noTYPE resU
			--
			--	Now, the Field type is nominally "OK" but not properly
			--		inferred:  Cannot type-check "field_fb",   so skip
			--		this result at all:
			--
			[_, OK_TCR _ _]   -> []

			--
			--	Field type is Bottom: produce Bottom:
			--
			[_, Bottom_TCR _] -> [Bottom_TCR (getAttr utcl)]

			--
			--	All other cases: The list is not of length 2: impossible!
			--
			_ -> error2 "typeCheckExpr_N--UnionCtor" "Integrity violation"


--------------------------
--	(N10) "UnionCase":  --
--------------------------
--	Basically, we just need to check that the "unionFB" is indeed of "UnionT".
--	We do NOT check that RHS tag belongs to the set of valid tags for that
--		"UnionT", as if it does not, evaluation of "UnionCase"  would just
--		produce "False".  On the other hand, if the RHS of the "UnionCase"
--		expr is not an Identifier, it's not even a type error, but a syntax
--		error!
--
typeCheckExpr_N    tcenv (UnionCase unionFB (Identifier _ _)) = resN
	where
	resN ::   [TC_Res]
	resN =    concat
			  (map isUnion (typeCheckFB_N (tcenv {_inferTypes=True}) unionFB))

	isUnion:: TC_Res -> [TC_Res]

	isUnion   (OK_TCR binds [Just (Literal (TypeVal (UnionT _)))]) =
			  -- The LHS is OK, the result is Boolean:
			  [OK_TCR binds [Just (Literal (TypeVal BooleanT)) ]]

	isUnion   bot@(Bottom_TCR _) = [bot]
			  -- Bottoms are propagated

	isUnion   _ = []
			  -- Other types (incl failed inferences) for the "unionFB" fail


typeCheckExpr_N	_ (UnionCase _  hmm) =
	error2 "TypeCheckExpr_N--UnionCase"
			("RHS is not an Identifier: "++(show hmm))

 
----------------------------
--	(N11) "RestrictType": --
----------------------------
--	Evaluate the "target_type" (unless the "_noEval" flag is set)  and  type-
--		check "fb" against it. The inferred type (if any) is the target type,
--		multiple versions of which are possible:
--
typeCheckExpr_N    tcenv (RestrictType any_fb target_type) = resN
	where
	etype ::  [EvalRes]
	etype =   if  _noEval tcenv
			  then
				--	No evaluation of "target_type":
				[Partial_ER (getBinds tcenv) target_type]

			  else
				--	Evaluate "target_type":
				EvalX.evalFB tcenv target_type

	resN  ::  [TC_Res]
	resN  =   concat (map restrictTo etype)

	restrictTo :: EvalRes   -> [TC_Res]
	--	Restricts the type of "any_fb" to the one given by the arg

	restrictTo   (Bottom_ER  binds) =  [Bottom_TCR binds]

	restrictTo   (Partial_ER binds targ_fb) =
		case targ_fb of
			SimpleFB _ targ_expr _  ->
				--	Extract the "targ_expr" and type-check against it.
				--		This is the main case:
					typeCheckFB_E (setBinds binds tcenv) targ_expr any_fb

			NestedFB _ _ innerFB _  ->
				--	Make a synthetic "EvalRes" with "innerFB" and proceed
				--		recursively:
				restrictTo (Partial_ER binds innerFB)

			--	The other two cases (VoidFB and FParmFB) should never occur:
			--
			hmm -> error2 "restritTo" ("Improper target FB: "++(show hmm))

	restrictTo   (Full_ER   binds targ_val) =
		case targ_val of
			TypeVal targT -> typeCheckFB_T (setBinds binds tcenv) targT any_fb
			_			  -> []


-------------------------
--	(N12) "ConvertTo": --
-------------------------
--	This is a "controlled pretend". It is type-checked via (actually) INFERRING
--		the type of the "convertible", which may not always work,  but the only
--		other way of doing this would be to EVALUATE the convertible  and check
--		it against the target type, which may be far too expensive:
--
typeCheckExpr_N    tcenv  (ConvertTo convertible target) = resN
	where
	--	Type-check the "convertible" and infer its type(s):
	--
	resC :: [TC_Res]
	resC =  typeCheckFB_N (tcenv {_inferTypes = True}) convertible

	--	For each inferred type in "resC", replace it with the ULTIMATE
	--		representation type  (which is different from the inferred
	--		type itself if the latter is an ADT):
	--
	resR :: [TC_Res]
	resR =  concat (map setUltimType resC)

	--
	--	"setUltimType":
	--
	setUltimType :: TC_Res ->   [TC_Res]

	setUltimType    bot@(Bottom_TCR _)		=   [bot]

	setUltimType    ok @(OK_TCR binds [Just inf_type]) =
		case inf_type of
			--
			--	An ADT Literal: Get its Ultimate Rep type:
			--
			Literal (TypeVal adt@(ADT _ rep_fb _)) ->
				let
					tcenv' :: ETC_Env
					tcenv' =  updateTCE ok tcenv
				in
					if  _noEval tcenv
					then
						--	Evaluation not allowed: get unevaluated "rep_fb":
						--
						[fb2tcr tcenv' rep_fb]
					else
						--	Multiple repr types are possible here,   since eva-
						--		luation is performed (by "getADTrep") to deter-
						--		mine them:
						map er2tcr (EvalLib.getUltADTrep tcenv' adt)
			--
			--	A "%" Identifier: if bound to an ADT, get the latter's ultimate
			--		  Rep. The "%" cannot be overloaded,  so we expect only one
			--		  target for it:
			--
			ident@(Identifier use_no ["%"]) ->
				case lookupBinds binds use_no of

					[PerCentTarg expr]      ->
						--	Go into that "expr" by constructing a synthetic TCR:
						setUltimType (OK_TCR binds [Just expr])

					hmm -> error2 "setUltimType"
								("Invalid binds for "++(show ident)++": "++
								(show hmm))
			_ ->
				-- For non-ADT inferred types, retain the original result:
				[ok]

	-- Type not properly inferred: skip this result...
	--
	setUltimType    _ =  []


	--	Now:
	--	Evaluate the "target" (unless the "_noEval" flag is set),  and convert
	--		the results into "TC_Res"s. Again, we need ULTIMATE "target" types,
	--		not just the evaluated "target":
	--
	resT :: [TC_Res]
	resT =  if _noEval tcenv
			then
				-- Take "target" as is:
				[fb2tcr tcenv target]
			else
				-- Evaluate "target":
				map er2tcr (EvalX.evalFB tcenv target)

	resU :: [TC_Res]
	resU =  concat (map setUltimType resT)

	--	Unify the attrs of "resR" and "resU":
	--
	resRU:: [ConsoList NameBinds TC_Res]
	resRU = crossWithAttrs  [resR, resU]

	--	Check each pair in "resRU" for AD-equivalence. There is a BIG problem
	--		here, however. Normally, if the inferred type is constructed as a
	--		result of type-checking, it would be evaluated (normalised), i.e.
	--		"typeE" below. However, by the nature of "ConvertTo", some callers
	--		would expect it to be the original "target". Which type should we
	--		infer? -- As there is no way of telling that, we produce TWO res-
	--		ults for each successful match!
	--	XXX: BUT: This would mean that we have TWO different type-checking
	--		results with the same NameBinds! -- This is OK in the current imp-
	--		lementation, since we do caching on evaluation results only,  NOT
	--		on type-checking results. To avoid propagation of multiple IDENTI-
	--		CAL results, we apply "nub" here:
	--
	resN :: [TC_Res]
	resN =  nub [jres | Just jres <- concat (map typeCheckRU resRU)]

	typeCheckRU :: ConsoList NameBinds TC_Res -> [Maybe TC_Res]
	typeCheckRU    cltcr =
		case getCList cltcr of
			[OK_TCR _ [Just typeC], OK_TCR _ [Just typeE]] ->
				--
				--	Compare type exprs here. If they match, install "typeT"
				--		in the result (if we do type indference for it):
				--	The type to be inferred, however, must be the original
				--		"target", NOT its ultimate evaluated version!
				--	XXX: Can "getExpr" here cause any problems if "target"
				--		is a FBBlock with a DefTable? -- Probably not,  as
				--		all binds of that expr will be preserved anyway:
				--
				case adeq    (etc2ade tcenv') typeC typeE  of
					Bottom_ADER      ->
						[Just (Bottom_TCR   binds)]
					Yes_ADER  binds' ->
						[Just (OK_TCR binds'
							  (maybeExpr tcenv (getExpr target))),
						Just (OK_TCR binds'
							  (maybeExpr tcenv typeE))]
					No_ADER          ->
						[Nothing]

			--
			--	At least one Bottom -- produce Bottom:
			--
			[Bottom_TCR _, _]	   -> [Just (Bottom_TCR   binds)]
			[_, Bottom_TCR _]	   -> [Just (Bottom_TCR   binds)]

			--
			--	All other cases (at least one type not fully inferred) --
			--		check fails:
			_  ->  [Nothing]
		where
		binds  ::  NameBinds
		binds  =   getAttr cltcr

		tcenv' ::  ETC_Env
		tcenv' =   setBinds binds tcenv


------------------------------
--	(N13) "Add*", "With*":  --
------------------------------
--	"Add1" and "With1" are easy:
--
typeCheckExpr_N tcenv (Add1  _ fb) =
	setType (maybeType tcenv TypeT) (typeCheckFB_N tcenv fb)

typeCheckExpr_N tcenv (With1 _ fb) =
	setType (maybeType tcenv TypeT) (typeCheckFB_N tcenv fb)

--	"Add2", "With2":
--	What are the constraints on FBs here? E.g. they must yield just a DefTable
--		or an ADT Literal... but there are no particular type rules for that!
--		So just check FBs for internal consistency, and infer "TypeT":
--
typeCheckExpr_N tcenv (Add2  _ fb1 fb2) =
	setType (maybeType tcenv TypeT)
	   ((typeCheckFB_N tcenv fb1) -&&- (typeCheckFB_N tcenv fb2))

typeCheckExpr_N tcenv (With2 _ fb1 fb2) =
	setType (maybeType tcenv TypeT)
	   ((typeCheckFB_N tcenv fb1) -&&- (typeCheckFB_N tcenv fb2))


-- ================== --
--	"typeCheckFB_N":  --
-- ================== --
--	The same as "typeCheckExpr_N", but for checking FBlocks:
--
typeCheckFB_N :: ETC_Env -> AldorFBlock -> [TC_Res]

--	(1) "VoidFB":
--		It should not apperar in  any legitimate exprs under type-checking, so
--		produce an error here:
--
typeCheckFB_N    _         (VoidFB	fbno) =
		error2 "typeCheckFB_N" ("VoidFB No."++(show fbno)++" encountered")

--
--	(2) "SimpleFB":
--		Type-check its expr:
--
typeCheckFB_N     tcenv    (SimpleFB _ expr _)	   =
		typeCheckExpr_N tcenv  expr

--
--	(3) "NestedFB":
--		Type-check its "innerFB" recursively,  and type-check the DefTable.
--		NB: "innerFB" is allowed to be "VoidFB", in that case it's skipped:
--
typeCheckFB_N     tcenv    (NestedFB _ dt innerFB _) =
	case innerFB  of
		VoidFB _  -> resD
		_	      -> resN
	where
	resI, resD, resN :: [TC_Res]

	resI= typeCheckFB_N tcenv innerFB
	resD= typeCheckDT   tcenv dt

	--	If the type of the whole NestedFB is inferred, it's that of the "resI",
	--		so we cross "resI" and "resD"   and lift the type of the former to
	--		the whole result:
	--
	resN= map liftLeftType (resI -&&- resD)

	--	"liftLeftType": --
	--
	liftLeftType :: TC_Res -> TC_Res
	liftLeftType    bot@(Bottom_TCR _)        = bot
	liftLeftType    (OK_TCR binds [typeI, _]) = OK_TCR binds [typeI]
	liftLeftType    (OK_TCR _     hmm)        =
			error2 "liftLeftType" ("Not a 2-element list: "++(show hmm))

--
--	(4) "FParmFB":
--		Just type-check the type-expr (against TypeT) in its singleton DT:
--
typeCheckFB_N       tcenv  fb@(FParmFB _ _ _)  =
	typeCheckExpr_T tcenv' TypeT type_expr
	where
	tcenv' :: ETC_Env
	tcenv' =  tcenv {_inferTypes = False}  -- Don't need type inference here!

	type_expr :: AldorExpr
	GenFDTE _ _  type_expr _ = getFParm fb
	--	Always "GenFDTE" here -- checked by "getFParm" itself


-- =============================== --
--	Other Type-Checking Functions: --
-- =============================== --
----------------------
--	"typeCheckDT":  --
----------------------
--	This function indeed belongs to the "_N" group:   for DefTables, there is
--		no single type-checking target!
--	The result of "typeCheckDT" is always  an empty (failure)  or a singleton
--		(success) list of "TC_Res", since type-checking DTEs is a  "reduction
--		point": ambiguities are not allowed there, so exactly 1 "OK_TCR" must
--		be produced for a DTE in question to type-check successfully:
--
typeCheckDT :: ETC_Env -> DefTable -> [TC_Res]

typeCheckDT    tcenv     dt       =  setType noTYPE ((=&&=) resDTEs)
	where
	--	For DTEs (at the top level at least), there is no need to do any type
	--		inference, since we do not infer the type of the DT itself:
	--
	tcenv1  :: ETC_Env
	tcenv1  =  tcenv {_inferTypes = False}

	--	Go through the "dt" and type-check all "GenDTE"s encountered. "RenDTE"s
	--		and Imports are not interesting. The results are accumulated in the
	--		"resDTE"s list:
	--
	resDTEs :: [[TC_Res]]
	--	Outer dim: DTEs; inner dim: result versions for each DTE

	resDTEs =  foldDT typeCheckDTE [] dt

	----------------------
	--	"typeCheckDTE": --
	----------------------
	typeCheckDTE :: FlatDTE -> [[TC_Res]] -> [[TC_Res]]

	typeCheckDTE   (GenFDTE def_no name type_expr body_fb) curr =
		let
			thisDef:: String
			thisDef = "\""++name++"\" (#"++(show def_no)++")"

			--	NB: We do NOT check here  that "type_expr" is indeed   type-
			--		forming (i.e. that it type-checks against "TypeT"): this
			--		is done later when it actually serves as a target.  This
			--		is because in many cases the "type_expr" is actually   a
			--		TypeVal object already, so it does not need any checks.
			--		If it is really an expr, it will be checked by the corr
			--		functions in the "TypeCheckTER" module.

			--	So:
			--	EVALUATE   (possibly to multiple values) the "type_expr":
			--
			resTE:: [TC_Res]
			resTE = case getDefEType tcenv1 (GenDTE def_no type_expr body_fb) of
						[] ->
							debug_msg tcenv1
								("\nWARNING: No target types for "++ thisDef  ++
								 "\n")	  []
						ok -> ok

			--	Then type-check "body_fb" against "resTE" elements (the env
			--		is adjusted automatically for each target).
			--	There is a special case here: "FunctionT" which is actually
			--		a shortcut to the Lambda type expr:
			--
			resB :: [TC_Res]
			resB =  nub (concat [checkBodyAgainstTE tcr | tcr <- resTE])

			----------------------------
			--	"checkBodyAgainstTE": --
			----------------------------
			--	The "body_fb" CAN be missing (Declaration without a Definition,
			--		but not visa versa, except for Renamings).  We will handle
			--		this condition HERE, as "typeCheckFB_R" to be called  does
			--		NOT allow "VoidFB" in the RHS, for safety reasons:
			--
			checkBodyAgainstTE :: TC_Res -> [TC_Res]
			checkBodyAgainstTE    type_tcr    = xxx
--				case xxx of
--					[] -> trace ("\nFAILED "++(show body_fb)++"\nAGAINST\n"++
--						  (show type_tcr)++"\n\n") []
--					ok -> ok
				where
				xxx = case body_fb of
					--	No Body -- just install the inferred type if required:
					--
					VoidFB _ ->
						case type_tcr of
							OK_TCR _ [Just dcl_type] ->
								[OK_TCR (getBinds  tcenv2)
										(maybeExpr tcenv2 dcl_type)]
								-- "type_expr" may be inferred even for VoidFB!

							OK_TCR _ _   ->
								[OK_TCR (getBinds tcenv2) noTYPE]

							Bottom_TCR _ ->
								[Bottom_TCR (getBinds tcenv2)]

					--	Body present -- type-check it for real:
					_  ->
						case type_tcr of
							OK_TCR _ [Just (Literal
										   (TypeVal (FunctionT _ retype)))]  ->
								--
								--	Check "body_fb" against "retype". The latter
								--		is already evaluated, since it origina-
								--		tes from one of the "resE"; the LHS
								--		is also already evaluated, as it comes
								--		from "resTE":
								let
									zzz = typeCheckFB_E tcenv2
													   (getExpr retype) body_fb
								in
--									trace ("\n===LHS: "++(show retype )++
--										   "\n===RHS: "++(show body_fb)++
--										   "\n===RESULT:\n"++(show zzz)++"\n")
									zzz

							_ ->
								--
								--	All other cases: Check "body_fb" against the
								--		whole "tcr" (which is aready evaluated):
								let
									zzz = typeCheckFB_R tcenv2 type_tcr body_fb
								in
--									trace ("\n===ENTRY: "++name++
--										"\n===CHECKING "++(show body_fb)++
--										"\n===AGAINST\n"++(show type_tcr)++
--										"\n===RESULT:\n"++(show zzz)++"\n")
									zzz
				tcenv2 :: ETC_Env
				tcenv2 =  updateTCE type_tcr tcenv1
		in
			--	Construct the intersection and analyse its structure:
			--
			case resB of
				--
				--	Single OK result:  really OK:
				--
				ok@[OK_TCR _ _]    ->
					debug_msg tcenv1 ("\nDef#"++(show def_no)++" SUCCEEDED\n")
							  (ok : curr)

				--
				--	Single Bottom result: also accepted, but then the whole DT
				--		type-checks to Bottom:
				--
				bot@[Bottom_TCR _] ->
					debug_msg tcenv1 ("\nWARNING: Def#"++(show def_no)++
									  ": BOTTOM\n") (bot : curr)

				--
				--	No results at all: will trigger failure of the whole DT:
				--	May need to produce a message here:
				--
				[]	-> debug_msg tcenv1
							("\nWARNING: Type-checking of Def#" ++
							(show def_no)++" FAILED: No matches\n")
							([]:curr)

				--
				--	Multiple results (OK and/or Bottom): ambiguity,  so also
				--		produce a failure.   Again, may need a message here:
				--
				amb -> debug_msg tcenv1
							("\nWARNING: Type-checking of Def#"   ++
							(show def_no)++" FAILED: Ambiguity:\n"++
							(show amb)++"\n")
							([]:curr)

		--	All done here!

	typeCheckDTE   (RenFDTE _ _ _) curr = curr -- "RenFDTE": Ignore it


-----------------------
--	"typeCheckAppl": --
-----------------------
--	Here checking against instances of dependent types takes place, e,g, for
--		Record Aggregates and Function Applications.   This function returns
--		type-checking results along with the EvalStack constructed to obtain
--		each of them,    so the latter can be used in further type inference
--		(e.g. to specialise the return type of a dependently-typed function).
--	The "protoFBs" must be of the same length as "args", or 1 element longer,
--		in that case the last "protoFB" is considered to be a Function return
--		type (not used for Record aggregates!):
--
typeCheckAppl :: ETC_Env -> Bool -> [AldorFBlock] -> [AldorFBlock] ->
				 [(EvalStack,   TC_Res)]

typeCheckAppl	 tcenv     evaldFlag    protoFBs    args
	| lengthDiff /= 0 && lengthDiff /= 1 =
	  debug_msg  tcenv
			("\nWARNING: Invalid number of function args:" ++
             "\nPROTO  : "++(show protoFBs)++
			 "\nARGS   : "++(show args)++"\n")
	  		[]
	| otherwise= resF
	where
	lengthDiff:: Int
	lengthDiff = (length protoFBs) - (length args)

	--	"parmFBs" are for function parms only, not the return type:
	--
	parmFBs   :: [AldorFBlock]
	parmFBs    = if   lengthDiff == 0
				 then protoFBs		-- All elements
				 else init protoFBs -- All but last elements

	--	Args are considered to be already evaluated (so no new evaluations will
	--		be done on them) if the "evaldFlag" or the "_noEval" flag is set:
	evalmFlag :: Bool
	evalmFlag =  evaldFlag || (_noEval tcenv)

	--	Form eval stacks and "new_tcenvs"  for types contained in "parmFB"s:
	--	Here the FULL "protoFBs" (including the function return type if pre-
	--		sent) are used to determine the dependent parameters:
	--
	ess       :: [EvalStack]
	ess 	  =  mkDepStacks tcenv evalmFlag lengthDiff protoFBs args

	new_tcenvs:: [ETC_Env]
	new_tcenvs = [tcenv {_evalStack = es}   | es <- ess]

	--	For each of the "new_tcenvs" formed, evaluate all types of "parmFB"s
	--		and match "args" against them:
	--
	resF :: [(EvalStack, TC_Res)]

	resF =  concat [[(_evalStack ntcenv, tcr) |
				tcr <- typeCheckArgs   ntcenv] | ntcenv <- new_tcenvs]

	-----------------------
	--	"typeCheckArgs": --
	-----------------------
	--	Type-checks all "args" in the given ETC_Env. There could be multiple
	--		evaluations for each parm,   so there are multiple parm vectors
	--		("xptypes") against which the "args" are checked:
	--
	typeCheckArgs :: ETC_Env -> [TC_Res]

	typeCheckArgs    tcenv' =  resX
		where
		eptypes :: [[TC_Res]]
		eptypes =  map (getFParmEType tcenv') parmFBs
		--	Inner dim: value versions,   outer dim: parms

		xptypes :: [TCR_CL]	 		   		
		xptypes =  crossWithAttrs  eptypes
		--	Inner dim: parms; outer dim: x-versions;  inconsistent vectors
		--		removed

		resX    :: [TC_Res]
		resX    =  concat    (map  checkAgainst xptypes)
		--	Build a list of all (OK and Bottom) results...

		----------------------
		--	"checkAgainst": --
		----------------------
		--	Checks "args" against ONE given (consistent) vector from "xptypes".
		--		Each pair (xptype, arg) is checked by "typeCheckFB_R" which can
		--		produce a multi-versioned results.  The (=&&=) function is then
		--		used to cross the versions:
		--	No type inference is done in "typeCheckAppl", so set "noTYPE" expl-
		--		icitly on the result of (=&&=).
		--	Also note that the LHS of checks performed here ("tcrcl") comes ul-
		--		timately from "eptypes" so it is already evaluated:
		--
		checkAgainst :: TCR_CL -> [TC_Res]
		checkAgainst    tcrcl  =
			setType  noTYPE ((=&&=) (zipWith tc1arg (getCList tcrcl)  args))
			where
			tc1arg :: TC_Res -> AldorFBlock -> [TC_Res]
			tc1arg    parm_tcr    arg =
--				trace ("\n===CHECKING:\n"++(show arg)++"\n===AGAINST\n"
--					++(show parm_tcr)  ++"\n===BINDS:\n"++
--					(show (getBinds tcenv'))
--					++"\n===RESULT: "++(show zzz)++"\n") zzz
--				where
				typeCheckFB_R tcenv' parm_tcr arg


----------------------
--  "mkDepStacks":  --
----------------------
--	Determine which of the "parms" are used as dependencies in type exprs of
--		subsequent parms, in order not to evaluate all "args" (that would be
--		totally wasteful! --  type-checking a function application would not
--		differ from really  applying that function!).  The "args" to be eva-
--		luated are then pushed on the stack. There may be multiple evaluati-
--		on results and hence multiple stacks returned. If the "evalDFlag" is
--		set, all "args" are assumed to be already evaluated,  so they can be
--		pushed on the stack at 0 cost. The "parms" are processed recursively
--		so if they contain nested dependent types,    all variables involved
--		get defined on the stack.
--	As for "TypeCheckAppl", "protoFBs" are either of the same length as "args"
--		or longer by 1 element:
--
mkDepStacks :: ETC_Env -> Bool-> Int -> [AldorFBlock] -> [AldorFBlock] ->
			  [EvalStack]

mkDepStacks       tcenv evalmFlag lengthDiff protoFBs args =

	mkDepStacks_R tcenv evalmFlag parms args dep_parms
	where
	--	"parms" are for function parms only, not the return type, so they may
	--		be different from the whole "protoFBs":
	--
	parms     :: [AldorFBlock]
	parms      = if   lengthDiff == 0
				 then protoFBs		-- All elements
				 else init protoFBs -- All but last elements

	--	Get all the parms (as (Index, FBNo, DefNo)) from "protoFBs",   upon
	--		which the types of other parms can depend. If this list is empty,
	--		the "protoFB"s are not from a dependent type.
	--	NB: we must use "protoFBs", not "parms", here, as the last of "proto
	--		FBs" may also be dependent on previous "protoFBs", but it is not
	--		in "parms"!
	--	Recursive invocations of "mkDepStack_R" below will propagate and re-
	--		use the "dep_parms" list computed here:
	--
	dep_parms :: [(Int, FBNo, DefNo)]
	dep_parms  = getDepParms tcenv protoFBs


----------------------
--	"getDepParms":  --
----------------------
getDepParms :: ETC_Env -> [AldorFBlock] -> [(Int, FBNo, DefNo)]

getDepParms    tcenv	  protoFBs      =  res
	where
	--	(1)
	--	First, get all UseNos involved in "ProtoFBs" type exprs (with structu-
	--		ral recursion).   XXX: the results of this analysis should better
	--		be cached for further calls to "mkDepStacks".
	--
	type_usenos  :: [UseNo]
	type_usenos  =  concat (map scanProtoType protoFBs)

	scanProtoType :: AldorFBlock -> [UseNo]

	--	For function parms, the arg must be an "FParmFB":
	scanProtoType    fb@(FParmFB  _ _ _)  =
		snd (foldFB_Gen  passFB  getUseNos  (undef,  []) fb)

	--	But for the return type, it can also be a "SimpleFB":
	scanProtoType	 fb@(SimpleFB _ _ _)  =
		snd (foldFB_Gen  passFB  getUseNos  (undef,  []) fb)

	--	Other FBs are not allowed:
	scanProtoType   fb = error2 "scanProtoType" ("Not an FParmFB: "++(show fb))

	getUseNos:: AldorFBlock -> AldorExpr -> [UseNo] -> [UseNo]
	getUseNos   _  (Identifier use_no _)    curr    =   use_no : curr
	getUseNos   _  _						curr    =   curr

	--	(2)
	--	For calculated "type_usenos", get all DefNos to which they may be bound.
	--		We can restrict ourselves to "FP_Targ"s only (but also wrapped into
	--		"AddedTarg"s!):
	--
	res :: [(Int, FBNo, DefNo)]
	res  = nub (concat (map getParmInfo type_usenos))
				 -- Will be [] if "evalmFlag" is set

	getParmInfo:: UseNo -> [(Int, FBNo, DefNo)]
	getParmInfo   useno =  catMaybes
						  (map mbParmInfo (lookupBinds (getBinds tcenv) useno))
		where
		mbParmInfo :: BindTarget -> Maybe (Int, FBNo, DefNo)
		mbParmInfo   (FP_Targ fb@(FParmFB fb_no _ _)) =
			case elemIndex fb_no all_fbnos of
				Just idx -> Just (idx, fb_no, def_no)
				Nothing  -> Nothing
			where
			--	Get "def_no" -- it always exists:
			GenFDTE  def_no _ _ _ = getFParm   fb

		mbParmInfo   (AddedTarg t)= mbParmInfo t
		mbParmInfo    _			  = Nothing

	all_fbnos :: [FBNo]
	all_fbnos =  [fb_no | FParmFB fb_no _ _ <- protoFBs]


------------------------
--	"mkDepStacks_R":  --
------------------------
--  Fills the given "EvalStack" (passed inside the "tcenv") by evaluated "args"
--		on which there are dependencies in "parms",   and produces a new set of
--		stacks. It then recurses to structural components of "parms" / "args":
--	Here "parms" always do NOT include the function return type,   so they must
--		be of the same length as "args":
--
mkDepStacks_R :: ETC_Env  -> Bool     -> [AldorFBlock] -> [AldorFBlock] ->
				 [(Int, FBNo, DefNo)] -> [EvalStack]

mkDepStacks_R    tcenv      evalmFlag    parms    args    dep_parms
	| length parms /= length args  = []
	| dep_parms    == []		   = [_evalStack tcenv]
	| otherwise					   = rnew_ess
	where
	--	Select those parms which produce definitions in the "dep_parms" list
	--		(i.e., restrict parms to those  on which there are dependencies);
	--		pair them with the resp args;     the result is "dep_defnos" and
	--		"dep_args":
	--
	dep_defnos:: [DefNo]
	dep_defnos = [def_no | (_, _, def_no) <- dep_parms]

	dep_idxs  :: [Int]
	dep_idxs  =  [i | (i, _, _)   <- dep_parms]

	dep_args  :: [AldorFBlock]
	dep_args  =  [args !!! i |  i <- dep_idxs]

	--	Evaluate all "dep_args" and cross over the results of multi-evaluation.
	--		All args are evaluated  independently in the given "tcenv", even if
	--		there are dependencies between the corresp parms' types.
	-- We assert that if the "_noEval" flag is set, "eargss" will not be requi-
	--		red at all in a lazy evaluation:
	--
	eargss    :: [[EvalRes]]
	--	Inner dim: eval versions; outer dim: args
	eargss    =  if _noEval tcenv
				 then
					error2 "mkDepStacks_R"
						   "NoEval mode, but still evaluating \"eargss\"?"
				 else
					map (EvalX.evalFB tcenv) dep_args

	xargss    :: [EvalRes_CL]
	xargss    =  crossWithAttrs  eargss
	--	Inner dim: args; outer dim:  X-versions;   vectors with inconsistent
	--		components (e.g. wrt binds) removed

	--	Produce new EvalStacks from the current stack in "tcenv":
	--
	curr_es::  EvalStack
	curr_es =  _evalStack tcenv

	new_ess:: [EvalStack]
	new_ess =
		if  evalmFlag
		then
			--	Don't use the evaluated args (so they will NOT be actually
			--		evaluated at all),  push the original ones:
			let
				dep_Oargs :: [EvalRes]
				dep_Oargs =  [fb2er (getBinds tcenv) (args !!! i) |
							  i  <- dep_idxs]
			in
				[pushES curr_es dep_defnos dep_Oargs]
		else
			--	Produce multiple stack versions, each for a particular vector
			--		of evaluated args:
			--
			case xargss of
				[] ->
						debug_msg    tcenv
						("\nWARNING: Evaluation of dependent args yielded []"++
						 "\nPARMS  : "++(show parms)++
						 "\nARGS   : "++(show args )++"\n")
						[]
				_  ->
					[pushES curr_es dep_defnos (getCList xrs) | xrs <-xargss]

	--
	--	That is not all!     If any of the parms have "complex" types which may
	--		themselves be dependent, this function is to be applied RECURSIVELY,
	--		and the results merged with "new_ecs":
	--
	rnew_ess ::  [EvalStack]
	rnew_ess  =  recDepTypes new_ess

	--	Find all "complex"  parm types (we can do that for "TypeVal"s only) and
	--		the FIELDS (if available) of their corresp actual args.   We do not
	--		verify here that the number of parm and actual fields is the same--
	--		a recursive call to "mkDepStacks_R" will do that:
	--
	complexPairs:: [([AldorFBlock], [AldorFBlock])]
	--			   Type Fields      Arg Fields

	complexPairs=  foldl findComplexTypes  [] (zip parms args)

	findComplexTypes :: [([AldorFBlock], [AldorFBlock])] ->
						 ( AldorFBlock,   AldorFBlock)   ->
						[([AldorFBlock], [AldorFBlock])]

	findComplexTypes    curr (fp, arg) =
		--
		--	NB: "fp" MUST be an "FParmFB" here -- already verified above:
		--
		case getFParm fp of
			--	(a)
			--	"RecordT" encountered: the actual must be "RecordVal" or
			--		"RecordCtor":
			--
			GenFDTE _ _ (Literal (TypeVal  (RecordT    t_fields))) _ ->
				case arg of
					SimpleFB   _ (Literal  (RecordVal  arg_vals))  _ ->
						(t_fields, map (simpleFB . Literal)  arg_vals) : curr

					SimpleFB   _ (RecordCtor arg_fbs) _   ->
						(t_fields,  arg_fbs) : curr
					_								      ->  curr

			--	(b)
			--	"FunctionT" encountered: the actual must be "FunctionCall":
			--
			GenFDTE _ _ (Literal (TypeVal   (FunctionT fparms _))) _ ->
				case arg of
					SimpleFB _   (FunctionCall _ fargs) _ ->
						(fparms, fargs) : curr
					_								      ->  curr

			--	(c)
			--	Nothing more: in particular, "UnionT"s cannot be dependent:
			_	->  curr

	--	We can now define "recDepTypes":
	--	For a given stack (result already obtained in "new_ecs"), it must call
	--		"mkDepStacks_R"  on all "complexPairs"  and  further  populate the
	--		current stack (possibly producing multiple versions of the result):
	--
	recDepTypes  :: [EvalStack]->  [EvalStack]
	recDepTypes  ess = foldl rdtAction ess complexPairs

	rdtAction    :: [EvalStack]-> ([AldorFBlock], [AldorFBlock])->
					[EvalStack]

	rdtAction    curr_ess     (parm_fbs, arg_fbs) =
		concat [let
					tcenv' :: ETC_Env
					tcenv' =  tcenv {_evalStack  = es}
				in
					mkDepStacks_R (tcenv') evalmFlag
							      parm_fbs arg_fbs (getDepParms tcenv' parm_fbs)
			   | es <-curr_ess]

	--	Uff... All done???


-- ==================================== --
--	Functions extracted from "ETC_Env":  --
-- ==================================== --
--	NB: Don't export them, or they will conflict with their real definitions!
--
typeCheckFB_T  :: ETC_Env -> AldorType -> AldorFBlock   -> [TC_Res]
typeCheckFB_T     tcenv      lhs   rhs = (_tcFB_T   tcenv) tcenv lhs rhs

typeCheckExpr_T:: ETC_Env -> AldorType -> AldorExpr     -> [TC_Res]
typeCheckExpr_T   tcenv      lhs   rhs = (_tcExpr_T tcenv) tcenv lhs rhs

typeCheckFB_E  :: ETC_Env -> AldorExpr -> AldorFBlock   -> [TC_Res]
typeCheckFB_E     tcenv      lhs   rhs = (_tcFB_E   tcenv) tcenv lhs rhs

typeCheckFB_R  :: ETC_Env -> TC_Res    -> AldorFBlock   -> [TC_Res]
typeCheckFB_R     tcenv      lhs   rhs = (_tcFB_R   tcenv) tcenv lhs rhs


-- =========== --
--	"error2":  --
-- =========== --
error2:: String -> String -> a
error2 = errorM    modName

