-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--								 "TypeCheckTER.hs":							 --
--		 Type-Checking Functions WITH Known Target ("_T", "_E" or _R")		 --
--							  L.A.Timochouk, 2001--2002						 --
-- ========================================================================= --
--	For all functions exported by this module, the pre-condition is that the
--		LHS (the type-checking Target) is already normalised, unlike the RHS
--		of "_N" functions:
--
module TypeCheckTER
	   (
			typeCheckExpr_T,   typeCheckFB_T,
			typeCheckExpr_E,   typeCheckFB_E,
			typeCheckExpr_R,   typeCheckFB_R
	   )
where

import DataTypes
import ParseAST
import NameBinds
import AlphaDelta
import EvalLib
import qualified EvalX
import TypeCheckLib
import TypeCheckN

import Maybe
import IOExts
import List

modName:: String
modName = "TypeCheckTER"


-- =================== --
--	"typeCheckExpr_T": --
-- =================== --
--	The LHS (the type-checking Target) is a Type constant. In many cases, type-
--		checking can then be performed by checking the RHS agaist no target (by
--		the corresp "_N" function), inferring the type of the RHS and comparing
--		it with the LHS for AD-equivalence ("typeCheckViaInfer*" functions). In
--		those cases when type inference is likely to fail, "more direct" checks
--		are performed.
--	Even in those cases when type-checking is done by inference, we would typi-
--		cally write 2 equations:   one for a restricted LHS which CAN match the
--		inferred type of the RHS,  and the generic onec when the LHS ccertainly
--		does NOT match the RHS's type. This is for efficiency -- to prevent the
--		system from inferring the type of the RHS when it is  immediately clear
--		that the check will fail anyway.
--
typeCheckExpr_T:: ETC_Env -> AldorType -> AldorExpr -> [TC_Res]

------------------------------
--	(T1) RHS is "VoidExpr": --
------------------------------
--  "VoidExpr" always evaluates to "TrivVal" of type "TrivT":
--
typeCheckExpr_T   tcenv  TrivT  VoidExpr =   okWith tcenv TrivT
typeCheckExpr_T   _      _      VoidExpr =   []

------------------------------
--	(T2) RHS is "Literal" : --
------------------------------
--	(a) "TrivVal":
--
typeCheckExpr_T   tcenv  TrivT  (Literal TrivVal)  = okWith tcenv TrivT
typeCheckExpr_T   _    _        (Literal TrivVal)  = []

--
--	(b) "TypeVal":
--		Any "TypeVal" against "TypeT": always OK (if the RHS is consistent):
--
typeCheckExpr_T   tcenv  TypeT rhs@(Literal (TypeVal _)) = resT
	where
	resN, resT :: [TC_Res]

	--	Check the RHS for internal consistency:
	resN = typeCheckExpr_N   (tcenv {_inferTypes = False}) rhs

	--  XXX Currently, we don't do anything more here -- if the RHS is an ADT
	--		object, we do NOT infer its most general Category type,  but just
	--		install the LHS (TypeT) -- is it OK?
	resT = setType (maybeType tcenv TypeT) resN

--
--	(b')
--	"TypeVal ADT" against "CategoryT": need to compare interfaces:
--		 (1) All "cdt" entries must also be present in (implemented by) the
--			 "adt", with the same types;
--		 (2) If so, this Category is the inferred type for the RHS:
--
typeCheckExpr_T   tcenv lhs@(CategoryT _ _)
						rhs@(Literal (TypeVal (ADT _ _ adt))) =
	resN -&&- (b_tcrs ++ c_tcrs)
	where
	--	(0)
	--	Check the RHS for internal consistency:
	--
	resN  :: [TC_Res]
	resN  =  typeCheckExpr_N  tcenv  rhs

	--	(1)
	--	Evaluate type exprs from the "adt" (the LHS entries are assumed to be
	--		already normalised). XXX: "RenDTE"s are ignored so the type check
	--		here is NOT modulo renamings:
	--
	adtes :: [(String, DefTableEntry)]
	adtes =  concat [[(name, dte) | dte <- dtes] | (name, dtes) <- dt2list adt]

	aress :: [[(DefNo, String, EvalRes)]]
	aress =  [[(def_no,  name, te_res) |
				te_res <-maybeEval type_expr] |
			   (name, GenDTE def_no type_expr _)  <- adtes]
	--	In "aress":
	--	Inner dim: eval versions; outer dim: different DTEs

	--	"maybeEval": evalues the expr,  or constructs a synthetic result
	--		if evaluations are not allowed:
	maybeEval :: AldorExpr -> [EvalRes]
	maybeEval    e
		| _noEval tcenv = [Partial_ER  (getBinds tcenv) (simpleFB e)]
		| otherwise		= EvalX.evalExpr tcenv e

	--	(2)
	--	Cross the "aress" vectors and retain those with compatible Binds
	--		for all entries:
	xress :: [(NameBinds, [(DefNo, String, EvalRes)]) ]
	xress =  foldl  consolBinds [] (crossVectors aress)
	--	In "xress":
	--	Inner dim: different DTEs with consolidated NameBinds;
	--	outer dim: crossed evaluated versions

	consolBinds :: [(NameBinds, [(DefNo, String, EvalRes)])] ->
								[(DefNo, String, EvalRes)]   ->
				   [(NameBinds, [(DefNo, String, EvalRes)])]

	consolBinds    curr      entries =
		case  consolidateList bindss of
			Just   binds  -> (binds, entries) :  curr
			Nothing		  -> curr
		where
		bindss  :: [NameBinds]
		bindss  =  [getAttr er | (_, _, er) <- entries]

	--	(3)
	--	Now split "xress" into two parts:
	--		(1) vectors each containing at least one Bottom   =>  their type-
	--			checking result is Bottom as well;
	--		(2) vectors without Bottoms => re-construct Categories from them:
	--
	bots, nbots     :: [(NameBinds, [(DefNo, String, EvalRes)])]
	(bots,nbots) =  foldl  separateBottoms ([], [])  xress

	--	The type expr for "separateBottoms" is far too tedious to type!
	--
	separateBottoms (curr_bots, curr_nbots) xres@(_, entries) =
		if	 or [isBottom er | (_, _, er) <- entries]
		then	(xres : curr_bots, curr_nbots)
		else	(curr_bots,  xres: curr_nbots)

	--	(4)
	--	"bots" are directly translated into "Bottom_TCR"s. XXX: This may not be
	--		the best thing to do, since if a Bottom occurred in a type expr of
	--		"adt" which does NOT implement any Category entries anyway, we may
	--		still be able to deduce a strict negative (or maybe even positive!)
	--		type-checking result.     However, we don't do that for the sake of
	--		simplicity:
	b_tcrs :: [TC_Res]
	b_tcrs =  [Bottom_TCR binds | (binds, _) <- bots]

	--	(5)
	--	For "nbots", re-construct Categories (NOT ENUMERATED) from them:
	--
	n_cats :: [(NameBinds, AldorType)]
	n_cats =  [(binds, CategoryT noNUMBER (reconstrDT entries)) |
			   (binds, entries) <- nbots]

	reconstrDT :: [(DefNo, String, EvalRes)] -> DefTable
	reconstrDT    entries  =  foldl attachEntry emptyDT  entries

	attachEntry:: DefTable -> (DefNo, String, EvalRes) -> DefTable
	attachEntry    curr_dt    (defno, name,   er) =
		attachFDTE curr_dt    (GenFDTE defno name (er2expr er) voidFB)

	--	(6)
	--	Now, for each of "n_cats", check that the LHS Category  is a "subcat"
	--		of the former. The inferred type (if any) is the LHS:
	c_tcrs  :: [TC_Res]
	c_tcrs  =  foldl catTCR [] n_cats

	catTCR  :: [TC_Res] -> (NameBinds, AldorType) -> [TC_Res]
	catTCR     curr		   (binds,     cat_type ) =
		case   subcat tcenv lhs cat_type of
			Nothing   ->
				--	"subcat" relation is undecided:
				(Bottom_TCR binds) : curr

			Just True ->
				--	Yes, "lhs" is a "subcat" of "cat_type":
				(OK_TCR binds (maybeType tcenv lhs)) : curr

			Just False->
				curr

	--	(7)
	--	The final result is (b_tcrs ++ c_tcrs) consoldated with "resN".


--	All other Type Literals against all other targets:  Fail:
--
typeCheckExpr_T   _		 _		   (Literal (TypeVal _))    = []

--
--	(c) "BooleanVal":
--
typeCheckExpr_T   tcenv  BooleanT  (Literal (BooleanVal _)) =
		okWith    tcenv  BooleanT
typeCheckExpr_T   _		 _		   (Literal (BooleanVal _)) = []

--
--	(d) "IntegerVal":
--
typeCheckExpr_T   tcenv  IntegerT  (Literal (IntegerVal _)) =
		okWith	  tcenv  IntegerT
typeCheckExpr_T   _	     _		   (Literal (IntegerVal _)) = []

--
--	(e) "FloatVal":
--
typeCheckExpr_T   tcenv  FloatT    (Literal (FloatVal _))   =
		okWith    tcenv  FloatT
typeCheckExpr_T   _	     _		   (Literal (FloatVal _))   = []

--
--	(f) "StringVal":
--
typeCheckExpr_T   tcenv  StringT   (Literal (StringVal _))  =
		okWith    tcenv  StringT
typeCheckExpr_T  _	     _		   (Literal (StringVal _))  = []

--
--	(g) "LambdaVal":
--		The LHS must then be a "FunctionT". The "LambdaVal" in the RHS is to be
--		checked for internal consistency, and the inferred Lambda Type compared
--		for AD-equivalence to the function type in the LHS:
--
typeCheckExpr_T	  tcenv  lhs@(FunctionT _ _) rhs@(Literal (LambdaVal _ _))  =
	typeCheckViaInfer_T  tcenv lhs rhs

typeCheckExpr_T	  _ 	 _	 (Literal (LambdaVal _ _))  = []

--
--	(h) "BuiltInOp":
--		The LHS must be a "BuiltInT". The type of the RHS is inferred and comp-
--		ared to the LHS:
--
typeCheckExpr_T   tcenv  lhs@(BuiltInT _ _) rhs@(Literal (BuiltInOp _ _ _)) =
	typeCheckViaInfer_T  tcenv lhs rhs

typeCheckExpr_T   _	     _   (Literal (BuiltInOp _ _ _)) = []

--
--	(i) "RecordVal":
--		The Record type can be dependent,  so we need  to use   "typeCheckAppl"
--		here (discarding the EvalStacks returned). The LHS must be a "RecordT":
--
typeCheckExpr_T   tcenv   lhs@(RecordT tfbs) (Literal (RecordVal vals)) = resT
	where
	resF, resT :: [TC_Res]

	resF= map snd (typeCheckAppl tcenv True 
								 tfbs  (map  (simpleFB . Literal) vals))
	--	Here "vals" are wrapped into "SimpleFB"s  in order to be passed to
	--		"typeCheckAppl" -- it introduces no inefficiency as fields are
	--		marked as already evaluated.

	--	Infer the resulting type if required:
	resT= setType (maybeType tcenv lhs) resF

typeCheckExpr_T   _		 _				(Literal (RecordVal _   )) = []

--
--	(j) "UnionVal":
--		The LHS must be a "UnionT". The Union type of the RHS cannot always be
--		inferred,    since its field selector may not be bound directly to the
--		corresp Union type field declaration, but only be AD-equivalent to ano-
--		ther Identifier which is in turn bound to that declaration.   So: just
--		try to find a compatible field    (with the same Name and the type sui-
--		table for the "val") in the LHS.   Union field names CAN in general be
--		overloaded, BUT only one overloaded field must be type-compatible with
--		each binding for the "sel" -- XXX  but we currently don't enforce this
--		restriction, since a single Union type would be inferred anyway:
--
typeCheckExpr_T    tcenv  ut@(UnionT _)
						 (Literal (UnionVal (Identifier _ [name]) val)) =
	typeCheckUnion tcenv  ut name val typeCheckVal_E
	where
	typeCheckVal_E :: ETC_Env -> AldorExpr -> AldorValue -> [TC_Res]
	typeCheckVal_E    tcenv'     type_expr    val'       =
		typeCheckExpr_E tcenv'   type_expr   (Literal val')


typeCheckExpr_T   _   (UnionT _) (Literal (UnionVal hmm _)) =
	--	The selector is not an Identifier in the required format:
	error2 "typeCheckExpr_T-UnionVal" ("Invalid selector: "++(show hmm))

typeCheckExpr_T   _		  _		 (Literal (UnionVal _ _))   =  []
	--	The LHS is not a "UnionT"...

--
--	(k) "ADT_Val":
--		Its type can be inferred accurately as it is contained in the LHS; it
--		must be an "ADT":
--
typeCheckExpr_T   tcenv  lhs@(ADT _ _ _) rhs@(Literal (ADT_Val _ _)) =
	typeCheckViaInfer_T  tcenv lhs rhs

typeCheckExpr_T   _		 _	(Literal (ADT_Val _ _))    =  []

--------------------------------
--	(T3) RHS is "Identifier": --
--------------------------------
--	Infer its type and compare the latter with the target in the LHS:
--
typeCheckExpr_T   tcenv  lhs  rhs@(Identifier _ _) =
	typeCheckViaInfer_T tcenv lhs rhs

-----------------------------
--	(T4) RHS is "AndList": --
--	(T5) ...     "OrList": --
-----------------------------
--	The target type must be Boolean. If it is, it's sufficient to check exprs
--		for internal integrity -- the corresp "_N" function  already verifies
--		that they produce a Boolean result:
--
typeCheckExpr_T   tcenv  BooleanT  rhs@(AndList _) = typeCheckExpr_N tcenv rhs
typeCheckExpr_T   _		 _			   (AndList _) = []

typeCheckExpr_T   tcenv  BooleanT  rhs@( OrList _) = typeCheckExpr_N tcenv rhs
typeCheckExpr_T   _      _             ( OrList _) = []

--------------------------------
--	(T6) RHS is "IfThenElse": --
--------------------------------
--	Type-checking through inference is NOT a good idea here, since the types of
--		the branches may not be accurately inferred. Use direct targeted checks;
--		no need for type inference and comparing types of the branches:
--
typeCheckExpr_T   tcenv  lhs  (IfThenElse  cond_fb  then_fb  else_fb) = resX
	where
	resC,  resT,  resE,  resX :: [TC_Res]

	resC = typeCheckFB_T tcenvC BooleanT cond_fb
	resT = typeCheckFB_T tcenvT lhs      then_fb
	resE = typeCheckFB_T tcenvF lhs	     else_fb

	--	We may need to enlarge NameBinds before type-checking branches (cf the
	--		corresp "_N" function):
	--
	bindsT, bindsF :: NameBinds
	bindsT  = mkCondTargs (getBinds tcenv) cond_fb  True
	bindsF  = mkCondTargs (getBinds tcenv) cond_fb  False

	tcenvC, tcenvT, tcenvF :: ETC_Env
	tcenvC  = tcenv {_inferTypes = False}
	tcenvT  = setBinds bindsT tcenvC
	tcenvF  = setBinds bindsF tcenvC

	--	Cross the results and possibly set the inferred type of the whole expr:
	--
	resX = setType (maybeType tcenv lhs) ((=&&=) [resC, resT, resE])

----------------------------------
--	(T7) RHS is "FunctionCall": --
----------------------------------
--	Use untargeted ("_N")   check for internal consistency of the call, invoked
--		through "typeCheckViaInfer*". Then it remains to check that the inferred
--		type of the call (i.e. the return type of the callable) matches the LHS:
--
typeCheckExpr_T   tcenv  lhs  rhs@(FunctionCall _ _)  =
	typeCheckViaInfer_T tcenv lhs  rhs

--------------------------------
--	(T8) RHS is "RecordCtor": --
--------------------------------
--	This is the same as "Literal RecordVal",  the only difference is that here
--		the field values are given by FBlocks, not AldorValues. Type inference
--		is NOT used here.  The "EvalStack"s   returned by  "typeCheckAppl" are
--		discarded. Note that "vfbs" may not yet be evaluated, hence "False" in
--		the call to "typeCheckAppl":
--
typeCheckExpr_T   tcenv  lhs@(RecordT tfbs) (RecordCtor vfbs) = resT
	where
	resF, resT:: [TC_Res]

	resF = map snd (typeCheckAppl   tcenv  False  tfbs  vfbs)
	resT = setType (maybeType  tcenv  lhs) resF

typeCheckExpr_T   _      _              (RecordCtor _   ) = []

--------------------------------
--	(T9) RHS is "UnionCtor":  --
--------------------------------
--	The Union type is NOT always inferrable (see also the "UnionVal" case), so
--		use the special "TypeCheckUnion" function:
--
typeCheckExpr_T    tcenv  ut@(UnionT _)
						 (UnionCtor (Identifier _ [name]) fb) =
	typeCheckUnion tcenv  ut name fb typeCheckFB_E


typeCheckExpr_T    _     (UnionT _) (UnionCtor hmm _) =
	--	The selector is not an Identifier in the required format:
	error2 "typeCheckExpr_T-UnionCtor" ("Invalid selector: "++(show hmm))

typeCheckExpr_T   _		  _			(UnionCtor _ _)   =  []
	--	The LHS is not a "UnionT"...

---------------------------------
--	(T10) RHS is "UnionCase":  --
---------------------------------
--	It only succeeds against "BooleanT":
--
typeCheckExpr_T   tcenv lhs@BooleanT    rhs@(UnionCase _ _) =
	typeCheckViaInfer_T  tcenv lhs rhs

typeCheckExpr_T   _   _ (UnionCase _ _) = []

-----------------------------------
--	(T11) RHS is "RestrictType": --
-----------------------------------
--	Further restrict the inferred type(s) (already restriction targets) to the
--		LHS target.
--	NB: No "generic" case (which always fails) here:
--
typeCheckExpr_T   tcenv  lhs  rhs@(RestrictType _ _) =
	typeCheckViaInfer_T tcenv lhs rhs

--------------------------------
--	(T12) RHS is "ConvertTo": --
--------------------------------
--	Restrict the (inferred) conversion target type to the LHS type.
--	NB: No "generic" case (which always fails) here:
--
typeCheckExpr_T   tcenv  lhs  rhs@(ConvertTo _ _)    =
	typeCheckViaInfer_T tcenv lhs  rhs

---------------------------------------
--	(T13) RHS is "Add*" or "With*":  --
---------------------------------------
--	"With*" is supposed to yield a Category, so it MUST be of type TypeT.
--	"Add*"  yields an ADT  which can be of more specific  Category  type,
--		but how to infer that type without evaluating the RHS? Currently
--		we only allow "TypeT" in the LHS, the same as the inferred type:
--
typeCheckExpr_T  tcenv  lhs   rhs@(Add1  _ _  ) =
	typeCheckViaInfer_T tcenv lhs  rhs

typeCheckExpr_T  tcenv  lhs   rhs@(Add2  _ _ _) =
	typeCheckViaInfer_T tcenv lhs  rhs

typeCheckExpr_T  tcenv  lhs   rhs@(With1 _ _  ) =
	typeCheckViaInfer_T tcenv lhs  rhs

typeCheckExpr_T  tcenv  lhs   rhs@(With2 _ _ _) =
	typeCheckViaInfer_T tcenv lhs  rhs


-- ==================== --
--	"typeCheckExpr_E":  --
-- ==================== --
--	Here the LHS is one of the Type-Forming Exprs (TFEs).
--	Currently recognised TFSs are:
--		TypeVal Literal; Identifier; IfThenElse; FunctionCall (XXX: not yet):
--
typeCheckExpr_E :: ETC_Env -> AldorExpr -> AldorExpr -> [TC_Res]

-------------------------------------
--	(E1) LHS is "TypeVal Literal": --
-------------------------------------
--	Delegete this case to the corresp "_T" function:
--
typeCheckExpr_E     tcenv    (Literal (TypeVal target_type))   rhs   =
	typeCheckExpr_T tcenv    target_type  rhs

--------------------------------
--	(E2) LHS is "Identifier": --
--------------------------------
--	Here the LHS is an unevaluated (symbolic) Identifier -- this case typically
--		occurs with dependent types. The RHS would typically be another Identi-
--		fier (eg defined with the type given by the LHS), IfThenElse, Function-
--		Call, maybe Literal. Other exprs are unlikely to occur (?) and unlikely
--		to succeed type-checking against an Identifier,    but we consider them
--		anyway.     All cases are treated via type inference, except IfThenElse
--		which allows a more direct approach.
--	NB: In general, this method can be applied  to other exprs in the LHS  (not
--		only Identifiers), but then AD-equivalence between the LHS and the inf-
--		erred RHS type is inlikely to succeed  (eg in case of "FunctionCall"s),
--		so more specific methods such as unrolling of function calls are needed:

--	(a) RHS is "IfThenElse":
--		The code here is similar to that in the corresp "_T" function:
--
typeCheckExpr_E    tcenv     lhs@(Identifier _ _)
							     (IfThenElse cond_fb then_fb else_fb) =
	resX
	where
	resI,  resC,  resT,  resE,  resX :: [TC_Res]

	--	First of all, ths LHS must be checked against "TypeT":
	--
	resI = typeCheckExpr_T tcenvC TypeT    lhs

	--	Then type-check the RHS itself: the condition must be of Boolean type,
	--		and both branches must conform to the LHS --  we DO NOT infer the
	--		type of the RHS branches, as we already know the target:
	--
	resC    = typeCheckFB_T   tcenvC  BooleanT cond_fb
	resT    = typeCheckFB_E   tcenvT  lhs      then_fb
	resE    = typeCheckFB_E   tcenvF  lhs 	   else_fb

	bindsT, bindsF :: NameBinds
	bindsT  = mkCondTargs (getBinds tcenv) cond_fb  True
	bindsF  = mkCondTargs (getBinds tcenv) cond_fb  False

	tcenvC, tcenvT, tcenvF :: ETC_Env
	tcenvC  = tcenv  {_inferTypes = False}
	tcenvT  = setBinds bindsT tcenvC
	tcenvF  = setBinds bindsF tcenvC

	--	Does the LHS Identifier need to be re-evaluated in new binds?  If it
	--		DOES evaluate further,   the result can only be Boolean True  or
	--		False, -- which is not of type "TypeT"; thus, only unevaluatable
	--		Identifiers in the LHS can really type-check;  thus, there is no
	--		point in further evaluation!

	--	Cross all the results and possibly set the inferred type of the whole
	--		expr:
	resX = setType (maybeExpr tcenv lhs) ((=&&=) [resI, resC, resT, resE])

--
--	(b) Generic RHS: Use type inference ("_E" version),   but don't forget to
--		type-check the LHS against "TypeT" first:
--
typeCheckExpr_E         tcenv  lhs@(Identifier _ _)  rhs = resX
--	if  length resX == 0
--	then
--		trace ("\nFAILED:\nLHS:\n"++(show lhs)++"\nRHS:\n"++(show rhs)++
--			   "\nBINDS:\n"++(show (getBinds tcenv)) ++
--			   "\nResL:\n"++(show resL)++"\nResR:\n"++(show resR)++"\n\n")[]
--	else
--		resX
	where
	resL,  resR,  resX :: [TC_Res]
	resL = typeCheckExpr_T    (tcenv {_inferTypes=False}) TypeT lhs
	resR = typeCheckViaInfer_E tcenv                      lhs   rhs

	resX = setType (maybeExpr tcenv lhs) (resL -&&- resR)


--------------------------------
--	(E3) LHS is "IfThenElse": --
--------------------------------
--		In that case, the RHS must also be "IfThenElse" with  EQUIVALENT cond,
--		and its "then" and "else" branches must type-check against the corresp
--		branches of the LHS.  The LHS itself must type-check against "TypeT".
--		Note that the conds are checked for STRICT (AD-) equivalence, not lo-
--		gical equivalence, -- the latter would require a theorem prover within
--		the type-checker:
--
typeCheckExpr_E     tcenv  lhs@(IfThenElse lcond_fb lthen_fb lelse_fb)
							   (IfThenElse rcond_fb rthen_fb relse_fb) =
	case adeq  (etc2ade tcenv)  lcond_fb rcond_fb of
		Bottom_ADER		->
			--	Equivalence of conds is undecided, so is the result:
			[Bottom_TCR (getBinds tcenv)]

		No_ADER         -> []
			--	Conds NOT equivalent: Check fails!

		Yes_ADER binds' -> setType (maybeExpr (setBinds binds' tcenv) lhs)
						   ((=&&=) [resL, resT, resE]  )
			--	Conds equivalent: type-check and consolidate branches. The
			--		inferred type is the whole "lhs", which still needs to
			--		be checked against "TypeT".
	where
	--	NB: XXX: "lthen_fb" and "lelse_fb" must be "SimpleFB"s or "NestedFB"s;
	--		the latter are unlikely to occur, but we still support them.  The
	--		exprs extracted from "lthen_fb" and "lelse_fb" are NOT evaluated,
	--		since we assume that the "lhs" has already been evaluated  before
	--		this function was called:
	--
	resL, resT, resE :: [TC_Res]

	resL= typeCheckExpr_T tcenv'  TypeT   lhs

	resT= concat [typeCheckFB_R   tcenvT  lthen_tcr rthen_fb |
				  lthen_tcr <- lthen_tcrs]

	resE= concat [typeCheckFB_R   tcenvF  lelse_tcr relse_fb |
				  lelse_tcr <- lelse_tcrs]

	--	Again, enlarge NameBinds for the branches, explicitly using  BOTH LHS
	--		and RHS conds (although they are equivalent).
	--	XXX: We then re-evaluate "lthen_fb" and "lelse_fb" in the LHS to make
	--		use of modified binds, BUT is it a right thing to do? The problem
	--		is that both branches of a conditional get evaluated,  albeit for
	--		the purpose of type-checking them (and that is the LHS only).
	--	In any case, no evaluation is done if the "_noEval" flag is set:
	--
	bindsT, bindsF :: NameBinds
	bindsT  = mkCondTargs
			 (mkCondTargs (getBinds tcenv) lcond_fb  True) rcond_fb True
	bindsF  = mkCondTargs 
			 (mkCondTargs (getBinds tcenv) lcond_fb False) rcond_fb False

	tcenv',   tcenvT,  tcenvF ::   ETC_Env
	tcenv'  = tcenv {_inferTypes = False}
	tcenvT  = setBinds bindsT tcenv'
	tcenvF  = setBinds bindsF tcenv'

	lthen_tcrs,   lelse_tcrs  ::   [TC_Res]
	lthen_tcrs  = extraEvalFB tcenvT lthen_fb
	lelse_tcrs  = extraEvalFB tcenvF lelse_fb

	extraEvalFB:: ETC_Env  -> AldorFBlock -> [TC_Res]
	extraEvalFB   tcenv''    branch_fb
		| _noEval tcenv'' =
			--	No re-evaluation:
			[fb2tcr tcenv'' branch_fb]
		| otherwise		  =
			--  Do re-evaluation:
			map er2tcr (EvalX.evalFB tcenv'' branch_fb)


----------------------------------
--	(E4) LHS is "FunctionCall": --
----------------------------------
--	The "callable" can be:
--
--	(1) "RecordVal" or "RecordCtor", in this case there must be just one "arg"
--		which is an Identifier, and the corresp Record field access must type-
--		check against "TypeT",  and the inferred type  of the RHS must be  AD-
--		equivalent to the LHS.
--	(2) A user-defined function of "FunctionT" type, with a "TypeT"-compatible
--		return type.
--	XXX
--	For the moment, we just infer the type of the RHS (with a special treatment
--		for IfThenElse, again) and compare it for AD-equivalence with the LHS:

typeCheckExpr_E  tcenv  lhs@(FunctionCall _ _)
						    (IfThenElse   cond_fb  then_fb  else_fb) =
		  ((=&&=)  [resL, resC, resT, resE])
	where
	--	Check LHS against "TypeT":
	resL :: [TC_Res]
	resL =  typeCheckExpr_T tcenv' TypeT  lhs

	--	Check "cond_fb" against "BooleanT":
	resC :: [TC_Res]
	resC =  typeCheckFB_T   tcenv' BooleanT cond_fb

	--	Re-evaluate LHS in extended environment for "then" and "else" branches:
	tcenv',   tcenvT,   tcenvF ::  ETC_Env
	tcenv' =  tcenv {_inferTypes = False}
	tcenvT =  setBinds  bindsT tcenv'
	tcenvF =  setBinds  bindsF tcenv'

	bindsT,   bindsF :: NameBinds
	bindsT =  mkCondTargs (getBinds tcenv) cond_fb  True
	bindsF =  mkCondTargs (getBinds tcenv) cond_fb  False

	lhssT, lhssF :: [TC_Res]
	lhssT  =  map  er2tcr (EvalX.evalExpr tcenvT  lhs)
	lhssF  =  map  er2tcr (EvalX.evalExpr tcenvF  lhs)

	--	Check branches against the re-evaluated LHSs:
	--	XXX:  Unrolling here???
	--
	resT, resE :: [TC_Res]
	resT   =  concat [typeCheckFB_R tcenvT lhsT then_fb | lhsT <- lhssT]
	resE   =  concat [typeCheckFB_R tcenvF lhsF else_fb | lhsF <- lhssF]

	--	Finally, all the results are crossed over (see above)

--
--	Generic case:
--	Infer the type of the RHS and check it for AD-equivalence with the LHS.
--	XXX: LHS unrolling is to be implemented here:
--
typeCheckExpr_E  tcenv  lhs@(FunctionCall _ _) rhs =
--	trace ("\nLHS: "   ++(show lhs)++
--		   "\nRHS: "   ++(show resRHS)++
--		   "\nRESULT: "++(show resAD) ++"\n")
	resAD
	where
	resRHS:: [TC_Res]
	resRHS=  typeCheckExpr_N (tcenv {_inferTypes = True}) rhs

	resAD :: [TC_Res]
	resAD =  [jres | Just jres <- map tcAD resRHS]

	--
	--	"tcAD":
	--
	tcAD  :: TC_Res -> Maybe TC_Res
	tcAD     bot@(Bottom_TCR _)      = Just bot

	tcAD     (OK_TCR binds  j@[Just  inf_type_expr]) =
		case  adeq  (etc2ade tcenv') lhs inf_type_expr of
			Yes_ADER binds'->
				if    _inferTypes  tcenv
				then  Just (OK_TCR binds' j)
				else  Just (OK_TCR binds' noTYPE)
			No_ADER		   ->
				Nothing
			Bottom_ADER	   ->
				Just  (Bottom_TCR binds)
		where
		tcenv' :: ETC_Env
		tcenv' =  setBinds binds tcenv

	tcAD     (OK_TCR _ _) = Nothing  		-- Type not inferred properly


----------------------------
--	(E5) All other cases: --
----------------------------
-- They fail: the LHS is not a type-forming expr:
--
typeCheckExpr_E  _ _ _ = []


-- ========================= --
--	Targeted "typeCheckFB*": --
-- ========================= --
--	The "_T" and "_E" forms of "typeCheckFB*"  are similar to each other  (and
--		somewhat different from the "N_" form), so they share a common implem-
--		entation.  Basically, it just invokes the corr "typeCheckExpr*" fnctn:
--	The "generic" "typeCheckFB" can be implemented either  on a disjoint union
--		of AldorType and AldorExpr,    or on a class of which the latter types
--		are instances. However, the latter approach leads to problems with GHC
--		(since GHC requires that all functions belonging to a mutually-recursv
--		group share the same class constraints, which is not possible for "_N"
--		functions -- they don't have any constraints!). We thus use "TorE":
--
-------------------------
--	"typeCheckFB_Gen": --
-------------------------
typeCheckFB_Gen:: ETC_Env -> TorE -> AldorFBlock -> [TC_Res]
--
--	(1) "VoidFB":
--		Always fails type-checking, as it has no valid type associated with it.
--		Furthermore, we produce an error here  (like in the resp "_N" functn):
--
typeCheckFB_Gen   _         _    hmm@(VoidFB	_)  =
		error2 "typeCheckFB_Gen" ((show hmm)++" encountered")

--
--	(2) "SimpleFB":
--		Type-check the expr contained in it:
--
typeCheckFB_Gen   tcenv     lhs			(SimpleFB _ expr _) =
		case lhs of
			T t  -> typeCheckExpr_T tcenv t expr
			E e  -> typeCheckExpr_E tcenv e expr

--
--	(3) "NestedFB":
--		Type-check its "innerFB" recursively, and also type-check the DefTable;
--		"innerFB"  is skipped if it's a "VoidFB" (as in "_N"):
--
--
typeCheckFB_Gen   tcenv     lhs         (NestedFB _ dt innerFB _) =
		case   innerFB of
			VoidFB _ -> resD
			_		 -> resX
		where
		resI, resD, resX ::  [TC_Res]

		resI= typeCheckFB_Gen tcenv' lhs innerFB
		resD= typeCheckDT	  tcenv' dt  -- Or use just "tcenv" here

		tcenv' :: ETC_Env
		tcenv' =  tcenv {_inferTypes = False}
		resX   =  setType inf_type (resI -&&- resD)

		inf_type  :: InferredType
		inf_type  =  case lhs of
			T t   -> maybeType tcenv t
			E e   -> maybeExpr tcenv e

--
--	(4) "FParmFB":
--		Should never occur, as it has no type by itself. Produce an error here:
--
typeCheckFB_Gen   _		    _		hmm@(FParmFB _ _ _) =
		error2 "typeCheckFB_Gen" ("FParmFB encountered: "++(show hmm))


------------------------
--	"typeCheckFB_T":  --
--	"typeCheckFB_E":  --
------------------------
--	We can now define these functions as specialisations of "typeCheckFB_Gen":
--
typeCheckFB_T :: ETC_Env -> AldorType -> AldorFBlock -> [TC_Res]
typeCheckFB_T    tcenv     t            fb          =
--	trace ("\nLHS: "   ++(show t)++
--		   "\nRHS: "   ++(show fb)++
--		   "\nRESULT: "++(show res)++"\n\n\n")
--	res
--	where
	typeCheckFB_Gen    tcenv  (T t) fb


typeCheckFB_E :: ETC_Env -> AldorExpr -> AldorFBlock -> [TC_Res]
typeCheckFB_E    tcenv     e			fb			=
	typeCheckFB_Gen    tcenv  (E e) fb


-- ================= --
--	"typeCheck*_R":  --
-- ================= --
--	Checking against targets represented as "TC_Res"s (through their inferred
--		types).  If the target is formally "OK" but its inferred type was not
--		set properly, the result is [].    If the target is Bottom, so is the
--		result.
--------------------------
--	"typeCheckExpr_R":  --
--------------------------
typeCheckExpr_R :: ETC_Env -> TC_Res -> AldorExpr  -> [TC_Res]

typeCheckExpr_R    _         bot@(Bottom_TCR _)     _	  =  [bot]

typeCheckExpr_R    tcenv     ok @(OK_TCR _ [Just type_expr]) expr =
	typeCheckExpr_E	   (updateTCE ok tcenv)      type_expr   expr

typeCheckExpr_R    _         (OK_TCR _ _)			_	  =  []

-----------------------
--	"typeCheckFB_R": --
-----------------------
typeCheckFB_R   :: ETC_Env -> TC_Res -> AldorFBlock-> [TC_Res]

typeCheckFB_R      _         bot@(Bottom_TCR _)     _     =  [bot]

typeCheckFB_R      tcenv	 ok @(OK_TCR _ [Just type_expr]) fb   =
--	trace ("\nLHS: "   ++(show type_expr)++
--		   "\nRHS: "   ++(show fb)++
--		   "\nBINDS: " ++(show binds)++
--		   "\nRESULT: "++(show res)++"\n\n\n")
--	res
--	where
	typeCheckFB_E  (updateTCE ok tcenv) type_expr fb

typeCheckFB_R	   _		 (OK_TCR _ _)			_	  =  []


-- =============================== --
--	Other Type-Checking Functions: --
-- =============================== --
----------------------------
--	"typeCheckViaInfer*": --
----------------------------
--	Applicable when the type of the RHS  can easily and accurately  be inferred.
--		It compares that inferred type for AD_equality with the LHS target. The
--		LHS can be either an AldorType or an AldorExpr.  These cases share much
--		of the common code, so we first define a generic version on "TorE":
--
typeCheckViaInfer_Gen:: ETC_Env ->  TorE-> AldorExpr -> [TC_Res]

typeCheckViaInfer_Gen	tcenv      lhs			rhs	=  resI
	where
	--	Infer the type of the RHS, if possible:
	resN :: [TC_Res]
	resN =  typeCheckExpr_N (tcenv {_inferTypes = True}) rhs

	--	Select those of the above results which match the LHS.  Perform the
	--		equivalence check for each inferred type (the "eqInfType" auto-
	--		matically updates the env prior to doing the check itself):
	--
	resE :: [TC_Res]
	resE =  case zzz of
				[] -> trace
					 ("NOT EQUIVALENT:\n"++
					 (show rhs) ++"\nOF TYPE\n" ++
					 (show resN)++",\nAND\n"++
					 (show lhs)++"\n\n")
					 []
				_  -> 
--					trace ("EQUIVALENT:\n"++
--					 (show rhs) ++"\nOF TYPE\n" ++
--					 (show resN)++",\nAND\n"++
--					 (show lhs)++"\n\n")
					 zzz
				where
				xxx = map (eqInfType tcenv lhs) resN		
				zzz = [tcr | Just tcr <- xxx]

--		[tcr | Just tcr <- map (eqInfType tcenv lhs) resN]

	-- If we are not requested to propagate the inferred types, reset them:
	resI :: [TC_Res]
	resI =  if _inferTypes tcenv  then resE else setType noTYPE resE

--
--	Now the specialisations:
--
typeCheckViaInfer_T  :: ETC_Env ->  AldorType -> AldorExpr -> [TC_Res]
typeCheckViaInfer_T     tcenv      t            rhs       =
	typeCheckViaInfer_Gen          tcenv  (T t) rhs

typeCheckViaInfer_E  :: ETC_Env ->  AldorExpr -> AldorExpr -> [TC_Res]
typeCheckViaInfer_E     tcenv      e            rhs       =
	typeCheckViaInfer_Gen          tcenv  (E e) rhs


-------------------
--	"eqInfType": --
-------------------
--	For a given "TC_Res", determine whether the InferredType contained in it is
--		AD-equivalent to the specified target type. "Bottom"s are considered to
--		be conformant (!) in order not to drop them prematurely,  but "TC_Res"s
--		with "noTYPE" inferred produce a negative result.
--	Return value: for a positive result, "Just" the original "TC_Res" is retur-
--		ned. If the equivalence check hits Bottom, "Just Bottom"  is  returned.
--		If the negative result is produced, "Nothing" is returned:
--
eqInfType :: ETC_Env -> TorE -> TC_Res -> Maybe TC_Res

eqInfType    _         _       bot@(Bottom_TCR _) = Just bot

eqInfType    tcenv     target  tcr@(OK_TCR   _ _) =
	case getInferredType tcr of
		Nothing			 ->
			Nothing	  -- Not a properly inferred type, check fails

		Just inf_type	 ->
			--	HERE we check the inferred (RHS) and the target (LHS) types for
			--		AD-equivalence.      There are SPECIAL EXTRA RULES here for
			--		Categories.
			--	NB: The special rules declared here SHOULD NOT be moved into the
			--		definitions of the "adeq" functions themselves,  as the are
			--		ASYMMETRICAL (e.g. LHS=TypeT, RHS=CategoryT) and thus valid
			--		for type-checking only, NOT for the general AD-equivalence:
			--
			case (target, inf_type) of
				--
				--	(1) Category-valued RHS against TypeT: OK:
				--
				(T TypeT, T (CategoryT _ _)) ->
					Just tcr

				(T TypeT, E (With1 _ _))     ->
					Just tcr

				(T TypeT, E (With2 _ _ _))   ->
					Just tcr

				--
				--	(2) Two Categories: do they need to be equivalent, or the
				--		inferred Category type is allowed to be more general
				--		than the Target? Do the latter -- allow for a "subcat"
				--		relation:
				(T cat1@(CategoryT _ _), T cat2@(CategoryT _ _)) ->
					case subcat tcenv cat1 cat2 of
						Just True  ->
							Just tcr
						Just False ->
							Nothing   -- "Nothing" for a negative result!

						Nothing    -> -- Bottom:
							Just (Bottom_TCR (getBinds tcenv'))
				--
				--	(3) ADT-valued RHS (that is, the RHS producing an ADT, not
				--		an ADT_Val !) against TypeT: also OK:
				--
				(T TypeT, T (ADT  _ _ _)) ->
					Just tcr

				(T TypeT, E (Add1 _ _  )) ->
					Just tcr

				(T TypeT, E (Add2 _ _ _)) ->
					Just tcr

				--
				--	BTW: It CANNOT happen  that  we have unevaluated
				--		 E (Literal (TypeVal TypeT)) in the LHS...

				--	(4) All other cases: do AD-test:
				--
				_ ->
					case adeq adenv inf_type target of
						Bottom_ADER	    ->
							Just (Bottom_TCR (_bindsR tcr))
						    -- Again, Bottom is retained

						Yes_ADER binds' ->
							Just (setAttr binds' tcr)
							-- Equivalence OK, keep "tcr" with "binds'":

						No_ADER			->
							trace ("\nAD-EQ BETWEEN "++(show inf_type)++
								   " AND "++(show target)++" FAILED\n")
							Nothing	  -- Equivalence failed
	where
	getInferredType :: TC_Res -> Maybe TorE
	getInferredType   (OK_TCR _ [(Just (Literal (TypeVal t)))]) = Just  (T t)
	getInferredType   (OK_TCR _ [ Just e ])					    = Just  (E e)
	getInferredType   _											= Nothing

	tcenv' :: ETC_Env
	tcenv' =  updateTCE tcr tcenv
	-- NB: "tcr" data are merged with "tcenv" before constructing "adenv"

	adenv  :: ADE_Env
	adenv  =  etc2ade (updateTCE tcr tcenv)


-----------------
--	"subcat":  --
-----------------
--	Tests whether the LHS Category provides an interface which is a subset of
--		that   in the RHS.  Both LHS and RHS are assumed to be already norma-
--		lised -- no evaluation is performed here.   If the result is Nothing,
--		a Bottom was encountered in AD-equivalence check:
--
subcat :: ETC_Env-> AldorType ->    AldorType -> Maybe Bool

subcat    tcenv    (CategoryT _ dt1)  (CategoryT _ dt2) = mres
	where
	--	Convert the LHS into a list of entries, XXX ignoring any renamings:
	--
	les:: [(String, AldorExpr)]
	les = concat [[(name, type_expr) |
				 GenDTE _ type_expr _ <- n_entries] |
				   (name, n_entries)  <- dt2list dt1]

	--	For all "les" names, search for the corresp entry in the RHS with an
	--		AD-equivalent type expr:
	--
	mress :: [Maybe Bool]
	mress =  map findInRHS les

	mres  :: Maybe Bool
	mres  =  if  any isNothing mress
			 then  Nothing
			 else  Just (and (map fromJust mress))

	findInRHS :: (String, AldorExpr) -> Maybe Bool
	findInRHS    (name,   ltype)     =
		case lookupDT dt2 name of
			[]    ->
				--	The "name" is not found at all in the RHS -- failure!
				Just False
			rdtes ->
				--	At least one of the "rdtes" must have a compatible type
				--		expr:
				let
					rtypes:: [AldorExpr]
					rtypes = [rt | GenDTE _ rt _ <- rdtes]

					aders :: [ADE_Res]
					aders  = [adeq adenv ltype rtype | rtype <- rtypes]

					hasTrue  :: Bool
					hasTrue  = length [ader | ader@(Yes_ADER _) <- aders] /= 0
				in
					--	Even if there were Bottoms here, just one True match
					--		is sufficient for a positive result:
					if  hasTrue
					then
						Just True
					else
						if  any isBottom aders
						then
							--	No "True", but there was a Bottom -- undecided:
							Nothing
						else
							--	Strictly negative:
							Just False
	adenv :: ADE_Env
	adenv  = etc2ade tcenv


subcat _ _ _ =  error2 "subcat" "Argument is not a Category"


-------------------------
--	"typeCheckUnion":  --
-------------------------
typeCheckUnion :: ETC_Env -> AldorType -> String -> a ->
				 (ETC_Env -> AldorExpr -> a -> [TC_Res]) -> [TC_Res]

typeCheckUnion    tcenv   ut@(UnionT flds) name a tcFunc =  resT
	where
	resT :: [TC_Res]
	resT =  [liftUF tcr | tcr <- nub (concat compFields)]

	compFields :: [[TC_Res]]
	--	Outer dim: fields, inner dim: results for each field
	compFields =  map compUF flds

	compUF :: AldorFBlock -> [TC_Res]
	compUF    fp@(FParmFB _ _ _) =
		if    fname == name
		then  -- Field name OK, type-check "val" against the "type_expr":
			  tcFunc tcenv type_expr a
		else  -- Field name is wrong:
			  []
		where
		GenFDTE _ fname type_expr (VoidFB _) = getFParm fp

	compUF hmm = error2 "typeCheckExpr_T--UnionT"
						("Union Field is not an FParmFB: "++(show hmm))

	liftUF :: TC_Res -> TC_Res
	liftUF	  (OK_TCR binds _)   = OK_TCR binds (maybeType tcenv ut)
	liftUF    bot@(Bottom_TCR _) = bot


typeCheckUnion _ _ _ _ _  =  []


-- =========== --
--	"error2":  --
-- =========== --
error2:: String-> String -> a
error2 = errorM   modName

